#include "XMLIO.hh"
#include <sstream>

using namespace BIAS;
using namespace std;

XMLIO::XMLIO() {

  isInitialized_ = false; 
  RootNode_ = NULL;
  CurrNode_ = NULL;

}


XMLIO::~XMLIO() {
  
  if (isInitialized_) {
    // Cleanup function for the XML library.
    xmlCleanupParser();
    xmlFreeDoc(xmlDoc_);
  }
}


void XMLIO::clear() {
 
  if (isInitialized_) {
    isInitialized_ = false; 
    xmlCleanupParser();
    xmlFreeDoc(xmlDoc_);
  }
  RootNode_ = NULL;
  CurrNode_ = NULL;

}


xmlNodePtr XMLIO::create(const std::string& RootNodeName) {

	if (xmlValidateName(BAD_CAST RootNodeName.c_str(), 0)!=0) {
    BIASERR("'" << RootNodeName << 
            "'is no valid node name, maybe contains blanks??\n");
    return NULL;
  }

  // Free the all existing XML pointers associated to the current document
  if (isInitialized_) {
    xmlCleanupParser();
    xmlFreeDoc(xmlDoc_);
    isInitialized_=false;
  }

  // reate a new XML DOM tree, to which the XML document will be  written
  xmlDoc_ = xmlNewDoc(BAD_CAST XML_DEFAULT_VERSION);
  if (xmlDoc_ == NULL) {
    BIASERR("Error creating the xml document tree\n");
    return NULL;
  }

  // Create a new XML node, to which the XML document will be appended
  CurrNode_ = RootNode_ = 
    xmlNewDocNode(xmlDoc_, NULL, BAD_CAST RootNodeName.c_str(), NULL);
  if (RootNode_ == NULL) {
    xmlFreeDoc(xmlDoc_);
    BIASERR("Error creating the root node\n");
    return NULL;
  }

  // Make ELEMENT the root node of the tree
  xmlDocSetRootElement(xmlDoc_, RootNode_);

  isInitialized_ = true;

  return RootNode_;

}


xmlNodePtr XMLIO::addChildNode(const xmlNodePtr ParentNode, 
                               const std::string& NewNodeName) {

	if (xmlValidateName(BAD_CAST NewNodeName.c_str(), 0)!=0) {
    BIASERR("'" << NewNodeName << 
            "'is no valid node name, maybe contains blanks??\n");
    return NULL;
  }

  if (!isInitialized_) {
    BIASERR("Not initialized: Please call createXML/loadXML first\n");
    return NULL;
  }

  CurrNode_ = xmlNewChild(ParentNode, NULL, 
                          BAD_CAST NewNodeName.c_str(), NULL);
  if (CurrNode_ == NULL) {
    BIASERR("Error creating the child node\n");
    return NULL;
  }

  return CurrNode_;

}

void XMLIO::addAttribute(const xmlNodePtr Node, 
                         const std::string& AttributeName, 
                         const int AttributeValue) {
  stringstream stst;
  stst << AttributeValue;
  addAttribute(Node,AttributeName,stst.str());

}

void XMLIO::addAttribute(const xmlNodePtr Node, 
                         const std::string& AttributeName, 
                         const double AttributeValue) {
  stringstream stst;
  stst << setprecision(10);
  stst << AttributeValue;
  addAttribute(Node,AttributeName,stst.str());

}
void XMLIO::addAttribute(const xmlNodePtr Node, 
                         const std::string& AttributeName, 
                         const std::string& AttributeValue) {

	if (xmlValidateName(BAD_CAST AttributeName.c_str(), 0)!=0) {
    BIASERR("'" << AttributeName << 
            "'is no valid attribute name, maybe contains blanks??\n");
    return;
  }

  if (!isInitialized_) {
    BIASERR("Not initialized: Please call createXML/loadXML first\n");
    return;
  }

  xmlAttrPtr attPtr;
	attPtr = xmlSetProp(Node, BAD_CAST AttributeName.c_str(), 
                      BAD_CAST AttributeValue.c_str());
  if (attPtr == NULL) {
    BIASERR("Error creating the child node\n");
    return;
  }

}


void XMLIO::addContent(const xmlNodePtr Node, const std::string& Content) {
	
  if (!isInitialized_) {
    BIASERR("Not initialized: Please call createXML/loadXML first\n");
    return;
  }

  xmlNodeAddContent(Node, BAD_CAST Content.c_str());
  
}

void XMLIO::addContent(const xmlNodePtr Node, const double Content) {
	
  if (!isInitialized_) {
    BIASERR("Not initialized: Please call createXML/loadXML first\n");
    return;
  }

  stringstream ss;
  ss << setprecision(10);
  ss << Content;
  xmlNodeAddContent(Node, BAD_CAST ss.str().c_str());
  
}

void XMLIO::addContent(const xmlNodePtr Node, const int Content) {
	
  if (!isInitialized_) {
    BIASERR("Not initialized: Please call createXML/loadXML first\n");
    return;
  }

  stringstream ss;
  ss << Content;
  xmlNodeAddContent(Node, BAD_CAST ss.str().c_str());
  
}

void XMLIO::addComment(const xmlNodePtr Node, const std::string& Comment) {

  if (!isInitialized_) {
    BIASERR("Not initialized: Please call createXML/loadXML first\n");
    return;
  }
  xmlNodePtr myNode;
  myNode = xmlNewComment(BAD_CAST Comment.c_str());
  xmlAddChild(Node, myNode);

}


int XMLIO::write(const std::string& Filename,
                 bool AutoAddCompressionSuffix) const {
  
  if (!isInitialized_) {
    BIASERR("Not initialized: Please call createXML/loadXML first\n");
    return -1;
  }
  string extFileName = Filename;
  if ((xmlGetDocCompressMode(xmlDoc_)>0) && AutoAddCompressionSuffix)
    extFileName+=".z" ;

  if (xmlSaveFormatFileEnc(extFileName.c_str(), xmlDoc_, 
                           XML_STD_ENCODING, 2)<0) {
    BIASERR("Error saving XML-File");
    return -1;
  }

  return 0;

}


xmlNodePtr XMLIO::read(const std::string& Filename) {
  
  // Free the all existing XML pointers associated to the current document
  if (isInitialized_) {
    xmlCleanupParser();
    xmlFreeDoc(xmlDoc_);
    isInitialized_=false;
  }
    
  xmlDoc_ = xmlParseFile(Filename.c_str());
  if (xmlDoc_ == NULL) {
    BIASERR("Error parsing the xml document, was not well formed: " 
            << Filename << "\n" );
    return NULL;
  }

  RootNode_ = CurrNode_ =  xmlDocGetRootElement(xmlDoc_);
  isInitialized_=true;
  
  return RootNode_;

}


xmlNodePtr XMLIO::getFirstChild(const xmlNodePtr ParentNode) {

  CurrNode_ = ParentNode->children;
  while (CurrNode_!=NULL) {
    if (CurrNode_->type == XML_ELEMENT_NODE) {
      return CurrNode_;
    }
    CurrNode_ = CurrNode_->next;
  } 
  return NULL;
   
}


xmlNodePtr XMLIO::getNextChild() {
  return getNextChild(NULL);
}


xmlNodePtr XMLIO::getNextChild(const xmlNodePtr SiblingNode) {

  if (SiblingNode!=NULL) {
    CurrNode_ = SiblingNode;
  }

  while (CurrNode_->next!=NULL) {
    CurrNode_ = CurrNode_->next;
    if (CurrNode_->type == XML_ELEMENT_NODE) {
      return CurrNode_;
    }
  } 

  return NULL;

}


xmlNodePtr XMLIO::getChild(const xmlNodePtr ParentNode, 
                           const std::string& ChildName) {

  xmlNodePtr myNode;
  myNode = getFirstChild(ParentNode);
  while(myNode!=NULL) {
    if (ChildName==getNodeName(myNode)) {
      return myNode;
    }
    myNode = getNextChild();
  }

  return NULL;

}


std::string XMLIO::getNodeName(const xmlNodePtr Node) const {

  std::string ret="";

  if ((Node!=NULL) && (Node->name!=NULL)) {
    ret = (char*) Node->name;
  }
  return ret;

}


std::string XMLIO::getNodeContentString(const xmlNodePtr Node) const {

  std::string ret="";

  if (Node->children!=NULL) {
    if (Node->children->content!=NULL) {
      ret = (char*) Node->children->content;
    }
  }  
  if (ret[0]==10) {
    ret = ret.substr(ret.find_first_of(10)+1);
  }
  if (ret[ret.size()-1]==32) {
    ret = ret.substr(0,ret.find_last_not_of(32)+1);
  }
  if (ret[ret.size()-1]==10) {
   ret = ret.substr(0,ret.size()-1);
  }

  return ret;

}

int XMLIO::getNodeContentInt(const xmlNodePtr Node) const {

  std::string s;
  s = getNodeContentString(Node);
  return atoi(s.c_str());

}

double XMLIO::getNodeContentDouble(const xmlNodePtr Node) const {

  std::string s;
  s = getNodeContentString(Node);
  return atof(s.c_str());

}

xmlAttrPtr XMLIO::getFirstAttribute(const xmlNodePtr Node) {

  CurrAttr_ = Node->properties;
  while (CurrAttr_!=NULL) {
    if (CurrAttr_->type == XML_ATTRIBUTE_NODE) {
      return CurrAttr_;
    }
    CurrAttr_ = CurrAttr_->next;
  } 
  return NULL;
  
}


xmlAttrPtr XMLIO::getNextAttribute() {
  
  while (CurrAttr_->next!=NULL) {
    CurrAttr_ = CurrAttr_->next;
    if (CurrAttr_->type == XML_ATTRIBUTE_NODE) {
      return CurrAttr_;
    }
  } 
  return NULL;

}


std::string XMLIO::getAttributeName(const xmlAttrPtr Attribute) const {

  std::string ret="";
  if (Attribute->name!=NULL) {
    ret = (char*) Attribute->name;
  }
  return ret;

}


std::string XMLIO::getAttributeValueString(const xmlAttrPtr Attribute) const {

  std::string ret="";
  if (Attribute->children!=NULL) {
    if (Attribute->children->content!=NULL) {
      ret = (char*) Attribute->children->content;
    }
  }
  return ret;

}

 
int XMLIO::getAttributeValueInt(const xmlAttrPtr Attribute) const {

  std::string s;
  s = getAttributeValueString(Attribute);
  return atoi(s.c_str());

}


double XMLIO::getAttributeValueDouble(const xmlAttrPtr Attribute) const {

  std::string s;
  s = getAttributeValueString(Attribute);
  return atof(s.c_str());

}

std::string XMLIO::getAttributeValueString(const xmlNodePtr Node, 
                                    const std::string& AttributeName) const{
  std::string ret="";
  char* pc;
  pc = (char*) xmlGetProp(Node, BAD_CAST AttributeName.c_str());
  if (pc!=NULL) {
    ret=pc;
    xmlFree(pc);
  }
  return ret;
}


int XMLIO::getAttributeValueInt(const xmlNodePtr Node, 
                         const std::string& AttributeName) const {
  std::string ret="";
  char* pc;
  pc = (char*) xmlGetProp(Node, BAD_CAST AttributeName.c_str());
  if (pc!=NULL) {
    ret=pc;
    xmlFree(pc);
  }
  return atoi(ret.c_str());
}


double XMLIO::getAttributeValueDouble(const xmlNodePtr Node, 
                               const std::string& AttributeName) const {
  std::string ret="";
  char* pc;
  pc = (char*) xmlGetProp(Node, BAD_CAST AttributeName.c_str());
  if (pc!=NULL) {
    ret=pc;
    xmlFree(pc);
  }
  return atof(ret.c_str());
}

int XMLIO::SetCompression(int level) {
  int ret = -1;
  if (isInitialized_) {
    // Cleanup function for the XML library.
    ret = xmlGetDocCompressMode(xmlDoc_);
    if (level>9) level = 9;
    if (level<0) level = 0;
    xmlSetDocCompressMode(xmlDoc_, level);
  } else {
    BIASERR("No document yet! Ignored compression level");
  }
  return ret;
} 
