/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef _TIMEMEASURE_H_
#define _TIMEMEASURE_H_

#ifdef WIN32

//windows implementation only returns system time, for user as well as for real time
#  include <time.h>
#  include <Base/Common/W32Compat.hh>

struct tms {
	clock_t tms_utime;  /* user time */
	clock_t tms_stime;  /* system time */
	clock_t tms_cutime; /* user time of children */
	clock_t tms_cstime; /* system time of children */
};
#else // WIN32
// Linux 
#  include <sys/time.h>
#  include <sys/times.h>
#  include <unistd.h>
#endif // WIN32


#include <fstream>
#include <errno.h>
#include <stdlib.h>
#include <limits.h>

// stl strings:
#include <string>
#include <iostream>

// debugging information
#include "Debug.hh"
#include "Error.hh"







namespace BIAS {

#ifdef WIN32
	/** @brief get "Time Stamp Counter"  on Pentium like CPU's. 
	Assembler function without prolog and epilog. 
	Should work on Intel Pentium and newer AMD CPU's, see rdtsc assembler instruction. 
	This functions can't be class member inlined because it collides with naked fastcall. 
	\todo cleanup TimeMeasure class in platform independant manner. (JW)
	\bug user/system time is implemented wrong on WIN32. see time and clock functions (JW)
	@returns the number of clock cycles since the CPU was powered up or reset. 
	@author Jan Woetzel, 12/2004 */
	// int64 has no ANSI equivalent (JW)
	//__int64 BIASGetTSC(void){ return 0; }; // fallback workaround implementation 
	__forceinline __int64 __fastcall GetTSC(){__asm rdtsc  };

	/** @brief compute the difference between two int64 cycle counts using 12 bit overflow mask .
		@author Jan Woetzel 12/2004 */
	inline __int64 GetTSCTicks(__int64 start, __int64 end) {return ((end - start) & 0xFFFFFFFFFFFF);};
#endif // WIN32





	/**  @class TimeMeasure
       @ingroup g_utils
       @brief class TimeMeasure contains functions for timing real time and 
       cpu (kernel) time.

       (c) 2002 MIP Uni Kiel,  all rights reserved. (copied in part from LFRender)

       Still to do:
       - non-static divisor from usec to sec
       - doxygen compatible full documentation

       added rdtsc assembler cycle count with int64 for Win32 jw 12/2004
       added cycle count functions for i386-processors woelk 07/2003
       added to namespace BIAS

       @author Jan Woetzel
       @status beta (03/04/2002) **/
	class DLLEXPORTAPI TimeMeasure {


	public:
		/// default constructor:
		inline TimeMeasure();

		/// print the output when the desctructor is called ?
		inline ~TimeMeasure() {};

		inline void Start();
		inline void start() { Start(); };

		inline void Stop();
		inline void stop() { Stop(); };

		inline void Print(std::ostream& os=std::cout);
		inline void print() { Print(); };

		inline void Reset();
		inline void reset() { Reset(); };

		/** return real time in usec */
		inline double GetRealTime();

		/** return real time in msec */
		inline double GetUserTime();

		/** returns the frequency in Hz based on real time (= fps)*/
		inline double GetFreq();

		/** returns the frequency in Hz based on user time (= fps)*/
		inline double GetUserFreq();

		inline void PrintRealTime(std::ostream& os=std::cout);
		inline void printRealTime() { PrintRealTime(); };

		inline void PrintUserTime(std::ostream& os=std::cout);
		inline void printUserTime() { PrintUserTime() ; };

		inline bool IsRunning();

		inline void SetTimerFile(const std::string &file);
		inline void LogToFileAndReset(const std::string &comment);
		inline void CloseTimerFile();

		/** return number of cycles between all subsequent calls 
        to Start() and Stop() since last call to Reset() 
        @author woelk */
		inline double GetCycleCount();

		/** prints the number of cycles between subsequent 
        calls to Start() and Stop()  since last call to Reset() to os  
        @author woelk */ 
		inline void PrintCycleCount(std::ostream& os=std::cout);


		/* Set *hi and *lo to the high and low order bits  of the cycle counter.  
       Implementation requires assembly code to use the rdtsc instruction.
       @author woelk (Linux), Win32 by Jan Woetzel */
		inline void ReadCycleCounter(unsigned int *hi, unsigned int *lo);


		/** returns the number of cycles between subsequent Start() and Stop() 
        calls !! implies a Reset(); !!!  
        @author woelk */
		inline double MeasureOverhead();

		/** measures the ProcessorFrequency in MHz  
        @author woelk */
		inline double MeasureProcessorFrequency();

	protected:
		timeval rtRealSum;
		timeval rtLast;
		timeval rtActual;
		timeval rtKernelSum;
		clock_t user;
		struct tms startstruct, stopstruct; // see `man 2 times`
		bool running; // boolean to tell wether the clock is running or not 


		double _dSum;

#ifdef WIN32
		__int64 _tscStart, _tscEnd, _tscTicks;
#else // WIN32
#  ifdef __i386__
		unsigned int _uiHiStart, _uiLoStart, _uiHiStop, _uiLoStop;
#  endif // __i386__
#endif // WIN32


		// have one file stream to log to for all timer instances
		static std::ofstream TimerStream_;

	}; // class -----------------------------------

	inline TimeMeasure::TimeMeasure() 
	{
		timerclear(&rtRealSum);
		timerclear(&rtKernelSum);
		timerclear(&rtLast);
		timerclear(&rtActual);

		user = 0;
		running = false;
		_dSum=0.0;

#ifdef WIN32
		_tscStart = _tscEnd = _tscTicks = 0;
		// \todo the timers are not correctly initailized on WIN32.
		/*
		startstruct.tms_utime=0;
		startstruct.tms_stime=0;
		startstruct.tms_cutime=0;
		startstruct.tms_cstime=0;
		stopstruct.tms_utime=0;
		stopstruct.tms_stime=0;
		stopstruct.tms_cutime=0;
		stopstruct.tms_cstime=0;
		*/
#else // WIN32
#  ifdef __i386__
		_uiHiStart=_uiLoStart=_uiHiStop=_uiLoStop=0;
#  endif
#endif //WIN32
	}

	inline void TimeMeasure::Start() 
	{
		if(!running) {
			running = true;
			//cerr << "start s: "<<rtLast.tv_sec<<"\tus: "<<rtLast.tv_usec<<std::endl;     

#ifdef WIN32
			if ((startstruct.tms_utime=clock()) == -1){
				BIASERR("error starting user timer "<<errno);
			}
			_tscStart = BIAS::GetTSC();

#else // WIN32
			gettimeofday(&rtLast,NULL);
			if (times(&startstruct) == -1){
				BIASERR("error starting user timer "<<errno);
			}
#  ifdef __i386__
			ReadCycleCounter(&_uiHiStart, &_uiLoStart);
#  endif //__i386__			
#endif // WIN32

			//cerr << "startstruct.tms_utime: " << startstruct.tms_utime << std::endl;
		} else {
			BIASERR("clock already started");
		}
	}

	inline void TimeMeasure::Stop() 
	{
		if(running) {

#ifdef WIN32
			_tscEnd=BIAS::GetTSC();

			if ((stopstruct.tms_utime=clock()) == -1){
				BIASERR("error stoping user timer "<<errno);
			}
			user += stopstruct.tms_utime - startstruct.tms_utime;
			rtRealSum.tv_sec = user/CLOCKS_PER_SEC;
			rtRealSum.tv_usec = user%CLOCKS_PER_SEC;
#else // WIN32
#  ifdef __i386__
			ReadCycleCounter(&_uiHiStop, &_uiLoStop);
#  endif // __i386__

			gettimeofday(&rtActual, NULL);
			if (times(&stopstruct) == -1){
				BIASERR("error stopping user timer "<<errno);
			}
			//cerr << "stop s: "<<rtActual.tv_sec<<"\tus: "<<rtActual.tv_usec<<std::endl;      
			rtRealSum.tv_sec += rtActual.tv_sec  - rtLast.tv_sec;
			long usec;
			if (rtActual.tv_usec >  rtLast.tv_usec) {
				usec = rtActual.tv_usec - rtLast.tv_usec;
			} else {
				usec = 1000000-rtLast.tv_usec+rtActual.tv_usec;
				rtRealSum.tv_sec--;
			}
			rtRealSum.tv_usec += usec;
			//     cerr << "act s: "<<rtActual.tv_sec<<"\tus: "<<rtActual.tv_usec<<endl;
			//     cerr << "lst s: "<<rtLast.tv_sec<<"\tus: "<<rtLast.tv_usec<<endl;
			//     cerr << "stop s: "<<rtRealSum.tv_sec<<"\tus: "<<rtRealSum.tv_usec<<endl;

			user += stopstruct.tms_utime - startstruct.tms_utime;
#endif // WIN32

			//cerr << "stopstruct.tms_utime: " << stopstruct.tms_utime << endl;
#ifdef WIN32
			_tscTicks = BIAS::GetTSCTicks(_tscStart, _tscEnd);
			_dSum += _tscTicks;
#else // WIN32
#  ifdef __i386__
			double count;
			if (_uiLoStop>_uiLoStart)
				count = ((double)_uiHiStop-(double)_uiHiStart)*(double)UINT_MAX+
          (double)_uiLoStop-(double)_uiLoStart;
			else
				count = ((double)_uiHiStop-(double)_uiHiStart-1)*(double)UINT_MAX+
          (double)_uiLoStop+(double)UINT_MAX-(double)_uiLoStart;
			_dSum+=count;
#  endif	//__i386__
#endif // WIN32


			running = false;
		} else {
			BIASERR("cannot stop clock befor starting it");
			BIASASSERT(running);
		} 
	}

	inline void TimeMeasure::Print(std::ostream& os)
	{
		PrintRealTime(os);
		PrintUserTime(os);
		PrintCycleCount(os);
	}

	inline void TimeMeasure::Reset()
	{
		if (running){
			BIASERR("cannot reset running TimeMeasure");
		} else {
#ifdef WIN32
			_tscStart = _tscEnd = _tscTicks = 0;
#else // WIN32
#  ifdef __i386__
			_uiHiStart=_uiLoStart=_uiHiStop=_uiLoStop=0;
#  endif // __i386__
#endif // WIN32

			_dSum=0.0;
			timerclear(&rtRealSum);
			timerclear(&rtKernelSum);
			user = 0;
		}
	}

	inline double TimeMeasure::GetRealTime() 
	{
		return (double)(rtRealSum.tv_usec + (rtRealSum.tv_sec*1E6));
	}

	inline double TimeMeasure::GetUserTime()
	{
		//cerr << "sysconf(_SC_CLK_TCK): " << sysconf(_SC_CLK_TCK) << endl;
		// see ` man 2 times`
#ifdef WIN32
		return (double)user/(double)CLOCKS_PER_SEC * 1e3;
#else // WIN32
		return (double)user/(double)sysconf(_SC_CLK_TCK) * 1e3;
#endif // WIN32
	}

	inline double TimeMeasure::GetFreq()
	{
		return 1E6/GetRealTime();
	}

	inline double TimeMeasure::GetUserFreq()
	{
		return 1E3/GetUserTime();
	}

	inline bool TimeMeasure::IsRunning()
	{
		return running;
	}

	inline void TimeMeasure::SetTimerFile(const std::string &file)
	{
		//   if (TimerStream_) TimerStream_.close();
		TimerStream_.open(file.c_str(), std::ios_base::app);
		if (!TimerStream_) {
			BIASERR("Can not open TimerFile "<<file
              <<" in TimeMeasure::SetTimerFile()")
				perror(file.c_str());
			abort();
		}
	}

	inline void TimeMeasure::LogToFileAndReset(const std::string &comment)
	{
		Stop();
		if (!TimerStream_) BIASERR("TimerStream_ not open");
		TimerStream_<<"real: "<<std::setw(10)<<GetRealTime()/1e3
                <<"\tuser: "<<std::setw(6)<<GetUserTime()
                <<"\t"<<comment<<std::endl;
		Reset();
	}

	inline void TimeMeasure::CloseTimerFile()
	{
		TimerStream_.close();
	}



	inline double TimeMeasure::GetCycleCount()
	{
		if(running) { 
			BIASERR("you called GetCycleCount() while timer is running"); 
		}
		return _dSum;
	}

	inline void TimeMeasure::PrintRealTime(std::ostream& os) 
	{
		os << "real: "<<GetRealTime()<<" us"<<std::endl;
	}

	inline void TimeMeasure::PrintUserTime(std::ostream& os) 
	{
		os << "user: "<<GetUserTime()<<" ms"<<std::endl;
	}




	inline void TimeMeasure::PrintCycleCount(std::ostream& os) 
	{
		os << "counted "<<GetCycleCount()<<" cycles"<<std::endl;
	}


#ifndef WIN32
#  ifdef __i386__
	inline 
  void TimeMeasure::ReadCycleCounter(unsigned int *hi, unsigned int *lo)
	{
		// Linux, Felix, see BIAS::GetTSC for WIN32 equivalent
		asm("rdtsc; movl %%edx,%0; movl %%eax,%1"   /* Read cycle counter */
        : "=r" (*hi), "=r" (*lo)                /* and move results to */
        : /* No input */                        /* the two outputs */
        : "%edx", "%eax");
	}
#  endif // __i386__
#endif // WIN32


	inline double TimeMeasure::MeasureOverhead()
	{
		/* Do it twice to eliminate cache effects */
		int i;
		double result;

		for (i = 0; i < 2; i++) {
			Reset();
			Start();
			Stop();
			result = GetCycleCount();
		}
		return result;
	}

	inline double TimeMeasure::MeasureProcessorFrequency()
	{
		double rate;
		int sleeptime=2;

		Start();
		sleep(sleeptime);
		Stop();
		rate = GetCycleCount() / (1e6*sleeptime);

		return rate;
	}


} // namespace 
#endif // _TIMEMEASURE_H_
