/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __EMatrixBase_hh__
#define __EMatrixBase_hh__

#include <Base/Math/Matrix3x3.hh>
#include "RMatrixBase.hh"
#include "HomgLine2D.hh"

#define EMATRIX_TYPE double

namespace BIAS {
  /** @class EMatrixBase 
      @ingroup g_geometry
      @brief describes the epipolar relationship between two images whith known
      internal camera calibration.

      @author woelk 07 2004 */
  class DLLEXPORTAPI EMatrixBase : public Matrix3x3<EMATRIX_TYPE>
  {
  public:
    inline EMatrixBase() : Matrix3x3<EMATRIX_TYPE>() { _isDecomposed=false; };
    inline ~EMatrixBase() {};

    /// see Set(Vector3<EMATRIX_TYPE>& epipole, RMatixBase& R);
    inline EMatrixBase(const Vector3<EMATRIX_TYPE>& epipole, 
                       const RMatrixBase& R)
      : Matrix3x3<EMATRIX_TYPE>() 
    { Set(epipole, R); }
    /// see Set(Vector3<EMATRIX_TYPE>& epipole, 
    ///         Vector3<EMATRIX_TYPE>& EulerAngXYZ);
    inline EMatrixBase(const Vector3<EMATRIX_TYPE>& epipole, 
                       const Vector3<EMATRIX_TYPE>& EulerAngXYZ)
      : Matrix3x3<EMATRIX_TYPE>() 
    { Set(epipole, EulerAngXYZ); };
    
    /** set it from epipole and rotation matrix
        in order to generate essential matrix from image 1 to image 2
        (i.e. generating the epipolar line in image 2, gfiven a point in 
        image 1),  use epipole in second image and rotation matrix from image 1
        to image 2 
        @author woelk 07/2004 */
    inline void Set(const Vector3<EMATRIX_TYPE>& epipole, 
                    const RMatrixBase& R);
    /** see constructor from epipole and rotation matrix
        builds rotation matrix in x, y, z order with moving axis
        see RMatrixBase
        @author woelk 07/2004 */
    inline void Set(const Vector3<EMATRIX_TYPE>& epipole, 
                    const Vector3<EMATRIX_TYPE>& EulerAngXYZ);

    
    inline double GetEpipolarErrorHomogenized(const BIAS::HomgPoint2D& P1,
                                              const BIAS::HomgPoint2D& P2);

    /** returns the cosine of the angle bewteen epipolar line 
        and match line */
    inline double GetCosAngleErrorHomogenized(const BIAS::HomgPoint2D& p1,
                                              const BIAS::HomgPoint2D& p2);

  protected:
    bool _isDecomposed;
    HomgPoint2D _epipole;
    RMatrixBase _R;

  };

  ///////////////////////////////////////////////////////////////////////////
  ///              implementation
  ///////////////////////////////////////////////////////////////////////////

  void EMatrixBase::Set(const Vector3<EMATRIX_TYPE>& epipole, 
                        const RMatrixBase& R)
  {
    Matrix3x3<EMATRIX_TYPE> e;
    HomgPoint2D he=epipole; he.Homogenize(); 
    e.SetAsCrossProductMatrix(he);
    e.Mult(R, *this);
    _isDecomposed=true;
    _epipole=he;
    _R=R;
  }
  
  void EMatrixBase::Set(const Vector3<EMATRIX_TYPE>& epipole, 
                        const Vector3<EMATRIX_TYPE>& EulerAngXYZ)
  {
    RMatrixBase R;
    R.SetXYZ(EulerAngXYZ[0], EulerAngXYZ[1], EulerAngXYZ[2]);
    Set(epipole, R);
  }
  
  double EMatrixBase::GetEpipolarErrorHomogenized(const BIAS::HomgPoint2D& P1,
                                                  const BIAS::HomgPoint2D& P2)
  {
    /*
    // old version, quite slow but readable ;-)
    
    // use F Matrix to get epipolar line in other image
    static BIAS::HomgLine2D EpiLineInImage;
    F.GetEpipolarLineImage1(P2,EpiLineInImage);
    
    // compute (perpendicular) distance of point to line and square it
    double dError1 = EpiLineInImage.DistanceSquared(P1);
    
    // do the same for the inverse direction and add error value
    F.GetEpipolarLineImage2(P1,EpiLineInImage);
    
    double dErrorAll = (dError1 + EpiLineInImage.DistanceSquared(P2));
    return dErrorAll;
    
    */
    static BIAS::HomgLine2D EL;
    
    // computation of epipolar line relying on homogenized point
    const register double *matP = GetData();	
    EL[0] = P2[0] * matP[0]
      + P2[1] * matP[3]
      + matP[6];
    EL[1] = P2[0] * matP[1]
      + P2[1] * matP[4]
      + matP[7];
    EL[2] = P2[0] * matP[2]
      + P2[1] * matP[5]
      + matP[8];    
  
    double dist = EL[0] * P1[0] + EL[1] * P1[1] + EL[2];
    // dist is now = x'Fx (zero for perfect correspondence)

    double dError = dist = dist * dist;
    dError /= (EL[0]*EL[0]+EL[1]*EL[1]);
    // dError containts now the perpendicular distance to epiline in image1

    // do the same for the inverse direction and add error value
    // EL = F * P1:
    EL[0] = matP[0]*P1[0]+ matP[1]*P1[1] + matP[2];
    EL[1] = matP[3]*P1[0]+ matP[4]*P1[1] + matP[5];
    // EL[2] is not needed

    return (dError + (dist / (EL[0]*EL[0]+EL[1]*EL[1])));
 
  }

  /// @todo optimze
  double EMatrixBase::GetCosAngleErrorHomogenized(const BIAS::HomgPoint2D& p1,
                                                  const BIAS::HomgPoint2D& p2)
  {
    BIASASSERT(_isDecomposed);
    double result;
    HomgLine2D l1, l2;
    HomgPoint2D rcp2;
    // l1 is the line between the epipole an p1
    l1.Set(_epipole, p1);
    l1.Homogenize();
    // now apply 'homography' to p2
    _R.Mult(p2, rcp2);
    // l2 is the line between p1 and R*p2
    l2.Set(p1, rcp2);
    l2.Homogenize();
    // now calculate the angle between these two lines
    result=l1[0]*l2[0]+l1[1]*l2[1];
    // crop inaccurate results
    if (result>=1.0) { result=1.0; }
    if (result<=-1.0) { result=-1.0; }
    BIASASSERT(result>=-1.0 && result <=1.0);
    return 1.0-result;
  }

} // namespace


#endif // __EMatrixBase_hh__
