/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __BresenhamCircle_hh__
#define __BresenhamCircle_hh__
#include "bias_config.h" 

namespace BIAS{
  /** @brief scans a circle using Bresenhams integer arithmetic algorithm
      - either Init or the according constructor have to be called befor 
      GetNext
      - points on circle are not given in continous manner, i.e. subsequent
      calls to GetNext() return (x,y), (y,x), (x,-y), (y,-x), 
      (-x,-y), (-y,-x), (-x,y), (-y,x) 
      - fast
      @author fw */
  class DLLEXPORTAPI BresenhamCircle{
  public:
    inline BresenhamCircle();
    
    inline BresenhamCircle(int center[2], int radius);
    
    inline ~BresenhamCircle();
    
    inline void Init(int center[2], int radius);

    inline bool GetNext(int next[2]);
  protected:
    int _iCenter[2]; // center coo
    int _iRadius; 
    int _iCurrent[2];
    int _iD;
    int _iDeltaE;
    int _iDeltaSE;
    int _iCount;
    
#ifdef BIAS_DEBUG
    bool _bInitialized;
#endif 
  };
  
  /** just like bresenham circle but only return 1/8th of the circle
      @author woelk */
  class DLLEXPORTAPI BresenhamCircleEighth{
  public:
    inline BresenhamCircleEighth();
    
    inline BresenhamCircleEighth(int center[2], int radius);
    
    inline ~BresenhamCircleEighth();
    
    inline void Init(int center[2], int radius);

    inline bool GetNext(int next[2]);
  protected:
    int _iCenter[2]; // center coo
    int _iRadius; 
    int _iCurrent[2];
    int _iD;
    int _iDeltaE;
    int _iDeltaSE;
    
#ifdef BIAS_DEBUG
    bool _bInitialized;
#endif 
  };

  /** makes 8 subsequent calls BresenhamCircleEighth
      slow does not work so far
      @author woelk */
  class DLLEXPORTAPI BresenhamCircleSlow{
  public:
    inline BresenhamCircleSlow();
    
    inline BresenhamCircleSlow(int center[2], int radius);
    
    inline ~BresenhamCircleSlow();
    
    inline void Init(int center[2], int radius);

    inline bool GetNext(int next[2]);
  protected:
    int _iZeroCenter[2];
    int _iCenter[2]; // center coo
    int _iRadius; 
    int _iCount;
    int _iNext[2];
    BresenhamCircleEighth _BresCirc;
#ifdef BIAS_DEBUG
    bool _bInitialized;
#endif 
  };

  BresenhamCircle::BresenhamCircle()
  {}

  BresenhamCircle::BresenhamCircle(int center[2], int radius)
  { Init(center, radius); }
    
  BresenhamCircle::~BresenhamCircle()
  {}

  void BresenhamCircle::Init(int center[2], int radius)
  {
    _iCenter[0]=center[0];
    _iCenter[1]=center[1];
    _iRadius=radius;
    _iCurrent[0]=0;
    _iCurrent[1]=_iRadius;
    _iD=1-_iRadius;
    _iDeltaE=3;
    _iDeltaSE=-2*_iRadius+5;
    // let GetNext return the starting point first 
    _iCurrent[0]--;
    if (_iD>=0) _iCurrent[1]++; 
    _iCount=0;
#ifdef BIAS_DEBUG
    _bInitialized=true;
#endif 
  }
  
  bool BresenhamCircle::GetNext(int next[2])
  {
#ifdef BIAS_DEBUG
    if (!_bInitialized)
      BIASERR("call Init befor GetNext");
#endif 
    bool result=true;
    if (_iCount==0){
      if (_iCurrent[1]>_iCurrent[0]){
        if (_iD<0){
          _iD+=_iDeltaE;
          _iDeltaE+=2;
          _iDeltaSE+=2;
        } else {
          _iD+=_iDeltaSE;
          _iDeltaE+=2;
          _iDeltaSE+=4;
          _iCurrent[1]--;
        }
        _iCurrent[0]++;
        next[0]=_iCurrent[0]+_iCenter[0];
        next[1]=_iCurrent[1]+_iCenter[1];
        _iCount=1;
      } 
    } else {
      next[0]=_iCenter[0];
      next[1]=_iCenter[1];
      switch (_iCount){
      case 1: 
        next[0]+=_iCurrent[1]; 
        next[1]+=_iCurrent[0]; break; 
      case 2: 
        next[0]+=_iCurrent[0]; 
        next[1]-=_iCurrent[1]; break; 
      case 3: 
        next[0]+=_iCurrent[1]; 
        next[1]-=_iCurrent[0]; break; 
      case 4: 
        next[0]-=_iCurrent[0]; 
        next[1]-=_iCurrent[1]; break; 
      case 5: 
        next[0]-=_iCurrent[1]; 
        next[1]-=_iCurrent[0]; break; 
      case 6: 
        next[0]-=_iCurrent[0]; 
        next[1]+=_iCurrent[1]; break; 
      case 7: 
        next[0]-=_iCurrent[1]; 
        next[1]+=_iCurrent[0]; break; 
      }
      _iCount++;
      if (_iCount==8) _iCount=0;
      if (_iCount==0 && _iCurrent[1]<=_iCurrent[0]){
#ifdef BIAS_DEBUG
        _bInitialized=false;
#endif 
        result = false;
      }
    }
    return result;
  }
  
  BresenhamCircleEighth::BresenhamCircleEighth()
  {}

  BresenhamCircleEighth::BresenhamCircleEighth(int center[2], int radius)
  { Init(center, radius); }
    
  BresenhamCircleEighth::~BresenhamCircleEighth()
  {}

  void BresenhamCircleEighth::Init(int center[2], int radius)
  {
    _iCenter[0]=center[0];
    _iCenter[1]=center[1];
    _iRadius=radius;
    _iCurrent[0]=0;
    _iCurrent[1]=_iRadius;
    _iD=1-_iRadius;
    _iDeltaE=3;
    _iDeltaSE=-2*_iRadius+5;
    // let GetNext return the starting point first 
    _iCurrent[0]--;
    if (_iD>=0) _iCurrent[1]++; 
#ifdef BIAS_DEBUG
    _bInitialized=true;
#endif 
  }
  
  bool BresenhamCircleEighth::GetNext(int next[2])
  {
#ifdef BIAS_DEBUG
    if (!_bInitialized)
      BIASERR("call Init befor GetNext");
#endif 
    bool result=true;
    if (_iCurrent[1]>_iCurrent[0]){
      if (_iD<0){
        _iD+=_iDeltaE;
        _iDeltaE+=2;
        _iDeltaSE+=2;
      } else {
        _iD+=_iDeltaSE;
        _iDeltaE+=2;
        _iDeltaSE+=4;
        _iCurrent[1]--;
      }
      _iCurrent[0]++;
      next[0]=_iCurrent[0]+_iCenter[0];
      next[1]=_iCurrent[1]+_iCenter[1];
    }  else {
#ifdef BIAS_DEBUG
      _bInitialized=false;
#endif 
      result = false;
    }
    return result;
  }

  BresenhamCircleSlow::BresenhamCircleSlow()
  {}

  BresenhamCircleSlow::BresenhamCircleSlow(int center[2], int radius)
  { Init(center, radius); }
    
  BresenhamCircleSlow::~BresenhamCircleSlow()
  {}

  void BresenhamCircleSlow::Init(int center[2], int radius)
  {
    _iZeroCenter[0]=_iZeroCenter[1]=0;
    _iCenter[0]=center[0];
    _iCenter[1]=center[1];
    _iRadius=radius;
    _iCount=0;
    _BresCirc.Init(_iZeroCenter, _iRadius);
#ifdef BIAS_DEBUG
    _bInitialized=true;
#endif 
 }

  bool BresenhamCircleSlow::GetNext(int next[2])
  {
#ifdef BIAS_DEBUG
    if (!_bInitialized)
      BIASERR("call Init befor GetNext");
#endif 
    bool res=true;
    switch (_iCount){
    case 0:
      if (_BresCirc.GetNext(_iNext)){
        next[0]=_iCenter[0]+_iNext[0];
        next[1]=_iCenter[1]+_iNext[1];
      } else {
        _BresCirc.Init(_iZeroCenter, _iRadius);
        _iCount++;
      }
      break;
    case 1:
      if (_BresCirc.GetNext(_iNext)){
        next[0]=_iCenter[0]+_iNext[1];
        next[1]=_iCenter[1]+_iNext[0];
      } else {
        _BresCirc.Init(_iZeroCenter, _iRadius);
        _iCount++;
      }
      break;
    case 2:
      if (_BresCirc.GetNext(_iNext)){
        next[0]=_iCenter[0]+_iNext[1];
        next[1]=_iCenter[1]-_iNext[0];
      } else {
        _BresCirc.Init(_iZeroCenter, _iRadius);
        _iCount++;
      }
      break;
    case 3:
      if (_BresCirc.GetNext(_iNext)){
        next[0]=_iCenter[0]+_iNext[0];
        next[1]=_iCenter[1]-_iNext[1];
      } else {
        _BresCirc.Init(_iZeroCenter, _iRadius);
        _iCount++;
      }
      break;
    case 4:
      if (_BresCirc.GetNext(_iNext)){
        next[0]=_iCenter[0]-_iNext[0];
        next[1]=_iCenter[1]-_iNext[1];
      } else {
        _BresCirc.Init(_iZeroCenter, _iRadius);
        _iCount++;
      }
      break;
    case 5:
      if (_BresCirc.GetNext(_iNext)){
        next[0]=_iCenter[0]-_iNext[1];
        next[1]=_iCenter[1]-_iNext[0];
      } else {
        _BresCirc.Init(_iZeroCenter, _iRadius);
        _iCount++;
      }
      break;
    case 6:
      if (_BresCirc.GetNext(_iNext)){
        next[0]=_iCenter[0]-_iNext[1];
        next[1]=_iCenter[1]+_iNext[0];
      } else {
        _BresCirc.Init(_iZeroCenter, _iRadius);
        _iCount++;
      }
      break;
    case 7:
      if (_BresCirc.GetNext(_iNext)){
        next[0]=_iCenter[0]-_iNext[0];
        next[1]=_iCenter[1]+_iNext[1];
      } else {
        _BresCirc.Init(_iZeroCenter, _iRadius);
        _iCount++;
      }
      break;
    case 8:
#ifdef BIAS_DEBUG
      _bInitialized=false;
#endif
      res=false;
      break;
    }

    return res;
  }
  

} // namespace

#endif // __BresenhamCircle_hh__
