/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __Image_hh__
#define __Image_hh__

#include "bias_config.h" 

//tell WIN32 to define math-constants 
//--- math.h is included by <complex>
#ifdef WIN32
#  define _USE_MATH_DEFINES
#endif //WIN32

#include <iostream>

#include <math.h>
#include <string>
#include <fstream>
#include "ImageBase.hh"
// for MaxSTValue and MinSTValue
#include <float.h>
#include <limits.h>

#ifdef _WIN32
#include "Base/Common/W32Compat.hh"
#pragma warning (disable: 4100)
#endif //_WIN32


#define MIP_DV_WIDTH 500
#define MIP_DV_HEIGHT 288

#define D_COPYIN        1<<1
#define D_CONVERT       1<<2 
#define D_CAST_CHECK    1<<3
#define D_MI_TOPLANAR   1<<4
#define D_MI_IOSTREAM   1<<5
#define D_MI_CONVGREY   1<<6
#define D_IMAGE_MINMAXCOO   1<<7

namespace BIAS {

  /** @class Image
      @ingroup g_image_base
      @brief The image template class for specific storage types.

      The real image data is stored as an array of the specific storage type.
      You can get the pointer by GetImageData() while  
      accessing the image data could also be more efficiently done 
      by GetImageDataArray()[y][x],
      which gives the first channel of pixel at position (x,y).
      The second and third channels for interleaved images are at 
      [y][channelcount*x+1] and [y][channelcount*x+2].<br>

      There is an additional MetaData array, which holds information like
      the Pmatrix, which are used by derived classes.<br>
     
      see also ExampleLoad.cpp , ExampleOperator.cpp, ExampleChessboard.cpp

      @internal The explicit instatiation of the template
      is handled in ImageInstantiation.cpp 
  **/
  template <class StorageType> 
  class Image : public ImageBase
  {
  public:    
    Image();
    virtual ~Image(){};

    /** constructor with size and channels
        allocates image data */
    Image(unsigned int width, unsigned int height, unsigned int channels=1,
          const bool interleaved=true);

    /** create image from imagebase. this is a dangerous operation and must 
        therefore not be used by implicit conversion. That means, you can
        manually convert ImageBase->Image, but you cannot pass an 
        Image<float> where Image<char> is expected. */
    explicit Image(const ImageBase& im);

    /** @brief calls Init from ImageBase  
        storageType is ignored, just dummy argument */
    void Init(unsigned int Width, unsigned int Height, 
              unsigned int channels=1, 
              enum EStorageType storageType= ST_unsignedchar,
              const bool interleaved=true);

    /** @brief reimplemented from ImageBase */
    void Release();

    /** @name Get Functions
        @{
    */
    /** overloaded GetImageData() from ImageBase
        @author Daniel Grest */
    inline StorageType* GetImageData() const
    {return (StorageType *)ImageData_;};
  
    /** overloaded GetImageData() from ImageBase
        @author Daniel Grest 
    */
    inline StorageType** GetImageDataArray() const
    {return (StorageType **)ImageDataArray_;};

    /** Returns value of pixel at specific position,
        using specific channel as offset
        @author Ingo Thomsen
        @date 03/27/2002
        @status tested (only with interleaved image) **/
    inline StorageType PixelValue(const unsigned int x,
                                  const unsigned int y,
                                  const unsigned short int channel=0)const;

    /** Returns value of pixel at specific position,
        using specific channel as offset.
	Faster than PixelValue because it has no error check,
	works only for interleaved images
        @author Johan Skoglund
        @date 06/10/2004
        @status untested **/
    inline StorageType PixelValueInterleaved(const int x,
					     const int y,
					     const int channel=0)const;

    /** Get the minimal pixel value
        if coo!=NULL the coo[0]=x of min and coo[1]=y of min
        @author Felix Woelk **/
    StorageType GetMinPixelValue(unsigned short int channel=0,
                                 unsigned int *coo=NULL) const;

    /** Get the maximal pixel value
        if coo!=NULL the coo[0]=x of max and coo[1]=y of max
        @author Felix Woelk **/
    StorageType GetMaxPixelValue(unsigned short int channel=0,
                                 unsigned int *coo=NULL) const;

    /** @brief returns the minimal and maximal pixel value in channel only
        Finds minimum and maximum pixel value in image in channel only, i.e.
        all other channels are ignored during the search. Returns the first 
        occurance of min/max pixelvalue, if they occure multiple times. 
        @param min The variable where to store the minimun value 
        @param min The variable where to store the maximum value 
        if mincoo or maxcoo !=NULL, also stores the location of
        min resp. max, works slow on  
        @tested woelk 10/2004 ( Examples/ExampleMinMax.cpp )
        @author Jan Frahm, woelk 01/2003, woelk 10/2004 **/
    void GetMinMaxPixelValue(StorageType& min, StorageType& max,
                             unsigned short int channel=0,
                             unsigned int *mincoo=NULL, 
                             unsigned int *coo=NULL) const;   

    /** @author woelk 08/2004 */
    void GetMeanPixelValue(StorageType mean[]);

    /** @brief Get both, minimal and maximal pixel value
        --actual implementation only for planar images--
        @param min The variable where to store the minimun value 
        @param min The variable where to store the maximum value 
        @param ignoreStart, ignoreEnd define teh range of values to be ignored
        @param channel to be searched fro multi-channel images
        @return true if a value within raneg was found, else false
        @author Jan Woetzel 01/2003 **/
    bool GetMinMaxPixelValueIgnoreRange(StorageType& min, StorageType& max,
                                        const StorageType ignoreStart=0,
                                        const StorageType ignoreEnd=0,
                                        const unsigned short int channel=0 ) 
      const;

    /** @brief Get the value of a specific channel offset of specific pixel 
        nearest to given double coordinates

        @author Ingo Thomsen
        @date 03/27/2002
        @status tested (only with interleaved grey images) **/
    inline StorageType& 
    GetPixelNearestNeighbor(const double x, const double y, 
                            const unsigned short int channel=0)const;

    /** @brief Get the value of a specific channel offset of specific pixel 
        nearest to given float coordinates, using the double version.
     
        @author Ingo Thomsen
        @date 03/28/2002
        @status tested (only with interleaved gray images) **/
    inline StorageType& 
    GetPixelNearestNeighbor(const float x, const float y, 
                            const unsigned short int channel=0) const;


    // @} 
    ////////end of Get Functions////////////////////////////////////////////
    
    /** @name Set Functions
        @{
    */
    /** 
        @brief sets all pixels to zero/value
        
        clear all pixels elements and set them to a 'default' value 
        JW 04/2003 
    */
    void Clear(const StorageType value=0) {
      StorageType tmp_v = value; 
      FillImageWithConstValue( tmp_v );
    };
    /// @brief fill grey images
    void FillImageWithConstValue(StorageType Value);
    /// @brief fill color images
    void FillImageWithConstValue(StorageType Value[]);
    
    /// @brief fills image with value depending on x coordinate
    void FillImageWithXValue();   

    /** set the value of a given pixel (x,y) in channel to value.
        @warning this function is slower than direct pointer acces if you 
        want to access ALL pixels of an image!        
        @author Jan Woetzel 08/2003 */
    inline void SetPixel(const StorageType & value, const unsigned int & x, 
                         const unsigned int & y, 
                         const unsigned short int channel=0);
    
    /** set the values of the first three channels of a given pixel (x,y) 
        in channel to value[0..2].
        @warning this function is slower than direct pointer acces if you 
        want to access ALL pixels of an image!
        @author Jan Woetzel 08/2003 */
    inline void SetPixel(const StorageType & value0,
			 const StorageType & value1,
			 const StorageType & value2,
			 const unsigned int & x, 
			 const unsigned int & y );
    
    // @}
    //////end of Set Functions /////////////////////////////////////////////

    /** @name Thresholding and Scaling
        @{
    */
   
    /** @brief scales and shifts image (all channels simultanously)

        does scale and shift every channel with same factor so that 
        (max of all channels) = Max and (min of all channels) = Min.
        First shifts, then scales.
        @author Felix Woelk **/
    int ScaleShift(double Scale, double Shift);

    /** @brief scales and shifts image so afterwards every pixel has a value 
        between Min and Max 

        scales and shifts every channel with same factor 
        @author Felix Woelk **/
    int ScaleShiftBetween(double Min, double Max); 
 

    /** @brief calculate the values needed by ScaleShiftBetween */
    int CalcScaleShift(double Min, double Max,
                       double& Scale, double &Shift) const;

    /// @brief sets alls pixels with values below Threshold to Value 
    int BelowThresholdToValue(StorageType Threshold, StorageType Value);

    /// @brief sets alls pixels with values above Threshold to Value 
    int AboveThresholdToValue(StorageType Threshold, StorageType Value);

    /** @brief sets all pixels >= Threshold to MaxPixelValue
        and all others to MinPixelValue 
        only for one channel / grey images
        @author woelk 01 2003 */
    void Binarize(StorageType Threshold);
   
    /** @brief Test for every pixel wether it is >= than the threshold. If so
        set to a MaxValue, otherwise to an MinValue.
       
        @author Ingo Thomsen
        @date 03/28/2002
        @status tested **/
    int BinaryThreshold(StorageType Threshold, 
                        StorageType MinValue, 
                        StorageType MaxValue);    
    //@}
    /////// end of thresholding //////////////////////////////////////////////
    /////////////////////////////////////////////////////////////////////////
    /** 
        @name Operators
        
        Operators -,+,*,/,-=,+=,*=,/= for scalar and image arguments <br>
        Operators do only work for argument of the same type <br> 
        Division Operators do zero checks (if BIAS_DEBUG is defined) 
        @{
    */
    /** copies ImageBase to an desired Image<>,
        be careful with this, because u can do 
        something like: Image<char> = Image<float>,
        what usualluy not makes sense          
        @author Friso Evers, Nov. 2002 */
    Image<StorageType>& operator=(const ImageBase& Source);


    /// Operator /= for scalar value
    Image<StorageType>& operator/= (const StorageType& argimage);


    /// Operator / for scalar value, returning a new image
    Image<StorageType> operator/  (const StorageType& argimage);


    /// Operator *= for scalar value  
    Image<StorageType>& operator*= (const StorageType& argimage);


    /// Operator * for scalar value, returning a new image
    Image<StorageType> operator*  (const StorageType& argimage);


    /// Operator += for scalar value  
    Image<StorageType>& operator+= (const StorageType& argimage);


    /// Operator + for scalar value, returning a new image
    Image<StorageType> operator+  (const StorageType& argimage);


    /// Operator -= for scalar value  
    Image<StorageType>& operator-= (const StorageType& argimage);


    /// Operator - for scalar value, returning a new image
    Image<StorageType> operator-  (const StorageType& argimage);


    /// Operator /= for another image as argument
    Image<StorageType>& operator/= (const Image<StorageType>& argimage);


    /// Operator / for another image as argument, returning a new image
    Image<StorageType> operator/  (const Image<StorageType>& argimage);


    /// Operator *= for another image as argument
    Image<StorageType>& operator*= (const Image<StorageType>& argimage);


    /// Operator * for another image as argument, returning a new image
    Image<StorageType> operator*  (const Image<StorageType>& argimage);


    /// Operator += for another image as argument
    Image<StorageType>& operator+= (const Image<StorageType>& argimage);


    /// Operator + for another image as argument, returning a new image
    Image<StorageType> operator+  (const Image<StorageType>& argimage);


    /// Operator -= for another image as argument
    Image<StorageType>& operator-= (const Image<StorageType>& argimage);


    /** subtracts every from every pixel (in ROI if defined)
        the pixel value from argimage 
        !!! negative results are clipped to zero !!!
        @author Ingo Thomsen
        @status tested (18/04/2002)
    */
    Image<StorageType> operator-  (const Image<StorageType>& argimage);
  

    Image<StorageType>& operator|  (const Image<StorageType>& argimage);
      
    //@}
    /////////// end of Operators ////////////////////////////////////////////

    /** @name Interpolating Functions
        @{
    */
    inline double LinearInterpolation(const unsigned int x, 
                                      const double y) const;
  
    inline double LinearInterpolation(const double x, 
                                      const unsigned int y) const;

    /** fast bilinear interpolation 
        specialisation for unsigned char exists, 20% faster than function
        below
        @author woelk 01 2003 */
    inline StorageType 
    FastBilinearInterpolationGrey(const double x, const double y) const;
	
    /** Returns the bilinear interpolated pixel value (using channel offset)
        value of pixel defined by double coordinates, ignoring an existingROI

        @author Ingo Thomsen
        @date 04.03.2002
        @status tested (only with gray images yet)      **/
    double BilinearInterpolation(const double x, const double y, 
                                 const unsigned short int channel=0) const;

   /** no boundary checking !
        ignores ROI
        @author Felix Woelk */
    inline double 
    BilinearInterpolationGrey(const double x, const double y) const;

    /** no boundary checking !
        ignores ROI
        @author Felix Woelk */
    inline double 
    BilinearInterpolationRGBPlanar(const double x, const double y, 
                                   unsigned int channel) const;

    /** no boundary checking !
        ignores ROI
        @author Felix Woelk*/
    inline double
    BilinearInterpolationRGBInterleaved(const double x, const double y, 
                                        unsigned int channel) const;

    /** interpolation with weights according to distances to x and y */
    inline double
    InterpolationGrey(const double x, const double y) const;

    /** @brief sample the continuous (bilinear interpolation of (*this)) image
        using a grid with one pixel spacing at subpixel 
        positions (x-hws, y-hws) , ... (x,y), ... (x-hws, y-hws)
        You get a window of size (2*hws_x+1)*(2*hws_y+1) which is exactly 
        centered at x,y (a subpixel position) of the original image
        
        @param Buffer must have size (2*hws_x+1)*(2*hws_y+1)
        @param hws_x half window size in x direction
        @param hws_y half window size in y direction
        @param x x-center of window
        @param y y-center of window
        @author koeser 01/2004 */
    inline void BilinearInterpolationShiftRegion(const double& x,
                                                 const double& y,
                                                 unsigned int hws_x, 
                                                 unsigned int hws_y,
                                                 unsigned char* Buffer);


    
    /** @brief bilinear interpolation pixel value with value check. 
        If at least one of the input values is <=minVal,
        then the unknown value minVal is returned, too.
        @param minVal all values <= minVal are invalid/unknown.
        @author Jan Woetzel 12/2004 */
    inline double
    BilinearInterpolationGreyMinVal(const double x, 
                                    const double y, 
                                    const StorageType & minVal) const;
    
    /** Returns the bicubic interpolated pixel value (using channel offset)
        value of pixel defined by double coordinates, ignoring an existing ROI
        Does not work for planar multi-channel images !

        The subpixel value is obtained by computing a weighted average of
        the surrounding 4x4 pixel block. The weights are computed by
        independent x and y-dir cubic polynomials which guarantee
        constant first order derivatives at the grid positions

        @attention 1.0<=x<Width-2 and 1.0<=y<Height-2 is required but not 
        checked
        @author koeser 02/2005
    **/
    double BicubicInterpolation(const double& x, const double& y, 
                                const unsigned short int channel=0) const;



    // @}
    
    /** @name Padding and Appending
        @{
    */
    /** enlarge the image by padding lines and colls with 0
        Only valid on interleaved images (now).
        @author Jan-Friso Evers-Senne*/  
    int ZeroPad(const unsigned int newwidth, const unsigned int newheight,
                unsigned char bgcolor= 0 );
      
    /** increase the size of this image to next power of two 
        (e.g. 384x100>512x128)
        the original image S is set to upper left rectangle
        @param padVal used by memset to pad the right and lower added pixels
        No scale, just padding!
        orig xx
        xxxxxxx
        @author Jan Woetzel 09/2003 */
    int PadToPowerOfTwo(BIAS::Image<StorageType> & dest, 
                        const int & padVal=0) const;
    int PadToPowerOfTwo(const int & padVal=0);
    
    /** first pad, then flip.
        useful for OpenGL texture image which must be pow2 size and 
        Y-flipped ics.
        example:
        0 1 2 
        3 4 5
        -->
        x x x x
        x x x x	
        3 4 5 x 
        0 1 2 x
        TODO: more efficient in place with memmove
        @author Jan Woetzel */
    int PadToPowerOfTwoAndFlip(const int & padVal=0);    
    
    /// Append an image with one channel of same StorageType
    /** Append means: RGB + A => RGBA */
    int AppendChannel(Image<StorageType> &img);               

    // @}

    /** @name Misc Functions
        @{
    */

    /// returns the maximal possible StorageType
    inline StorageType MaxSTValue();
      
    /// returns the minimal possible StorageType
    inline StorageType MinSTValue();
        
    /// writes pointer of IplImage_ to os (ascii)
    std::ostream& PrintPointer(std::ostream &os) const;

    /// writes data of IplImage_ to os (ascii)
    std::ostream& PrintData(std::ostream &os) const;

    /** @brief (*this) = | im1 - im2 |
        sets this as the absolute difference between two arg images
        @author Felix Woelk */
    void AbsDiff(const Image<StorageType>& im1, 
                 const Image<StorageType>& im2);

    /** @return the reciprocal for each pixel: x' = 1/x
        computes for exactly one channel (in case of multi channel images)
        returns float/double image
        @author Jan Woetzel 01/2003 */
    Image<float> Reciprocal(const float factor=1.0, 
                            const unsigned short int channel=0);
             
    // @}

  };

  // inline functions
#include "ImageInline.hh"

} // namespace BIAS

#endif // __Image_hh__
