/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <bias_config.h>

#define ImageIO_DEBUG_IO false

#include "ImageIO.hh"

// ImageMagick support: Jan Woetzel and Sun Min 10/2004
#ifdef BIAS_HAVE_IMAGEMAGICKLIB 
#  ifdef WIN32
#    pragma warning( disable : 4251)  // JW
#  endif
#  include <Magick++.h>
//#  include <magick/api.h>
#endif

#ifdef BIAS_HAVE_IMLIB
#  include <Imlib.h>
#endif

#include <Base/Debug/DebugSimple.hh>

#include <fstream>
#include <sstream>

#include "ImageConvert.hh"
#include "ViffHeader.hh"

// for transfrom:
#include <algorithm>
#include <cctype>

// for O_APPEND
#include <fcntl.h>

  // for perror
#include <stdio.h>

using namespace BIAS;
using namespace std;

char *ImageIO::MemAlignedBuffer_ = NULL;
unsigned long int ImageIO::MemAlignedBufferSize_ = 0;

int ImageIO::Load(const string& FileName, ImageBase &img)
{
  int result = 0;
  ifstream is(FileName.c_str());
  char identifier[IDENTIFIER_LENGTH];

  if (!is){
    result = -1;
  } else {
    is.read(identifier, IDENTIFIER_LENGTH);
    if (!is.good()){
      result = -2;
      is.close();
    } else {
      if (identifier[0] == 'M' && identifier[1] == 'I' && 
          identifier[2] == 'P'){ // mipimage
        is.close();
#ifdef BIAS_DEBUG
        if (ImageIO_DEBUG_IO) COUT("mip format detected\n");
#endif
        result = Read(FileName, img);
      } else { // foreign image format
        is.close();
#ifdef BIAS_DEBUG
        if (ImageIO_DEBUG_IO) COUT("image not in mip format\n");
#endif
        result = ImageIO::ImportImage(FileName, img);
      }
    }
  }
  return result;
}


#ifdef BIAS_HAVE_IMAGEMAGICKLIB
int ImageIO::ImportMagickPP(const std::string& FileName, ImageBase& result)
{
  try {
//#ifdef BIAS_DEBUG
//    if (ImageIO_DEBUG_IO) COUT("using ImageMagick to import image\n");
//#endif
#ifdef WIN32
    //initialize DLL loaded library (not required on Linux but once on Windows)
    Magick::InitializeMagick(NULL);
#endif
    Magick::Image image; 
    
    image.read( FileName );

    // now extract image info from Magick::Image
    unsigned width=image.image()->columns;
    unsigned height=image.image()->rows;
    unsigned bitdepth=image.image()->depth;
    unsigned depth=0;
    while (depth<bitdepth) depth+=8;
    depth=depth>>3; // divide by eight
//#ifdef BIAS_DEBUG
//    COUT("depth: "<<depth<<"\tbitdepth: "<<bitdepth<<endl);
//#endif

    unsigned cc=0;
    ImageBase::EColorModel cm=ImageBase::CM_invalid;
    ImageBase::EStorageType st=ImageBase::ST_invalid;
    Magick::StorageType imst=MagickLib::UndefinedPixel;

    /// LibMagick: imcs is RGB even for one channel images, like imlib does, too. (bug?)
    Magick::ColorspaceType imcs=image.image()->colorspace;
    std::string cm_map="";

    // chose appropriate storage type
    switch (depth){
    case sizeof(unsigned char):
      st=ImageBase::ST_unsignedchar;
      imst=MagickLib::CharPixel;
      break;
    case sizeof(unsigned short):
      st=ImageBase::ST_unsignedshortint;
      imst=MagickLib::ShortPixel;
      break;
    case sizeof(float): // could this also be int?
      st=ImageBase::ST_float;
      imst=MagickLib::FloatPixel;
      break;
    default:
      BIASERR("funny depth "<<depth<<", maybe just unfinished code");
      return -4;
      break;
    }

    // choose appropriate color model
    switch (imcs){
    case MagickLib::RGBColorspace:
      cm=ImageBase::CM_RGB;
      cm_map="RGB";
      cc=3;
      break;
    case MagickLib::GRAYColorspace:
      cm=ImageBase::CM_Grey;
      cm_map="K";
      cc=1;
      break;
    default:
      // look in ImageMagick.h for further ColorspaceType's
      BIASERR("color model not implemented ");
      return -3;
      break;
    }
//#ifdef BIAS_DEBUG
//    COUT("cm_map: "<<cm_map<<endl);
//#endif



//#ifdef BIAS_DEBUG
//    COUT("MagickLib::Image: Read "<<image.columns()<<"x"<<image.rows()<<" "
//         <<image.magick()<<" image "<<FileName<<" using ImageMagick++"<<endl);
//    COUT("MagickLib::Image: depth "<<image.depth()<<" bit."<<endl );
//    COUT("MagickLib::Image: comment "<<endl<<image.comment()<<endl );
//    COUT("MagickLib::Image: depth "<<depth<<endl);
//    COUT("MagickLib::Image: size "<<width<<"x"<<height<<endl);
//    COUT("MagickLib::Image: storage class "<<image.image()->storage_class
//         <<endl);
//    COUT("MagickLib::Image: color space "<<image.image()->colorspace<<endl);
//    COUT("MagickLib::Image: channel count "<<cc<<endl);
//#endif
    
    if (!result.IsEmpty()) result.Release();

    // do error checking if result has already a specific StorageType
    if (result.GetStorageType()!=ImageBase::ST_invalid){
      if (result.GetDepth()!=depth){
        BIASERR("error: you tried to load an "<<depth<<" byte "
                <<"image into an image of storage type "
                <<result.GetStorageType()<<" which has depth "
                <<result.GetDepth()<<" byte");
        result.PrintHeader(cout);
        assert(false);
        return -2;
      }
    } 
    
    if (!result.IsEmpty()) result.Release();
    
    result.Init(width, height, cc, st, true);
    //result.PrintHeader(cout);
    result.SetBitDepth(bitdepth);
   
    // now get the meta data from the comment
    stringstream strstr;
    //strstr << image.comment();
    /// \\todo crashes JW
    //strstr >> *result.GetMetaData();

    //result.GetMetaData()->Dump();

    image.write(0, 0, width, height, cm_map, imst, result.GetImageData());

  } catch( Magick::Exception &error_ ) {
    BIASERR("Reading "<<FileName<<" with Magick++ failed."<<endl
            <<"  Caught exception: "<<endl
            <<"  "<<error_.what()<<endl );
    //assert(false);
    return -1; // error
  }
  
  return BIAS_IO_SUCCESS;
}
#endif


int ImageIO::ImportImage(const string& filename, ImageBase &result, 
                         bool assumeGrey) 
{
#ifdef BIAS_DEBUG
    if (ImageIO_DEBUG_IO) COUT("ImportImage\n");
#endif
  
  int res=-1; // init as file not read
  ifstream ifs;
  unsigned char tmpline[512];
  
#ifdef   WIN32
  ifs.open(filename.c_str(),ios::binary);
#else  //not WIN32
  ifs.open(filename.c_str());
#endif //WIN32
  
  if (!ifs) {
    BIASERR("ImportImage(): open of file \""<<filename<<"\" failed");
    perror(filename.c_str());
    ifs.close();
    return -1;
  }

  // invalidate old uid, since it may refer to a different image
  result.InvalidateUID();

  // read first two chars to get a hint what this file is
  ifs >>tmpline[0]>>tmpline[1];
  if (!ifs){
    res=-1;
    BIASERR("unable to read from "<<filename);
    ifs.close();
    return res;
  }
  
  // Check for famous P5/P6 in PNM, PGM, PBM to determine the fileformat
  // and call our own reader if appropriate.
  if (tmpline[0] =='P' && tmpline[1] == '5') {
#ifdef BIAS_DEBUG
    if (ImageIO_DEBUG_IO) COUT("reading single channel pnm/pgm/pbm image using ImportImagePnm_\n");
#endif
    res = ImportImagePnm_(ifs, result, 1); 
    if (res!=BIAS_IO_SUCCESS) 
      BIASERR("error reading "<<filename<<" although P5 format detected.");
    result.SetColorModel(ImageBase::CM_Grey);
  } else if (tmpline[0] =='P' && tmpline[1] == '6') {
#ifdef BIAS_DEBUG
    if (ImageIO_DEBUG_IO) COUT("reading 3 channel pnm/pgm/pbm image using ImportImagePnm_\n");
#endif
    res = ImportImagePnm_(ifs, result, 3); 
    if (res!= BIAS_IO_SUCCESS) 
      BIASERR("error reading "<<filename<<" although P6 format detected.");
    result.SetColorModel(ImageBase::CM_RGB);
  } 
  
  // other loaders do not use file stream but operate on filename directly 
  // and open theirselves.
  ifs.close();
  
  //
  // Try to read using external image loaders.
  //
  
#ifdef BIAS_HAVE_IMAGEMAGICKLIB
  if (res!=BIAS_IO_SUCCESS) res=ImportMagickPP(filename, result); 
#endif

#ifdef BIAS_HAVE_IMLIB
  if (res!=BIAS_IO_SUCCESS) res=ImportImlib(filename, result, assumeGrey);
#endif

  // JW: try to read VIFF if anything else failed
  if (res!=BIAS_IO_SUCCESS) {
      // open fstream 
      std::ifstream ifs(filename.c_str(), std::ios::binary );
      if (!ifs) {
          BIASERR("could not open "<<filename);
      } else {
          // OK
        res=ImportImageViff_(ifs, result/* assumeGrey*/ );
      }
  };
  
  if (res==0){
    // --- handle meta data ----------------
    AppData ad;
    if (result.GetMetaData()->Find(AppData::MD_UUID,"#[UUID]", ad)>=0){
      if (ad.tag==AppData::MD_USE_ASCII){
        //cerr << "before |"<< ad.sdata <<"|"<< endl ;
        while (ad.sdata.length()>0 && (ad.sdata[0]=='#' || ad.sdata[0]==' ' ||
                                       ad.sdata[0]=='\n'))
          ad.sdata.erase(0, 1);
        if (ad.sdata.length()>36)
          ad.sdata.erase(36, ad.sdata.length()-36);
        //cerr<<"after |"<<ad.sdata<<"|"<<" length: "<<ad.sdata.length()<<endl;
        UUID TempID(ad.sdata);
        result.SetUID(TempID);
      } else {
        BIASERR("UUID is no meta datum in binary images");
        BIASERR("possibly two concurrent UUIDs in one image");
        //result.SetUID(*((UUID*)ad.data));
        BIASASSERT(false);
      }
    } else {
#ifdef BIAS_DEBUG
#  ifdef BIAS_EXTRA_WARN
      BIASERR("warning, no valid UID read, UUID must be set manually");
#  endif
#endif
      result.InvalidateUID();
    }
    //cerr <<"header in ImportImage"<<endl;
    //result.PrintHeader(cout);
  }
  return res;
} // ImportImage





#ifdef BIAS_HAVE_IMLIB
///JW extracted from ImportImage 10/2004 
int ImageIO::ImportImlib(const string& filename, ImageBase &result, 
                         bool assumeGrey){
#ifdef BIAS_DEBUG
    if (ImageIO_DEBUG_IO) COUT("using Imlib to import image\n");
#endif
  // Data structures
  ImlibImage *img;
  //Static since I have not been able to find a function that frees Imlibdata
  //If we read one Image then it is reasonable to think that we will call 
  //this function again
  static ImlibData  *iData=NULL;
  static Display    *disp=NULL;
  
  if(disp==NULL)
    // Imlib really needs a pointer to the current display (on Unix)
    disp=XOpenDisplay(NULL);
  if(iData==NULL)
    iData=Imlib_init(disp);
  img = Imlib_load_image(iData, (char*) filename.c_str());
  
  if (img == NULL) {
    BIASERR("Could not read or decode image file "<<filename
            <<" with Imlib. Nothing read!");
    return -1;
  }
  
  bool isGrey = true;
  // Do I have to test for grey at all?
  if (assumeGrey) {
    isGrey = true;
  } else {
    
    // Go through the image and test for equal R,G,B values for each pixel
    register unsigned char* start = img->rgb_data;
    register unsigned char* stop  = start + 
      ( img->rgb_width * img->rgb_height * 3);
    
    while (start < stop) {    
      if ( (*(start) == *(start+1)) && *(start)  == *(start+2)) {
	start+=3;
      } else {
	isGrey = false;
	break;
      }
    }
  }
  
  // must be empty before new image information can be inserted
  if (!result.IsEmpty()){
    result.Release();
  }
  
  // copy the raw data in
  register unsigned char* sourceImageData = img->rgb_data;
  register unsigned char* imageData;
  register unsigned char* imageDataStop;
  
  /// Friso said: imlib seems to read three channel images even for 
  // single channel grey images (JW 10/2004)
  if (isGrey) {
    // Grey image
    result.Init(img->rgb_width, img->rgb_height, 1);
    result.SetColorModel(ImageBase::CM_Grey);
    imageData     = (unsigned char*)result.GetImageData();
    imageDataStop = imageData + result.GetPixelCount();
    
    // Take every third value (r=g=b) to create an image with grey values
    for (; imageData < imageDataStop; imageData++, sourceImageData += 3) {
      *imageData = *sourceImageData;
    }
    
  } else {
    // Color image
    result.Init(img->rgb_width, img->rgb_height, 3);
    result.SetColorModel(ImageBase::CM_RGB);
    imageData     =  (unsigned char*)result.GetImageData();
    imageDataStop = imageData + 3 *  result.GetPixelCount();
    for (; imageData < imageDataStop; imageData++, sourceImageData++) {
      *imageData = *sourceImageData;
    }
  }

  Imlib_kill_image(iData,img); 
  //XCloseDisplay(disp);
  return BIAS_IO_SUCCESS;
}
#endif  // BIAS_HAVE_IMLIB






int ImageIO::ImportImagePnm_(ifstream& ifs, ImageBase &result, int depth)
{
#ifdef BIAS_DEBUG
    if (ImageIO_DEBUG_IO) COUT("ImportImagePnm_\n");
#endif
  int xres=-1, yres=-1, maxval=-1, read=0;
  int res=0;
  char c;

  if (!result.IsEmpty())
    result.Release();
  // start from the beginning and read line with P5 or P6
  ifs.seekg(0,ios::beg);

  ifs >> c; 
  if (c!='P'){
    BIASERR("invalid magic cookie");
    return -1;
  }
  ifs >> c; 
  if (c!='5' && c!='6'){
    BIASERR("invalid magic cookie");
    return -1;
  }

  // this allows an arbitary amount of whitespace in between the header
  while (read<3){
    ifs >> c; 
    ifs.putback(c);
    if (c=='#'){
      ifs >> *result.GetMetaData();
      if (!ifs){
        BIASERR("error reading meta data");
        return -1;
      }
    } else if (c>29 || c<40){ // number 0-9
      switch (read){
      case 0:
        ifs >> xres;
        //cerr << "xres: "<<xres<<endl;
        read++;
        break;
      case 1:
        ifs >> yres;
        //cerr << "yres: "<<yres<<endl;
        read++;
        break;
      case 2:
        ifs >> maxval;
        //cerr << "maxval: "<<maxval<<endl;
        read++;
        break;
      default:
        BIASERR("funny error: "
                <<"there should be only 3 numerical values in pgm header");
        return -1;
        break;
      }; // switch
    } else { // if (c>29 || c<40){
      BIASERR("corrupted image header "<<c);
    }
  }
  ifs.read(&c, 1*sizeof(char)); // last whitespace
 
  //cerr <<"image has size "<<xres<<"x"<<yres<<"  with maxval "<<maxval<<endl;
  if (maxval==255){
#ifdef BIAS_DEBUG
    if (result.GetStorageType()!=ImageBase::ST_unsignedchar &&
        result.GetStorageType()!=ImageBase::ST_invalid){
      BIASERR("warning trying to load an unsigned char image into an image of"
              <<" storage type "<<result.GetStorageType());
      BIASASSERT(false);
    }
#endif
    result.Init(xres, yres, depth, ImageBase::ST_unsignedchar);
    ifs.read(reinterpret_cast<char *>(result.GetImageData()), xres * yres * depth);
  } else if (maxval==4095){ // 12 bit images
#ifdef BIAS_DEBUG
    if (result.GetStorageType()!=ImageBase::ST_unsignedshortint &&
        result.GetStorageType()!=ImageBase::ST_invalid){
      BIASERR("warning trying to load an unsigned short int image into an "
              <<"image of storage type "<<result.GetStorageType());
    }
#endif
    result.Init(xres, yres, depth, ImageBase::ST_unsignedshortint);  
    result.SetBitDepth(12);
    ifs.read(reinterpret_cast<char *>(result.GetImageData()), 
             sizeof(unsigned short int) * xres * yres * depth);
    // check if bytes need swapping
    unsigned short int *p=(unsigned short int *)result.GetImageData(), *pe;
    pe=p+result.GetPixelCount()*result.GetChannelCount();
    // check if pixel > maxval exist
    bool swap=false;
    while (p<pe) { swap = (swap || (*p > maxval)); p++; }
    // swap bytes if pixel > maxval existed
    if (swap){
      BIASERR("maxval > "<<maxval<<" : swapping LSB and MSB");
      char *cp=(char *)result.GetImageData(), *cpe;
      cpe=(char *)((unsigned short int *)cp + result.GetPixelCount()*
        result.GetChannelCount());
      char c;
      while (cp<cpe){
        c=cp[0]; cp[0]=cp[1]; cp[1]=c;
        cp+=2;
      }
    }
#ifdef BIAS_DAIMLERCHRYSLER
    // check image, since bosch fills last line with values above 4095
    p=((unsigned short int *)result.GetImageData());
    pe=p+result.GetPixelCount()*result.GetChannelCount();
    while (p<pe) {
      if (*p>maxval) *p=maxval; 
      p++; 
    }
#endif
  }
  
  if (!ifs.good()){
    BIASERR("error reading from ifs");
    res=-1;
  }
  //ExportImage("test", result, MIP_jpg);
  return res;
}


int ImageIO::ExportImage( const string& filename, ImageBase &img,
                          enum TFileFormat FileFormat, bool sync,
                          int jpeg_quality, bool forceNewID) 
{ 
  // do NOT scale or shift in this routine. 
  // Do it by hand if you need it. (JW 04/2003)
  int result = -1;
  
  string s(filename);
#ifdef BIAS_HAVE_IMLIB
  // special saving information, only needed for jpeg images
  struct _ImlibSaveInfo saveinfo;
#endif
  
  // first get copy of image in outpic, which is unsigned char
  // because nearly every export format is based on unsigned char
  Image<unsigned char> outpic;
  
  switch (img.GetStorageType()){
  case ImageBase::ST_unsignedchar:
    outpic = img;
    break;
    
  case ImageBase::ST_float: 
    // do NOT scale or shift in this routine. 
    // Do it by hand if you need it. (JW 04/2003) 
    // Image<float> tmp;
    // tmp = img;
    // tmp.ScaleShiftBetween(0,255);
    // outpic = tmp;
    outpic = img;
    break;
    
#ifdef BUILD_IMAGE_CHAR
  case ImageBase::ST_char: 
    outpic = img;
    break;
#endif
#ifdef BUILD_IMAGE_USHORT
  case ImageBase::ST_unsignedshortint : 
    outpic = img;
    break;
#endif
#ifdef BUILD_IMAGE_SHORT
  case ImageBase::ST_shortint : 
    outpic = img;
    break;
#endif
#ifdef BUILD_IMAGE_UINT 
case ImageBase::ST_unsignedint : 
    outpic = img;
    break;
#endif
#ifdef BUILD_IMAGE_INT  
  case ImageBase::ST_int : 
    outpic = img;
    break;
#endif  

  default:
    BIASERR("Unknow StorageType: "<<img.GetStorageType());
    break;
  };

  // asign new id if forced or there is none yet
  if ((forceNewID) || (!outpic.GetUID().IsValid())) {
    outpic.SetUID(UUID::GenerateUUID());
  }
  // write uuid to meta data if valid (only temporary to utilize metadata i/o)
  if (outpic.GetUID().IsValid()) {
    string suuid;
    outpic.GetUID().GetString(suuid);
    outpic.GetMetaData()->Add("#[UUID]", suuid);
  } else BIASERR("id of outpic is invalid");


  switch (FileFormat) {
  case FF_ppm:
    // s += ".ppm";
    s = ImageIO::ExtensionName( s, "ppm");
    result = ExportImagePnm_( s, outpic, sync); 
    break;
    
  case FF_pgm:
    // s += ".pgm";
    s = ImageIO::ExtensionName( s, "pgm");
    if (img.GetChannelCount() != 1)
      ImageConvert::IP_ToGrey(outpic);
    result = ExportImagePnm_( s, outpic, sync);
    break;
    
  case FF_jpg:
    if (jpeg_quality < 0 || jpeg_quality > 100) {
      BIASERR("The quality of a jpeg image to be saved must be between "<<
              "0 and 100 %. Nothing saved.");
      return -1;
    }
#ifdef BIAS_HAVE_IMLIB
    saveinfo.quality = (int) (2.56 * (double) jpeg_quality);
#endif
    // s += ".jpg";
    s = ImageIO::ExtensionName( s, "jpg");
    break;
  case FM_tif:
    // s += ".tif";
    s = ImageIO::ExtensionName( s, "tif");
    break;
  case FF_png:
    // s += ".png";
    s = ImageIO::ExtensionName( s, "png");
    break;
  
    // JW 01/2003
  case FF_viff:
    // s += ".viff";
    s = ImageIO::ExtensionName( s, "viff");
    break;
    
  default:
    BIASERR("Unknown image file format for saving! Nothing saved!");
    return -1;
  } // switch

#ifdef BIAS_HAVE_IMLIB
  if (result==-1) {
    // Imlib needs an RGB image, interleaved with unsigned char as data type
    Image<unsigned char> RGB_version;
    ImageConvert::ToRGB(outpic,RGB_version);
  
    /// \todo create separate ExportImlib method like for import, see comment there (jw)
    // create Imlib data structures
    ImlibImage imgImLib;

    /// static because we didn't find free for xopendisplay... (jw)
    static ImlibData  *idata=NULL;
    static Display    *imgDisp=NULL;

    /// ? kommen hier die DISPLAY fehler her?
      if (imgDisp==NULL)
	imgDisp = XOpenDisplay(NULL);
      
      if (idata==NULL)
	idata = Imlib_init(imgDisp);
  
    // create an Imlib image and take the Image image data for this purpose
    imgImLib.rgb_width  = RGB_version.GetWidth();
    imgImLib.rgb_height = RGB_version.GetHeight();
    imgImLib.rgb_data   = RGB_version.GetImageData();
    imgImLib.filename   = "dummy_file_name";
    // call the Imlib saving routine
    if (Imlib_save_image(idata, &imgImLib, (char*)s.c_str(), &saveinfo) == 0) {
      BIASERR("Error saving " << s.c_str() << " occured. Nothing saved!");
      return -1;
    }
    result = BIAS_IO_SUCCESS;
  }
#endif

  // if we assigned a new id to outpic, update img with this id !
  img.SetUID(outpic.GetUID());
  return result;
}





int ImageIO::ExportImagePnm_(const string &FileName, ImageBase &img, bool sync)
{
  ofstream ofs;
  // bool sync=false;

#ifdef   WIN32
    ofs.open(FileName.c_str(),ios::binary);
#else  //not WIN32
    ofs.open(FileName.c_str());
#endif //WIN32

  if (!ofs) {
    BIASERR("ExportImagePnm_() : open of file "<<FileName<<" failed");
    perror(FileName.c_str());
    return -1;
  }
  
  switch (img.GetChannelCount()) {
  case 3:
    ofs <<"P6"<<endl;
    break;
    
  case 1:
    ofs << "P5"<<endl;
    break;
  default:
      BIASERR("ExportImagePnm_() : unsupported number of channels: "<<img.GetChannelCount());
    return -1;
  }

  img.GetMetaData()->WriteAscii(ofs);
  if (!ofs){
    BIASERR("writing MetaData failed");
    return -2;
  } 

  ofs << img.GetWidth()<<" "<<img.GetHeight()<<endl;
  ofs << "255"<<endl;
  
  ofs.write(reinterpret_cast<char *>(img.GetImageData()), img.GetSize());
  if(!ofs) {
    BIASERR("ExportImagePnm_() : write failed");
    return -1;
  }

  if (sync) {
    // flush streambuf from c++
    ofs.flush();
    // flush system buffers
#ifndef _WIN32
#if __GNUC__ < 3
     fsync(ofs.rdbuf()->fd());
#else
	 ofs.rdbuf()->pubsync();
#endif
#else
     ofs.rdbuf()->pubsync();
#endif   
  }
  
  ofs.close();
  return 0;
}






/// helper function for ExtensionName:
char myTolower(char c) {
  return char(tolower(c));
}


std::string ImageIO::ExtensionName(const std::string & oldName, 
                                   const std::string & newExtension) {
  string name = oldName; // full name
  string ext; // file extension after dot '.' converted to lower case
  string newLCExtension; // lower case
  
  // find dot from right
  string::size_type pos = name.rfind('.'); 

  if (pos!=std::string::npos) { 
    // found dot and extension, copy from char after dot to end
    ext = name.substr(pos+1, name.size()-1 );
    // lower case extension
    transform(ext.begin(), ext.end(), ext.begin(), myTolower); 
    
    newLCExtension = newExtension;
    transform( newLCExtension.begin(), newLCExtension.end(), 
               newLCExtension.begin(), myTolower ); // lower case extension
    
    // append extension if it is not correct case insensitive extension:
    if (ext!=newLCExtension) name += "." + newExtension;
    
  } else {
    // no pos=no extension, add one:
    name += '.' + newExtension;
  };
  return name;
}


int ImageIO::Write(const string& FileName, ImageBase &img, 
                   bool sync, bool forceNewID) {
  int res = 0;
  ofstream os;

  // assign new id if forced or there is none yet
  if ((forceNewID) || (!img.GetUID().IsValid())) {
    img.SetUID(UUID::GenerateUUID());
  }
  
  // avoid ".mip.mip" (JW) 
  string file = ImageIO::ExtensionName( FileName, "mip");
  
#ifdef   WIN32
  os.open(file.c_str(),ios::binary);
#else  //not WIN32
  os.open(file.c_str());
#endif //WIN32

  if (!os){
    string Errorstring="Error opening "+file+" writing";
    perror(Errorstring.c_str());
    return -1;
  }

  os << img;
  res = (os.good())?(0):(-5);

  if (res==0){
    //cerr <<"Writing MetaData\n";
    //img.GetMetaData()->Dump();
    img.GetMetaData()->WriteBinary(os);
    if (!os){
      BIASERR("writing MetaData failed");
      res=-6;
    }
  }

  if (sync) {
    // flush streambuf from c++
    os.flush();
    // flush system buffers
#ifndef _WIN32
#  if __GNUC__ < 3 // in gcc 2.9x we can get a POSIX file descriptor 
    fsync(os.rdbuf()->fd());
#  else   // in gcc 3.x this is prohibited :-(
    os.rdbuf()->pubsync(); 
//     ::sync(); // this is systemcall sync() to sync all buffers    
#  endif
#else
    /** @todo Sync written files and buffers to disk under WIN32 
	(evers 2004-10-14) */
// 	os.rdbuf()->pubsync();
//     ::sync();
#endif
  }
  os.close();

#ifndef _WIN32 // alternative way to sync, reopen file to get a handle
  int fd = open(file.c_str(),O_APPEND);
  fsync(fd); // and now fsync()
  close(fd);
#endif 
  return res;
}


int ImageIO::Read(const string& FileName, ImageBase &img)
{
  int res=0;
  ifstream is;

  #ifdef   WIN32
    is.open(FileName.c_str(),ios::binary);
  #else  //not WIN32
    is.open(FileName.c_str());
  #endif //WIN32

  if (!is){
    string Errorstring="Error opening "+FileName+" for reading";
    perror(Errorstring.c_str());
    return -1;
  }
  is >> img;
  
  res = (is.good())?(0):(-5);

  if (res==0){
    //cerr << "ImageIO::Read() : "<<is.tellg()<<endl;
    is >> *img.GetMetaData();
    if (!is && !is.eof()){
      BIASERR("error reading MetaData");
      res = -6;
    }
  }

  //img.GetMetaData()->Dump();

  is.close();

  
  return res;
}



 int ImageIO::ImportImageViff_(std::ifstream& ifs, ImageBase &result)
{  

  struct xvimage viff;
  
  if (!ifs) return -1;
  
  if (!result.IsEmpty())  result.Release();
  
  ifs.seekg(0, ios::beg);  
  
  // read header struct (jw)
  ifs.read((char*)&viff,sizeof(viff));
  
  cout <<"read viff size "<<viff.row_size<<" x "<<viff.col_size<<endl;
  
  //   if (xres %4 >0) xresaligned = xres - xres%4;
  //   else xresaligned = xres;

  //result.Init(xresaligned, yres, depth, ImageBase::ST_unsignedchar);
  result.Init(viff.row_size, viff.col_size, 
      1, ImageBase::ST_float); // 1 is the number of channels
  //  cerr << "initialized with " << xres << " " << yres 
  // << " unsigned char" << endl;

  // read data area (jw)
  ifs.read(reinterpret_cast<char*>(result.GetImageData()), result.GetSize());
  result.SetColorModel(ImageBase::CM_Grey);

  return 0; 
}


#ifdef BIAS_HAVE_IMAGEMAGICKLIB
/** @brief: This routine is working inefficiently, 
    because it doesn't know what to do about conversion if read and requested 
    types don't match (JW 10/2004)
    However it works and assumes RGB in case of doubt.
    \todo rewrite by making more use of type Image<..> instead of ImageBase
    because lossless conversions and warnings for unwanted lossy conversions
    are easier, then (JW).
    @author Jan Woetzel 10/2004 */
int ImageIO::ImportMagickPPAutoconvert(const std::string& FileName,
                                       ImageBase& result,
                                       const bool &assumeGrey,
                                       const bool &readMetaData )
{
  try {
#ifdef WIN32
    // initialize DLL loaded library (not required on Linux but once on Windows)
    Magick::InitializeMagick(NULL);
#endif
    Magick::Image image; // temporary converter for lib usage
    
    // Read a file into image object using Magick++ library function:
    image.read( FileName );
    
    COUT("DBG Read "<<image.columns()<<"x"<<image.rows()<<" "<<image.magick()<<" image "<<FileName<<" using ImageMagick++"<<endl);
    COUT("DBG   depth: "<<image.depth()<<" bit."<<endl );
    COUT("DBG   comment: "<<endl<<image.comment()<<endl );
    
    // determine if import image format/channels and bit depth match:
    Magick::StorageType outputST=Magick::CharPixel; // output storage type

    /** keep the output storagetype
	\todo TODO HACK add dynamic cast test for templated Image<> to convert loaded float imaeg automatically to uc etc.. (JW)
	\todo add more knowledge about conversion and the option of errors if requested and read type don't match (JW).    
	check if at least StorageType and bit depth match:
	
    */
    //
    // first switch: ouput storagetype 
    //
    switch (result.GetStorageType()) {
    case BIAS::ImageBase::ST_invalid:
      BIASERR("invalid Storagetype ST_invalid FileName="<<FileName<<endl);
      break;
      
    case BIAS::ImageBase::ST_unsignedchar:
    case BIAS::ImageBase::ST_char:
      // 8 bit
      if (image.depth() != 8*sizeof(char)){
	BIASERR("StorageType is "<<8*sizeof(char)<<" bit but data has "<<image.depth()<<" bit depth. FileName="<<FileName<<endl);
	/** \todo handle conversion in extra routine 
	    with extra warnings in case of scaling or lossy conversion.
	    -Add a 'normalized scale' per conversion for each Storagetype.
	    -Add a flag to return errors instead of run-time warning. (discussion 26.Oct.2004) (JW)  */
	return -1; // HACK
      } else {
	outputST=Magick::CharPixel;
	COUT("is an "<<8*sizeof(char)<<" bit (u)char image\n");
      };
      break;
      
    case BIAS::ImageBase::ST_unsignedshortint:
    case BIAS::ImageBase::ST_shortint:
      // 16 bit
      if (image.depth()!=8*sizeof(short)){
	BIASERR("StorageType is "<<8*sizeof(short)<<" bit but data has "<<image.depth()<<" bit depth. FileName="<<FileName<<endl);
	return -2; // HACK 
      } else {
	outputST=Magick::ShortPixel;
	COUT("is a "<<8*sizeof(short)<<" bit (u)short image\n");
      };
      break;
      
    case BIAS::ImageBase::ST_unsignedint:
    case BIAS::ImageBase::ST_int:
      // 32 bit int
      if (image.depth()==8*sizeof(int)){
	BIASERR("StorageType is "<<8*sizeof(int)<<" bit but data has "<<image.depth()<<" bit depth. FileName="<<FileName<<endl);
	return -3; // HACK 
      } else {
	outputST=Magick::IntegerPixel;
	COUT("is a "<<8*sizeof(int)<<" bit (u)int image\n");
      };
      break;

      /**  \todo float and double pixel are 0..1 normlaized range in ImageMagick++. Min, max meta data required (JW) */
    case BIAS::ImageBase::ST_float:
      // 32 bit float
      if (image.depth()!=8*sizeof(float)) {
	BIASERR("StorageType is "<<8*sizeof(float)<<" bit but data has "<<image.depth()<<" bit depth. FileName="<<FileName<<endl);
	return -4; // HACK
      } else {
	outputST=Magick::FloatPixel;
	COUT("is a "<<8*sizeof(float)<<" bit float image\n");
	BIASERR("warning: You are using ImageMagick to read a float image. Be aware that ImageMagick float values are normalized to 0..1 !!!"<<endl);
      };
      break;
      
    case BIAS::ImageBase::ST_double:
      // 64 bit double 
      if (image.depth()!=8*sizeof(double)){
	BIASERR("StorageType is "<<8*sizeof(double)<<" bit but data has "<<image.depth()<<" bit depth. FileName="<<FileName<<endl);
	return -4; // HACK 
      } else {
	outputST=Magick::DoublePixel;
	COUT("is a "<<8*sizeof(double)<<" bit float image\n");
	BIASERR("warning: You are using ImageMagick to read a double image. Be aware that ImageMagick float values are normalized to 0..1 !!!"<<endl);
      };
      break;
      
    default:
      BIASERR("StorageType "<<result.GetStorageType()<<" not supported. FileName="<<FileName<<endl);
      return -9; // error
    };
    
    
    //
    // second switch for channels: number and color model
    //
    string magickMap="RGB"; // default
    unsigned int nChannels=3;
    //determine import image format/channels:
    switch (result.GetColorModel()) {
    case BIAS::ImageBase::CM_invalid:
      COUT("DBG CM_invalid. Setting default map RGB\n");
      nChannels=3;
      magickMap="RGB";
      break;
      
    case BIAS::ImageBase::CM_Grey:
      COUT("DBG CM_Grey\n");
      magickMap="K"; // blacK=grey
      nChannels=1;
      break;
    case BIAS::ImageBase::CM_GreyA:
      COUT("DBG CM_GreyA\n");
      magickMap="KA";
      nChannels=1;
      break;
    case BIAS::ImageBase::CM_RGB:
      COUT("DBG CM_RGB\n");
      magickMap="RGB";
      nChannels=3;
      break;
    case BIAS::ImageBase::CM_RGBA:
      COUT("DBG CM_RGBA\n");
      magickMap="RGBA";
      nChannels=4;
      break;
    case BIAS::ImageBase::CM_BGR:
      COUT("DBG CM_BGR\n");
      magickMap="BGR";
      nChannels=3;
      break;
      
    default:
      BIASERR("unknown result Color Model "<<result.GetColorModel()<<" reading "<<FileName<<endl);
    };
    
    // init/allocate result color image data: 
    if (!result.IsEmpty()) result.Release();
    result.Init(image.columns(), image.rows(), nChannels);
    result.SetColorModel(ImageBase::CM_RGB);
    
    
    // get the pointer to our result BIAS image data: 
    unsigned char* p_imageData     =  (unsigned char*)result.GetImageData();
    BIASASSERT( p_imageData != NULL);
    // write data array from read in ImageMagick++ image to our own ImageData Array.
    // details (JW): http://studio.imagemagick.org/Magick++/Image.html#write 
    image.write(0,0, // upper left corner of region to extract 
		result.GetWidth(), result.GetHeight(), // region size to extract
		"RGB", // map=output format in our pixel data, e.g RGBA
		outputST, 
		p_imageData // pointer to our data array of type
		);
    if (readMetaData){
      BIASERR("reading of MetaData is not implemented for IMageMagick++ but could be done using BIAS::ImageAttributes class (JW 10/2004)\n");
    }
    
  } catch( Magick::Exception &error_ ) {
    BIASERR("Reading "<<FileName<<" with Magick++ failed."<<endl
	    <<"  Caught exception: "<<endl
	    <<"  "<<error_.what()<<endl );
    return -1; // error
  };
  return  BIAS_IO_SUCCESS;// OK
}


// JW
int ImageIO::ExportMagickPP(const std::string& FileName, 
                            const ImageBase& input,
                            const int &quality,
                            const bool &writeMetaData)
{
  try {    
#ifdef WIN32
    // initialize DLL loaded library (not required on Linux but once on Windows)
    Magick::InitializeMagick(NULL);
#endif
    
    Magick::Image image; // temporary used for saving to disk
    
    // determine export image format/channels:
    std::string pixmap;
    switch (input.GetChannelCount()) {
    case 1:
      pixmap="K"; break;
    case 2:
      pixmap="KA"; break;
    case 3:
      pixmap="RGB"; break;
    case 4:
      pixmap="RGBA"; break;
    default:
      BIASERR("unsupported channelcount. Could not write image with Magick++");
      return -1;
    }
    
    // determine per pixel input/output storagetype: 
    Magick::StorageType magick_ST=Magick::CharPixel; // default
    // CharPixel, ShortPixel, IntegerPixel, FloatPixel, or DoublePixel
    switch (input.GetStorageType()){
    case BIAS::ImageBase::ST_unsignedchar:
    case BIAS::ImageBase::ST_char:
      magick_ST=Magick::CharPixel; break;
    case BIAS::ImageBase::ST_unsignedshortint:
    case BIAS::ImageBase::ST_shortint:
      magick_ST=Magick::ShortPixel; break;
    case BIAS::ImageBase::ST_unsignedint:
    case BIAS::ImageBase::ST_int:
      magick_ST=Magick::IntegerPixel; break;
    case BIAS::ImageBase::ST_float:
      magick_ST=Magick::FloatPixel; break;
    case BIAS::ImageBase::ST_double:
      magick_ST=Magick::DoublePixel; break;
    default:
      BIASERR("unsupported input image BIAS StorageType "<<input.GetStorageType() );
    };
    
    // Read a file into image object using Magick++ library function:
    unsigned char* p_imageData     =  (unsigned char*)input.GetImageData();
    BIASASSERT(p_imageData!=NULL);
    COUT("DBG writing "<<input.GetWidth()<<"x"<<input.GetHeight()<<" "<<pixmap<<" image "<<FileName<<" using ImageMagick++ "<<endl);
    image.read(input.GetWidth(), // dimension
	       input.GetHeight(),
	       pixmap, // output map type. e.g. RGBA
	       magick_ST, // desired storagetype
	       p_imageData  // pointer to our own data array to be written to disk
	       );
    image.quality(quality);
    if (writeMetaData) BIASERR("writing MetaData is not implemented");
	image.comment( "Hello world, this is a comment written by BIAS::ImageIO::WriteMagickPP.\n");
    image.write(FileName); // write to disk using Magick++ 
    
  } catch( Magick::Exception &error_ ) {
    BIASERR("Writing "<<FileName<<" with Magick++ failed."<<endl
	    <<"  Caught exception: "<<endl
	    <<"  "<<error_.what()<<endl );
    return -1; // error
  };
  return  BIAS_IO_SUCCESS;// OK
}

#endif // BIAS_HAVE_IMAGEMAGICKLIB  





int ImageIO::WriteUnbuffered(const std::string& FileName, ImageBase &img)
{
#ifdef WIN32
  BIASERR("ImageIO::WriteUnbuffered() not implemented in Win32");
  abort();
#else
  if (!img.GetUID().IsValid()) 
    img.SetUID(UUID::GenerateUUID());
  
std::ostringstream FlatImage;
  FlatImage << img;
  img.GetMetaData()->WriteBinary(FlatImage);
  FlatImage.flush();
  unsigned long int is = FlatImage.str().size();
//   cout <<"FlatImage.size: "<<is<<endl;;
  if (MemAlignedBufferSize_ < is) {
    if (MemAlignedBuffer_ != NULL) free(MemAlignedBuffer_);
    MemAlignedBufferSize_ = 512;
    while (MemAlignedBufferSize_ < is) MemAlignedBufferSize_ +=512;
//     cout <<"New MemAlignedBuffer_ with size "<<MemAlignedBufferSize_ <<endl;
    posix_memalign((void**)&MemAlignedBuffer_, 512, MemAlignedBufferSize_);
    if (MemAlignedBuffer_ == NULL) {
      BIASERR("Can not posix_memalign()"<<MemAlignedBufferSize_<<" bytes");
      return -1;
    }
  }
  // copy flattened image into Buffer
  memcpy(MemAlignedBuffer_, FlatImage.str().c_str(), is);

  int fd = open(FileName.c_str(),O_WRONLY | O_CREAT | O_DIRECT);
  if (fd <0) {
    BIASERR("Can not open/create file ");
    perror(FileName.c_str());
    return -1;
  }
  fchmod(fd,S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
  int res = write(fd,MemAlignedBuffer_, MemAlignedBufferSize_);
  if (res <0) {
    BIASERR("Can not write with O_DIRECT to file ");
    perror(FileName.c_str());
    return -1;
  }

  close(fd);

  return 0;
#endif
}










