/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "UUID.hh"

#include <sstream>


using namespace BIAS;
using namespace std;

timeval BIAS::UUID::LastGenerationTime_ = {0,0};


BIAS::UUID::UUID(const std::string &sID){
  SetFromString(sID);
}

BIAS::UUID::UUID(const unsigned long & n){
    SetFromLong(n);
}


bool BIAS::UUID::SetFromString(const std::string &sID) {
#ifdef _WIN32
  char *tmp = const_cast<char *>(sID.c_str());
  unsigned char * cUUIDwin = reinterpret_cast<unsigned char *>(tmp);
  if (UuidFromString(cUUIDwin, &theID_)!=RPC_S_OK) {
    BIASERR("UUID could not be read!");
    IsValid_ = false;
  }
  else IsValid_ = true;
//  std::cout <<"After Parser: "<<std::hex<<theID_.Data1<<"-"
//     <<theID_.Data2<<std::dec<<std::endl;
#else
  if (sID.length()!=36) 
    BIASERR("Wrong size of string to construct from: "<<sID);
  if (uuid_parse(sID.c_str(), theID_)!=0) {
    BIASERR("UUID could not be read!");
    IsValid_ = false;
  }
  else   IsValid_ = true;
#endif
  return IsValid_;
}


// jw
bool BIAS::UUID::SetFromLong(const unsigned long & n) {
    std::stringstream ss;
    // unused prefix:
    ss<<"00000000-0000-0000-0000-";
    // 12 digit hex format with leading zeroes
    ss<<setfill('0')<<setw(12)<<noshowbase<<hex
        <<n;
    return SetFromString( ss.str() );
}


BIAS::UUID BIAS::UUID::GenerateUUID() {

#ifdef _WIN32
  //JW details:  http://msdn.microsoft.com/library/default.asp?url=/library/en-us/rpc/rpc/uuid_1.asp 

  uuid_t* theID; 
  theID = new uuid_t;
  // creation uses ethernet/token ring addresses; UUID is guaranteed to be unique for the network
  HRESULT hr;
  hr = UuidCreate(theID);
  if (hr!=RPC_S_OK) {
    if (hr==RPC_S_UUID_LOCAL_ONLY) {
      BIASERR("UUID only guarantied unique for the generating machine!");
    } else {
      uuid_t emptyID = {0, 0, 0, {0x0,0x0, 0x0,0x0, 0x0,0x0, 0x0,0x0} };
      *theID = emptyID;
      BIASERR("Could not create UUID!");
    }
  }
  return BIAS::UUID(*theID);

#else //_WIN32
  uuid_t theID;
  uuid_generate_time(theID);
  /*
  for (int i=0; i<sizeof(uuid_t); i++) 
    std::cout << (short)(((unsigned char*)(&theID))[i])<<" ";
  std::cout<<std::endl;
  */
  
  timeval CurTime;
  uuid_time(theID, &CurTime);

  if ((CurTime.tv_sec  > LastGenerationTime_.tv_sec) ||
     ((CurTime.tv_sec == LastGenerationTime_.tv_sec) &&
      (CurTime.tv_usec > LastGenerationTime_.tv_usec))) {
    LastGenerationTime_.tv_sec  = CurTime.tv_sec;
    LastGenerationTime_.tv_usec = CurTime.tv_usec;
  } else {
    BIASERR("There is a system clock problem !  The last time used for"
            <<" UUID generation lies in the future ! Maybe someone has turned"
            <<" back the clock (NTP ?). UUIDs may not be unique anymore."
            <<" CurTime="<<CurTime.tv_sec<<":"<<CurTime.tv_usec
            <<" LastTime="<<LastGenerationTime_.tv_sec<<":"
            <<LastGenerationTime_.tv_usec); 
  }
  return BIAS::UUID(theID);
#endif //_WIN32
}


void BIAS::UUID::GetString(std::string& sUUID) const {
#ifdef _WIN32
  unsigned char * cUUIDwin;
  UuidToString (const_cast<uuid_t*>(&theID_), &cUUIDwin);
  sUUID = (char *)cUUIDwin;
#else
  char cUUID[37];
  uuid_unparse(theID_, cUUID);
  sUUID = cUUID;
#endif
}

// jw
std::string BIAS::UUID::GetString() const {
    std::string str;
    GetString(str);
    return str;
}


time_t BIAS::UUID::GetUUIDCreationTime() const {
#ifdef _WIN32
  BIASERR("Not yet implemented!");
  return time_t();
#else
  timeval CreationTime;
  time_t thetime;
  thetime = uuid_time(theID_, &CreationTime);
  return thetime;
#endif
}


