/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004, 2005    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __POLYNOM_HH__
#define __POLYNOM_HH__

#include <vector>
#include "Base/Debug/Debug.hh"
#include "Base/Debug/Error.hh"

namespace BIAS{
  /**@class Polynom
     @ingroup g_math
     @brief class for Polynoms of arbitary order

     class Polynom represents a Polynom of arbitary order, where 
     ith element is the coefficient of the ith exponent

     @author Marcel Lilienthal, June 2005
  */
  class Polynom : public Debug
  {
  public:
    Polynom(){
      order_=0;
    };
    ~Polynom(){};
    
    /**
       constructs a polynom of order=i 
       the coefficients are 0

       @author Marcel Lilienthal
     */
    Polynom(unsigned int i){
      order_=i; 
      poly_.resize(i,0.0); 
    };
    
    /**
       sets the order of polynom to i 
       new generated coefficients are 0

       @author Marcel Lilienthal
     */
    inline void SetOrder(unsigned int i){
      order_=i; 
      poly_.resize(i,0.0);
    };

    inline unsigned int GetOrder() const { return order_; };

    //operators
    /**
       returns a reference to ith coefficent of the polynom readonly

       @author Marcel Lilienthal
     */
    inline const double& operator[](const unsigned int i) const;

    /**
       returns a reference to ith coefficent of the polynom writable

       @author Marcel Lilienthal
     */
    inline double& operator[](const unsigned int i);


    inline Polynom& operator=(const Polynom& a);
    inline bool operator==(const Polynom& a) const;
    inline bool operator!=(const Polynom& a) const;

  private:
    std::vector<double> poly_; 
    unsigned int order_;
  };//class

  inline const double& Polynom::operator[](const unsigned int i) const{
    BIASASSERT(i<order_);
    return poly_[i];
  }

  inline double& Polynom::operator[](const unsigned int i){
    BIASASSERT(i<order_);
    return poly_[i];
  }

  inline Polynom& Polynom::operator=(const Polynom& a){
    if (this != &a){
      order_ = a.order_;
      poly_.resize(order_);
      for(unsigned int i=0;i<order_; i++){
        poly_[i]=a.poly_[i];
      }
    }
    return *this;
  }
  
  inline bool Polynom::operator==(const Polynom& a) const{
    bool res = (this == &a);

    if(!res){
      res = order_ == a.order_;
      if(res){
        for(unsigned int i=0; i< order_; i++){
          res = poly_[i]==a.poly_[i];
          if(!res)
            return res;
        }
      }
    } 
    return res;
  }

 

  inline bool Polynom::operator!=(const Polynom& a) const{
    return !operator==(a);
  }

  inline Polynom operator+(const Polynom& a, const Polynom& b){
    Polynom c; 
    c=a;
    if(a.GetOrder() < b.GetOrder())
      c.SetOrder(b.GetOrder());
    for(unsigned int i=0; i<b.GetOrder(); i++)
      c[i]+=b[i];
    return c;
  }
  
  inline Polynom operator-(const Polynom& a, const Polynom& b){
    Polynom c;
    c=a;
    if(a.GetOrder() < b.GetOrder())
      c.SetOrder(b.GetOrder());
    for(unsigned int i=0; i<b.GetOrder(); i++)
      c[i]-=b[i];
    return c;
  }
  
  inline Polynom operator*(const Polynom& a, const Polynom& b){
    Polynom c(a.GetOrder() + b.GetOrder() - 1);

    for(unsigned int i=0; i<a.GetOrder(); i++){
      for(unsigned int j=0; j<b.GetOrder(); j++){
        c[i+j] += a[i] * b[j];
      }
    }
    return c;
  }

  inline Polynom operator*(const double d, const Polynom& a){
    Polynom c(a.GetOrder());

    for(unsigned int i=0; i<a.GetOrder(); i++){
      c[i]=a[i]*d;
    }
    return c;
  }
  
  inline Polynom operator*(const Polynom& a, const double d){
    Polynom c(a.GetOrder());

    for(unsigned int i=0; i<a.GetOrder(); i++){
      c[i]=a[i]*d;
    }
    return c;
    }



}//namespace

#endif //__POLYNOM_HH__
