/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __Tensor3D_hh__
#define __Tensor3D_hh__

#include "Vector.hh"
#include <Base/Debug/Error.hh>

namespace BIAS {

  template <class T=double>
  class DLLEXPORTAPI Tensor3D 
  {
  public:
    Tensor3D();
    Tensor3D(const int i, const int j, const int k, 
                const T& def=(T)0.0);
    Tensor3D(const Tensor3D<T>& t);
    
    ~Tensor3D();

    /// allocates internal data
    void NewSize(const int i, const int j, const int k, 
                 const T& def=(T)0.0);
    /// calls NewSize(i, j, k, default)
    void newsize(const int i, const int j, const int k, 
                 const T& def=(T)0.0);

    /// frees internal data 
    void Destroy();
    /// calls Destroy() 
    void destroy();

    int GetDim1() const;

    int GetDim2() const;

    int GetDim3() const;

    /** @return all tensor elements in a vector
        @author mdunda */
    BIAS::Vector<T> ToVector();

    /** @brief set the tensor by the rows of the vector. Vector size must
        be at least Dim1*Dim2*Dim3!
        @author mdunda */
    void SetFromVector(const BIAS::Vector<T> &v);
    
    inline T* GetData()
    { return _Data; }

    inline T*** GetDataArray()
    { return _DataArray; }

    inline const T* GetData() const
    { return (const T*)_Data; }

    /// set all elements to zero
    void SetZero();

    T GetLength() const;

    void Scale(T factor);

    Tensor3D<T>& operator=(const Tensor3D<T>& t);
    /** @brief checks if all tensor entries are equal (up-to-scale-equality
        is not checked)
        @author mdunda 07 2004 */
    bool operator==(const Tensor3D<T>& t); 

    /** @brief direct data access through () operator: T_i^{jk}
        @author mdunda 07 2004 */
    inline T& operator()(unsigned int i, unsigned int j, unsigned int k)
    { 
      BIASASSERT((int)i<_Dim1);
      BIASASSERT((int)j<_Dim2);
      BIASASSERT((int)k<_Dim3);
      return _DataArray[i][j][k]; 
    }

  protected:
    // the actual data, in memory as row first, column second, depth third
    // i.e. _Data[9] is the Tensor3D_{0, 0, 1}
    T *_Data;
    // inverse pointer representation: depth z, row y and column x
    // can be addressed by _DataArray[z][y][x]
    // this is analogue to the ImageDataArray
    T ***_DataArray;
    int _Dim1, _Dim2, _Dim3, _Size;

  }; // class

} // namespace

#endif // __Tensor3D_hh__
