/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "FFT2D_Tiles.hh"
#include <fftw3.h>
#include <Base/Debug/Error.hh>


using namespace BIAS;
using namespace std;



//////////////////////////////////////////////////////////////////////////
//                 instantiation
//////////////////////////////////////////////////////////////////////////
#define FILTER_INSTANTIATION_CLASS FFT2D_Tiles
#define FILTER_INSTANTIATION_NO_UNSIGNED_OUTPUT
#include "Filterinst.hh"


template <class InputStorageType, class OutputStorageType>
FFT2D_Tiles<InputStorageType, OutputStorageType>::
FFT2D_Tiles()
{
  _TileSizeX = 0;
  _TileSizeY = 0;
  _NumTilesX = 0;
  _NumTilesY = 0;
}

template <class InputStorageType, class OutputStorageType>
FFT2D_Tiles<InputStorageType, OutputStorageType>::
FFT2D_Tiles::~FFT2D_Tiles()
{
  Release();
  fftw_cleanup();
}

template <class InputStorageType, class OutputStorageType>
void FFT2D_Tiles<InputStorageType, OutputStorageType>::
Init(int width, int height, int tilewidth, int tileheight)
{
#ifdef BIAS_DEBUG
  if (_SizeX!=0 || _SizeY!=0 ||  _OutNum!=0 || _out || _in || _p_forward){
    BIASERR("FFT2D_Tiles: call Release() before calling Init() a second time");
    BIASASSERT(false);
  }
#endif
  _SizeX=height;
  _SizeY=width;

  _TileSizeX = tileheight;
  _TileSizeY = tilewidth;
  _NumTilesX = _SizeX / _TileSizeX;
  if (_SizeX % _TileSizeX >0) _NumTilesX++;
  _NumTilesY = _SizeY / _TileSizeY;
  if (_SizeY % _TileSizeY >0) _NumTilesY++;



  _Size = _TileSizeX*_TileSizeY;
  _OutSizeY = (_TileSizeY>>1)+1;
  _OutNum=_TileSizeX * _OutSizeY;
  _out =(fftw_complex*)fftw_malloc(sizeof(fftw_complex) * _OutNum);
  _in = (double*)fftw_malloc(sizeof(double) * _Size);
 // this is slow a few seconds
  _p_forward = fftw_plan_dft_r2c_2d(_TileSizeX, _TileSizeY, _in, _out, FFTW_MEASURE);
  _p_reverse = fftw_plan_dft_c2r_2d(_TileSizeX, _TileSizeY, _out, _in, FFTW_MEASURE);
}






template <class InputStorageType, class OutputStorageType>
void FFT2D_Tiles<InputStorageType, OutputStorageType>::
Release()
{
  if (_p_forward) fftw_destroy_plan(_p_forward);
   if (_in) fftw_free(_in); 
  if (_out) fftw_free(_out);
  _in=NULL;
  _out=NULL;
  _p_forward=NULL;
   _SizeX=_SizeY=_Size=_OutNum=0;
}


template <class InputStorageType, class OutputStorageType>
int FFT2D_Tiles<InputStorageType, OutputStorageType>::
Forward_(const Image<InputStorageType>& src)
{
  return 0;

}


template <class InputStorageType, class OutputStorageType>
int FFT2D_Tiles<InputStorageType, OutputStorageType>::
Reverse_(const Image<OutputStorageType>& src)
{
  return 0;
}




    /** dst.GetChannelCount()==2*src.GetCHannelCount() */
template <class InputStorageType, class OutputStorageType>
int FFT2D_Tiles<InputStorageType, OutputStorageType>::
Filter(const Image<InputStorageType>& src, Image<OutputStorageType>& dst)
{
 
  if (!dst.IsEmpty()) dst.Release();
  dst.Init(_OutSizeY, _SizeX, 2);
  // copy and cast back int dst, two channel mode
  fftw_complex *outp = _out;
  OutputStorageType *oimgp = dst.GetImageData();
  for ( int i=0;i<_OutNum; i++){
    *oimgp = (OutputStorageType) (*outp)[0];
    oimgp++;
    *oimgp = (OutputStorageType) (*outp)[1];
    oimgp++;
    outp++;
  }
  return 0;
}



/** dstX.GetChannelCount()==src.GetCHannelCount() */
template <class InputStorageType, class OutputStorageType>
int FFT2D_Tiles<InputStorageType, OutputStorageType>::
Filter(const Image<InputStorageType>& src, Image<OutputStorageType>& dst1, 
       Image<OutputStorageType>& dst2)
{

  return 0;
}


/** dst.GetChannelCount()==src.GetCHannelCount() */
template <class InputStorageType, class OutputStorageType>
int FFT2D_Tiles<InputStorageType, OutputStorageType>::
TransformAbs(const Image<InputStorageType>& src, Image<OutputStorageType>& dst)
{
  Image<InputStorageType> tmpImg;
  InputStorageType **Iida ;

  if ( ((int) src.GetHeight()  != _TileSizeX * _NumTilesX)
       || ((int) src.GetWidth() != _TileSizeY * _NumTilesY)) {
    tmpImg = src;
    tmpImg.ZeroPad( _TileSizeX * _NumTilesX, _TileSizeY * _NumTilesY );
    Iida = tmpImg.GetImageDataArray();
 }
  else Iida = src.GetImageDataArray();
  // prepare Image
  if (!dst.IsEmpty()) dst.Release();
  dst.Init( _OutSizeY * _NumTilesY,_TileSizeX * _NumTilesX,1);

  OutputStorageType **Oida = dst.GetImageDataArray();
  // for all tiles
  for ( int ty=0; ty<_NumTilesY; ty++) {
    for ( int tx=0; tx<_NumTilesX; tx++) {
      double *inp = _in;
      fftw_complex *outp = _out;

      // copy this tile int _in;
      for ( int y = ty*_TileSizeY; y<(ty+1)*_TileSizeY; y++)
	for ( int x = tx*_TileSizeX; x<(tx+1)*_TileSizeX; x++) {
	  *inp++ = Iida[y][x];
	}
      // fft it
      fftw_execute(_p_forward);
      // copy result back
      for ( int y = ty*_OutSizeY; y<(ty+1)*_OutSizeY; y++)
	for ( int x = tx*_TileSizeX; x<(tx+1)*_TileSizeX; x++) {
	  // remeber: x is height and y is width
	  Oida[x][y] = (OutputStorageType) sqrt((*outp[0]* *outp[0]) + 
						(*outp[1]* *outp[1]));
	  outp++;
	}
    }
  }

  return 0;
}





template <class InputStorageType, class OutputStorageType>
int FFT2D_Tiles<InputStorageType, OutputStorageType>::
TransformReverse(const Image<OutputStorageType>& src,
		Image<InputStorageType>& dst)
{
  Reverse_(src);
  if (!dst.IsEmpty()) dst.Release();
  dst.Init(_SizeX, _SizeY,1);
  // copy and cast back int dst, abs mode
  double *inp = _in;
  InputStorageType *oimgp = dst.GetImageData();
  for ( int i=0;i<_Size; i++){
    *oimgp = (InputStorageType) *inp;
    oimgp++;
    inp++;
  }

  return 0;
}



