/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "GradientSimple.hh"
#include "FilterMask.hh"

using namespace BIAS;
using namespace std;

//////////////////////////////////////////////////////////////////////////
//                 instantiation
//////////////////////////////////////////////////////////////////////////
#define FILTER_INSTANTIATION_CLASS GradientSimple
#define FILTER_INSTANTIATION_NO_UNSIGNED_OUTPUT
#include "Filterinst.hh"

//////////////////////////////////////////////////////////////////////////
//                 implementation
//////////////////////////////////////////////////////////////////////////
template <class InputStorageType, class OutputStorageType> void
GradientSimple<InputStorageType, OutputStorageType>::
InitKernel_() 
{
  // this is a filtermask in the sense of convolution
  // it is reflected at its symmetry center before being "dropped"
  // onto the image !
  Vector<FM_INT> h(3), v(1);
  h[0] = 1;
  h[1] = 0;
  h[2] = -1;
  v[0] = 1;

  FilterMask f(h, v, 1, 0);
  f.CreateFloatFilter();
  _Conv.SetKernel(f);

}

template <class InputStorageType, class OutputStorageType>
GradientSimple<InputStorageType, OutputStorageType>::
GradientSimple()
  : FilterNTo2N<InputStorageType, OutputStorageType>() 
{
  InitKernel_();
}

template <class InputStorageType, class OutputStorageType>
GradientSimple<InputStorageType, OutputStorageType>::
GradientSimple(Param & para, const std::string& postfix)
  : FilterNTo2N<InputStorageType, OutputStorageType>(para,
                                                     "-GradientSimple"+postfix)
{
  InitKernel_();
}

template <class InputStorageType, class OutputStorageType>
GradientSimple<InputStorageType, OutputStorageType>::
GradientSimple(Param & para, const int grp, const std::string& postfix)
  : FilterNTo2N<InputStorageType, OutputStorageType>(para, grp, 
                                                     "-GradientSimple"+postfix)
{
  InitKernel_();
}

template <class InputStorageType, class OutputStorageType>
GradientSimple<InputStorageType, OutputStorageType>::
~GradientSimple()
{

}


template <class InputStorageType, class OutputStorageType>
int GradientSimple<InputStorageType,OutputStorageType>::
Filter(const Image<InputStorageType>& src, Image<OutputStorageType>& grad) 
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "GradientSimple::Filter(src, dst)\n");
  Image<OutputStorageType> gy;
  int res=Filter(src, grad, gy);
  grad.AppendChannel(gy);
  return res;
}

template <class InputStorageType, class OutputStorageType>
int GradientSimple<InputStorageType,OutputStorageType>::
Filter(const Image<InputStorageType>& src, Image<OutputStorageType>& gx, 
       Image<OutputStorageType>& gy)
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "GradientSimple::Filter(src, gx, gy)\n");
  int res=0;
  if (!gx.SamePixelAndChannelCount(src)){
    gx.Release();
    gx.Init(src.GetWidth(), src.GetHeight(), src.GetChannelCount());
  }
  if (!gy.SamePixelAndChannelCount(src)){
    gy.Release();
    gy.Init(src.GetWidth(), src.GetHeight(), src.GetChannelCount());
  }
 
  if (src.GetChannelCount()==1) {
    if (_Conv.CalculationInFloat()){
      res = SimpleGreyValidFloat_(src, gx, gy);
    } else {
      res = SimpleGreyValidInt_(src, gx, gy);
    }
  } else {
    if (_Conv.Filter(src, gx)!=0) {
      BIASERR("Error in gradient x computation...");
      return -1;
    }
    FilterMask f;
    _Conv.GetKernel(f);
    f.Transpose();
    _Conv.SetKernel(f);
    res =  _Conv.Filter(src, gy);
    // re-set original kernel
    f.Transpose();
    _Conv.SetKernel(f);
  }
  return res;
}

template <class InputStorageType, class OutputStorageType>
int GradientSimple<InputStorageType,OutputStorageType>::
Filter(const Image<InputStorageType>& src, Image<OutputStorageType>& gx, 
       Image<OutputStorageType>& gy, Image<OutputStorageType>& absg) 
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, 
            "GradientSimple::Filter(src, gx, gy, abs)\n");
  if (!src.SamePixelAndChannelCount(absg)){
    if (!absg.IsEmpty()) absg.Release();
    absg.Init(src.GetWidth(), src.GetHeight(), src.GetChannelCount());
  }
  
  int res, absres=0;
  res = Filter(src, gx, gy);

  absres = VecLen(gx, gy, absg);
  
  return (res==0 && absres==0)?0:-1;
}




template <class InputStorageType, class OutputStorageType>
int GradientSimple<InputStorageType, OutputStorageType>::
SimpleGreyValidFloat_(const Image<InputStorageType>& src,
                      Image<OutputStorageType>& gWE,
                      Image<OutputStorageType>& gNS)
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "GradientSimple::SimpleGreyValidFloat_\n");
 
  OutputStorageType *gx=gWE.GetImageData();
  OutputStorageType *gy=gNS.GetImageData();
  const int width=(int)src.GetWidth();
  const int nwidth = - width;
  
#ifdef BIAS_DEBUG
  BIASASSERT(src.GetChannelCount()==1);
  BIASASSERT(src.SamePixelAndChannelCount(gWE));
  BIASASSERT(src.SamePixelAndChannelCount(gNS));
  BIASASSERT(gx!=gy);
  BIASASSERT(_Conv.CalculationInFloat());
#endif

  int minx, miny, maxx, maxy;
  src.GetROI()->GetCorners(minx, miny, maxx, maxy);
  gWE.GetROI()->SetCorners(minx+1, miny+1, maxx-1, maxy-1);
  gNS.GetROI()->SetCorners(minx+1, miny+1, maxx-1, maxy-1);

  InputStorageType *ph;
  
  int offset;
  OutputStorageType *pgx, *pgy, *end2, *lend2;
 
 
  int step=width-maxx+minx;

  offset=minx+1+(miny+1)*width;
  ph = src.GetImageData() + offset;
  pgx=gx+offset;
  pgy=gy+offset;
  step+=2;
  end2=gx+maxx+width*(maxy-2)-1;
  lend2=gx+maxx+width*(miny+1)-1;
  while (pgx<end2){
    while (pgx<lend2){
      *pgx++ = (OutputStorageType)(ph[1] - ph[-1]);
      *pgy++ = (OutputStorageType)(ph[width] - ph[nwidth]);
      ph++;
    }
    ph+=step;
    pgx+=step;
    pgy+=step;
    lend2+=width;
  }
  return 0;
}


template <class InputStorageType, class OutputStorageType>
int GradientSimple<InputStorageType, OutputStorageType>::
SimpleGreyValidInt_(const Image<InputStorageType>& src,
                      Image<OutputStorageType>& gWE,
                      Image<OutputStorageType>& gNS)
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "GradientSimple::SimpleGreyValidInt_\n");
 
  OutputStorageType *gx=gWE.GetImageData();
  OutputStorageType *gy=gNS.GetImageData();
  const int width=(int)src.GetWidth();
  const int nwidth = - width;
  
#ifdef BIAS_DEBUG
  BIASASSERT(src.GetChannelCount()==1);
  BIASASSERT(src.SamePixelAndChannelCount(gWE));
  BIASASSERT(src.SamePixelAndChannelCount(gNS));
  BIASASSERT(gx!=gy);
  BIASASSERT(!_Conv.CalculationInFloat());
#endif



  int minx, miny, maxx, maxy;
  src.GetROI()->GetCorners(minx, miny, maxx, maxy);
  gWE.GetROI()->SetCorners(minx+1, miny+1, maxx-1, maxy-1);
  gNS.GetROI()->SetCorners(minx+1, miny+1, maxx-1, maxy-1);

  InputStorageType *ph;
 
  int offset;
  OutputStorageType *pgx, *pgy, *end2, *lend2;
  
 
  int step=width-maxx+minx;

   offset=minx+1+(miny+1)*width;
  ph = src.GetImageData()+offset;
 
  pgx=gx+offset;
  pgy=gy+offset;
  step+=2;
  end2=gx+maxx+width*(maxy-2)-1;
  lend2=gx+maxx+width*(miny+1)-1;
  while (pgx<end2){
    while (pgx<lend2){
      *pgx++ = (OutputStorageType)(ph[1] - ph[-1]);
      *pgy++ = (OutputStorageType)(ph[width] - ph[nwidth]);
      ph++;
    }
    ph+=step;
    pgx+=step;
    pgy+=step;
    lend2+=width;
  }

  return 0;
}

