/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "GradientSobel3x3.hh"
#include "FilterMask.hh"

using namespace BIAS;
using namespace std;

//////////////////////////////////////////////////////////////////////////
//                 instantiation
//////////////////////////////////////////////////////////////////////////
#define FILTER_INSTANTIATION_CLASS GradientSobel3x3
#define FILTER_INSTANTIATION_NO_UNSIGNED_OUTPUT
#include "Filterinst.hh"

//////////////////////////////////////////////////////////////////////////
//                 implementation
//////////////////////////////////////////////////////////////////////////
template <class InputStorageType, class OutputStorageType> void
GradientSobel3x3<InputStorageType, OutputStorageType>::
InitKernel_() 
{
  // this is a filtermask in the sense of convolution
  // it is reflected at its symmetry center before being "dropped"
  // onto the image !
  Vector<FM_INT> h(3), v(3);
  h[0] = 1;
  h[1] = 0;
  h[2] = -1;
  v[0] = v[2] = 1;
  v[1] = 2;

  FilterMask f(h, v, 1, 2);
  //FilterMask f(h, v, 2, 2);
  f.CreateFloatFilter();
  _Conv.SetKernel(f);

}

template <class InputStorageType, class OutputStorageType>
GradientSobel3x3<InputStorageType, OutputStorageType>::
GradientSobel3x3()
  : FilterNTo2N<InputStorageType, OutputStorageType>() 
{
  InitKernel_();
  _fwork = _fwork2 = NULL;
  _iwork = _iwork2 = NULL;
}

template <class InputStorageType, class OutputStorageType>
GradientSobel3x3<InputStorageType, OutputStorageType>::
GradientSobel3x3(Param & para, const std::string& postfix)
  : FilterNTo2N<InputStorageType, OutputStorageType>(para, "-GradientSobel3x3"+postfix)
{
  InitKernel_();
  _fwork = _fwork2 = NULL;
  _iwork = _iwork2 = NULL;
}

template <class InputStorageType, class OutputStorageType>
GradientSobel3x3<InputStorageType, OutputStorageType>::
GradientSobel3x3(Param & para, const int grp, const std::string& postfix)
  : FilterNTo2N<InputStorageType, OutputStorageType>(para, grp, "-GradientSobel3x3"+postfix)
{
  InitKernel_();
  _fwork = _fwork2 = NULL;
  _iwork = _iwork2 = NULL;
}

template <class InputStorageType, class OutputStorageType>
GradientSobel3x3<InputStorageType, OutputStorageType>::
~GradientSobel3x3()
{
  _DeleteInternalMem();
}


template <class InputStorageType, class OutputStorageType>
int GradientSobel3x3<InputStorageType,OutputStorageType>::
Filter(const Image<InputStorageType>& src, Image<OutputStorageType>& grad) 
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "GradientSobel3x3::Filter(src, dst)\n");
  Image<OutputStorageType> gy;
  int res=Filter(src, grad, gy);
  grad.AppendChannel(gy);
  return res;
}

template <class InputStorageType, class OutputStorageType>
int GradientSobel3x3<InputStorageType,OutputStorageType>::
Filter(const Image<InputStorageType>& src, Image<OutputStorageType>& gx, 
       Image<OutputStorageType>& gy)
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "GradientSobel3x3::Filter(src, gx, gy)\n");
  int res=0;
  if (!gx.SamePixelAndChannelCount(src)){
    gx.Release();
    gx.Init(src.GetWidth(), src.GetHeight(), src.GetChannelCount());
  }
  if (!gy.SamePixelAndChannelCount(src)){
    gy.Release();
    gy.Init(src.GetWidth(), src.GetHeight(), src.GetChannelCount());
  }
 
  if (src.GetChannelCount()==1) {
    if (_InternalMemWidth < (int)src.GetWidth() || 
        _InternalMemHeight < (int)src.GetHeight() ) _DeleteInternalMem();
    if (_Conv.CalculationInFloat()){
      if (_fwork == NULL) _AllocInternalMem(src.GetWidth(), src.GetHeight());
      switch (*_FilterBorderHandling){
      case TBH_valid:
        res = Sobel3x3GreyValidFloat_(src, gx, gy);
        break;
      case TBH_same:
        res = Sobel3x3GreySameFloat_(src, gx, gy);
        break;
      default:
        BIASERR("unfinished");
        res = -1;
        break;
      }
    } else {
      if (_iwork == NULL) _AllocInternalMem(src.GetWidth(), src.GetHeight());
      res = Sobel3x3GreyValidInt_(src, gx, gy);
    }
  } else {
    if (_Conv.Filter(src, gx)!=0) {
      BIASERR("Error in gradient x computation...");
      return -1;
    }
    FilterMask f;
    _Conv.GetKernel(f);
    f.Transpose();
    _Conv.SetKernel(f);
    res =  _Conv.Filter(src, gy);
    // re-set original kernel
    f.Transpose();
    _Conv.SetKernel(f);
  }
  return res;
}

template <class InputStorageType, class OutputStorageType>
int GradientSobel3x3<InputStorageType,OutputStorageType>::
Filter(const Image<InputStorageType>& src, Image<OutputStorageType>& gx, 
       Image<OutputStorageType>& gy, Image<OutputStorageType>& absg) 
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "GradientSobel3x3::Filter(src, gx, gy, abs)\n");
  if (!src.SamePixelAndChannelCount(absg)){
    if (!absg.IsEmpty()) absg.Release();
    absg.Init(src.GetWidth(), src.GetHeight(), src.GetChannelCount());
  }
  
  int res, absres=0;
  res = Filter(src, gx, gy);

  absres = VecLen(gx, gy, absg);
  
  return (res==0 && absres==0)?0:-1;
}




template <class InputStorageType, class OutputStorageType>
int GradientSobel3x3<InputStorageType, OutputStorageType>::
Sobel3x3GreyValidFloat_(const Image<InputStorageType>& src,
                        Image<OutputStorageType>& gWE,
                        Image<OutputStorageType>& gNS)
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "GradientSobel3x3::Sobel3x3GreyValidFloat_\n");
  InputStorageType *im=src.GetImageData();
  OutputStorageType *gx=gWE.GetImageData();
  OutputStorageType *gy=gNS.GetImageData();
  const int width=(int)src.GetWidth();
  const int nwidth = - width;
  const int height=(int)src.GetHeight();

#ifdef BIAS_DEBUG
  BIASASSERT(src.GetChannelCount()==1);
  BIASASSERT(src.SamePixelAndChannelCount(gWE));
  BIASASSERT(src.SamePixelAndChannelCount(gNS));
  BIASASSERT(gx!=gy);
  BIASASSERT(_Conv.CalculationInFloat());
#endif

  if (width>this->_InternalMemWidth || height>this->_InternalMemHeight){
    this->_DeleteInternalMem();
    this->_AllocInternalMem(width, height);
  }

  int minx, miny, maxx, maxy;
  src.GetROI()->GetCorners(minx, miny, maxx, maxy);
  gWE.GetROI()->SetCorners(minx+1, miny+1, maxx-1, maxy-1);
  gNS.GetROI()->SetCorners(minx+1, miny+1, maxx-1, maxy-1);

  InputStorageType *end, *p, *lend;
  CONV_FLOAT *ph, *pv;
  int offset;
  OutputStorageType *pgx, *pgy, *end2, *lend2;
  CONV_FLOAT tmp, tmpx, tmpy;
  int roiw=maxx-minx-1;
  int step=width-maxx+minx;

  // separated sobel 3x3 for gx and gy
  
  // first line
  offset=minx+miny*width;
  p=im+1+offset;
  ph=this->_fwork[0]+1+offset;
  pv=this->_fwork2[0]+offset+width;
  lend=p+roiw-1;
  while (p<lend){
    tmp = (CONV_FLOAT(*p))*2.0f;
    *ph++ = CONV_FLOAT(p[-1]) + tmp + CONV_FLOAT(p[1]);
    p++;
  }
  p+=step+1;
  ph+=step+1;
  lend+=width;
  end=im+maxx-1+(maxy-2)*width;
  while (p<end){
    tmp = (CONV_FLOAT(*p))*2.0f;
    *pv++ = CONV_FLOAT(p[nwidth]) + tmp + CONV_FLOAT(p[width]);
    ph++;
    p++;
    while (p<lend){
      tmp = (CONV_FLOAT(*p))*2.0f;
      *ph++ = CONV_FLOAT(p[-1]) + tmp + CONV_FLOAT(p[1]);
      *pv++ = CONV_FLOAT(p[nwidth]) + tmp + CONV_FLOAT(p[width]);
      p++;
    }
    tmp = (CONV_FLOAT(*p))*2.0f;
    *pv++ = CONV_FLOAT(p[nwidth]) + tmp + CONV_FLOAT(p[width]);
    ph++;
    p++;
    lend+=width;
    p+=step;
    ph+=step;
    pv+=step;
  }
  p++; ph++; 
  // last line
  while (p<lend){
    tmp = (CONV_FLOAT(*p))*2.0f;
    *ph++ = CONV_FLOAT(p[-1]) + tmp + CONV_FLOAT(p[1]);
    p++;
  }

  // now second run from temporary to gradients

  offset=minx+1+(miny+1)*width;
  ph=this->_fwork[0]+offset;
  pv=this->_fwork2[0]+offset;
  pgx=gx+offset;
  pgy=gy+offset;
  step+=2;
  end2=gx+maxx+width*(maxy-2)-1;
  lend2=gx+maxx+width*(miny+1)-1;
  while (pgx<end2){
    while (pgx<lend2){
      tmpx = (pv[1] - pv[-1])/8.0f; 
      tmpy = (ph[width] - ph[nwidth])/8.0f;
      *pgx++ = (OutputStorageType)tmpx;
      *pgy++ = (OutputStorageType)tmpy;
      pv++;
      ph++;
    }
    pv+=step;
    ph+=step;
    pgx+=step;
    pgy+=step;
    lend2+=width;
  }
  return 0;
}

template <class InputStorageType, class OutputStorageType>
int GradientSobel3x3<InputStorageType, OutputStorageType>::
Sobel3x3GreySameFloat_(const Image<InputStorageType>& src,
                        Image<OutputStorageType>& gWE,
                        Image<OutputStorageType>& gNS)
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "GradientSobel3x3::Sobel3x3GreySameFloat_\n");
  int res=0;
  res = Sobel3x3GreyValidFloat_(src, gWE, gNS);

  int minx, miny, maxx, maxy;
  src.GetROI()->GetCorners(minx, miny, maxx, maxy);
  register int x, y, xp, xm, yp, ym;

  InputStorageType **srcida=src.GetImageDataArray();
  OutputStorageType **gweida=gWE.GetImageDataArray();
  OutputStorageType **gnsida=gNS.GetImageDataArray();
  
  // upper left pixel
  x = minx;
  y = miny;
  xp = x+1;
  xm = x;
  yp = y+1;
  ym = y;
  gweida[y][x] = 
    OutputStorageType((float(srcida[ym][xp])-float(srcida[ym][xm])+
                       2.0*(float(srcida[y][xp])-float(srcida[y][xm]))+
                       float(srcida[yp][xp])-float(srcida[yp][xm]))/8.0);
  gnsida[y][x] = 
    OutputStorageType((float(srcida[yp][xm])-float(srcida[ym][xm])+
                       2.0*(float(srcida[yp][x])-float(srcida[ym][x]))+
                       float(srcida[yp][xp])-float(srcida[ym][xp]))/8.0);
  
  // upper right pixel
  x = maxx-1;
  y = miny;
  xp = x;
  xm = x-1;
  yp = y+1;
  ym = y;
  gweida[y][x] = 
    OutputStorageType((float(srcida[ym][xp])-float(srcida[ym][xm])+
                       2.0*(float(srcida[y][xp])-float(srcida[y][xm]))+
                       float(srcida[yp][xp])-float(srcida[yp][xm]))/8.0);
  gnsida[y][x] = 
    OutputStorageType((float(srcida[yp][xm])-float(srcida[ym][xm])+
                       2.0*(float(srcida[yp][x])-float(srcida[ym][x]))+
                       float(srcida[yp][xp])-float(srcida[ym][xp]))/8.0);

  // lower right pixel
  x = maxx-1;
  y = maxy-1;
  xp = x;
  xm = x-1;
  yp = y;
  ym = y-1;
  gweida[y][x] = 
    OutputStorageType((float(srcida[ym][xp])-float(srcida[ym][xm])+
                       2.0*(float(srcida[y][xp])-float(srcida[y][xm]))+
                       float(srcida[yp][xp])-float(srcida[yp][xm]))/8.0);
  gnsida[y][x] = 
    OutputStorageType((float(srcida[yp][xm])-float(srcida[ym][xm])+
                       2.0*(float(srcida[yp][x])-float(srcida[ym][x]))+
                       float(srcida[yp][xp])-float(srcida[ym][xp]))/8.0);

  // lower left pixel
  x = minx;
  y = maxy-1;
  xp = x+1;
  xm = x;
  yp = y;
  ym = y-1;
  gweida[y][x] = 
    OutputStorageType((float(srcida[ym][xp])-float(srcida[ym][xm])+
                       2.0*(float(srcida[y][xp])-float(srcida[y][xm]))+
                       float(srcida[yp][xp])-float(srcida[yp][xm]))/8.0);
  gnsida[y][x] = 
    OutputStorageType((float(srcida[yp][xm])-float(srcida[ym][xm])+
                       2.0*(float(srcida[yp][x])-float(srcida[ym][x]))+
                       float(srcida[yp][xp])-float(srcida[ym][xp]))/8.0);

  // upper row
  y = miny;
  yp = y+1;
  for (x=minx+1; x<maxx-1; x++){
    xp = x+1;
    xm = x-1;
    gweida[y][x] =
      OutputStorageType((2.0*(float(srcida[y][xp])-float(srcida[y][xm]))+
                         float(srcida[yp][xp])-float(srcida[yp][xm]))/6.0);
    gnsida[y][x] = 
      OutputStorageType((2.0*(float(srcida[yp][x])-float(srcida[y][x]))+
                         float(srcida[yp][xp])+float(srcida[yp][xm])-
                         float(srcida[y][xp])-float(srcida[y][xm]))/8.0);
  }
  // lower row
  y=maxy-1;
  ym = y-1;
  for (x=minx+1; x<maxx-1; x++){
    xp = x+1;
    xm = x-1;
    gweida[y][x] =
      OutputStorageType((2.0*(float(srcida[y][xp])-float(srcida[y][xm]))+
                         float(srcida[ym][xp])-float(srcida[ym][xm]))/6.0);
    gnsida[y][x] = 
      OutputStorageType((2.0*(float(srcida[y][x])-float(srcida[ym][x]))-
                         float(srcida[ym][xp])-float(srcida[ym][xm])+
                         float(srcida[y][xp])+float(srcida[y][xm]))/8.0);
  }
  // left column
  x = minx;
  xp = x+1;
  for (y=miny+1; y<maxy-1; y++){
    yp = y+1;
    ym = y-1;
    gweida[y][x] =
      OutputStorageType((2.0*(float(srcida[y][xp])-float(srcida[y][x]))+
                         float(srcida[ym][xp])+float(srcida[yp][xp])-
                         float(srcida[ym][x])-float(srcida[yp][x]))/8.0);
    gnsida[y][x] = 
      OutputStorageType((2.0*(float(srcida[yp][x])-float(srcida[ym][x]))+
                         float(srcida[yp][xp])-float(srcida[ym][xp]))/6.0);
  }

  // right column
  x = maxx-1;
  xm = x-1;
  for (y=miny+1; y<maxy-1; y++){
    yp = y+1;
    ym = y-1;
    gweida[y][x] =
      OutputStorageType((2.0*(float(srcida[y][x])-float(srcida[y][xm]))-
                         float(srcida[ym][xm])-float(srcida[yp][xm])+
                         float(srcida[ym][x])+float(srcida[yp][x]))/8.0);
    gnsida[y][x] = 
      OutputStorageType((2.0*(float(srcida[yp][x])-float(srcida[ym][x]))+
                         float(srcida[yp][xm])-float(srcida[ym][xm]))/6.0);
  }

  gWE.GetROI()->SetCorners(minx, miny, maxx, maxy);
  gNS.GetROI()->SetCorners(minx, miny, maxx, maxy);
  
  return res;
}

template <class InputStorageType, class OutputStorageType>
int GradientSobel3x3<InputStorageType, OutputStorageType>::
Sobel3x3GreyValidInt_(const Image<InputStorageType>& src,
                      Image<OutputStorageType>& gWE,
                      Image<OutputStorageType>& gNS)
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "GradientSobel3x3::Sobel3x3GreyValidInt_\n");
  InputStorageType *im=src.GetImageData();
  OutputStorageType *gx=gWE.GetImageData();
  OutputStorageType *gy=gNS.GetImageData();
  const int width=(int)src.GetWidth();
  const int nwidth = - width;
  const int height=(int)src.GetHeight();

#ifdef BIAS_DEBUG
  BIASASSERT(src.GetChannelCount()==1);
  BIASASSERT(src.SamePixelAndChannelCount(gWE));
  BIASASSERT(src.SamePixelAndChannelCount(gNS));
  BIASASSERT(gx!=gy);
  BIASASSERT(!_Conv.CalculationInFloat());
#endif

  if (width>this->_InternalMemWidth || height>this->_InternalMemHeight){
    this->_DeleteInternalMem();
    this->_AllocInternalMem(width, height);
  }

  int minx, miny, maxx, maxy;
  src.GetROI()->GetCorners(minx, miny, maxx, maxy);
  gWE.GetROI()->SetCorners(minx+1, miny+1, maxx-1, maxy-1);
  gNS.GetROI()->SetCorners(minx+1, miny+1, maxx-1, maxy-1);

  InputStorageType *end, *p, *lend;
  CONV_INT *ph, *pv;
  int offset;
  OutputStorageType *pgx, *pgy, *end2, *lend2;
  CONV_INT tmp, tmpx, tmpy;
  int roiw=maxx-minx-1;
  int step=width-maxx+minx;

  // separated sobel 3x3 for gx and gy
  
  // first line
  offset=minx+miny*width;
  p=im+1+offset;
  ph=this->_iwork[0]+1+offset;
  pv=this->_iwork2[0]+offset+width;
  lend=p+roiw-1;
  while (p<lend){
    tmp = (CONV_INT(*p))<<1;
    *ph++ = CONV_INT(p[-1]) + tmp + CONV_INT(p[1]);
    p++;
  }
  p+=step+1;
  ph+=step+1;
  lend+=width;
  end=im+maxx-1+(maxy-2)*width;
  while (p<end){
    tmp = (CONV_INT(*p))<<1;
    *pv++ = CONV_INT(p[nwidth]) + tmp + CONV_INT(p[width]);
    ph++;
    p++;
    while (p<lend){
      tmp = (CONV_INT(*p))<<1;
      *ph++ = CONV_INT(p[-1]) + tmp + CONV_INT(p[1]);
      *pv++ = CONV_INT(p[nwidth]) + tmp + CONV_INT(p[width]);
      p++;
    }
    tmp = (CONV_INT(*p))<<1;
    *pv++ = CONV_INT(p[nwidth]) + tmp + CONV_INT(p[width]);
    ph++;
    p++;
    lend+=width;
    p+=step;
    ph+=step;
    pv+=step;
  }
  p++; ph++; 
  // last line
  while (p<lend){
    tmp = (CONV_INT(*p))<<1;
    *ph++ = CONV_INT(p[-1]) + tmp + CONV_INT(p[1]);
    p++;
  }

  // now second run from temporary to gradients

  offset=minx+1+(miny+1)*width;
  ph=this->_iwork[0]+offset;
  pv=this->_iwork2[0]+offset;
  pgx=gx+offset;
  pgy=gy+offset;
  step+=2;
  end2=gx+maxx+width*(maxy-2)-1;
  lend2=gx+maxx+width*(miny+1)-1;
  while (pgx<end2){
    while (pgx<lend2){
      tmpx = (pv[1] - pv[-1])>>3;  // /8.0;
      tmpy = (ph[width] - ph[nwidth])>>3; // /8.0;
      *pgx++ = (OutputStorageType)tmpx;
      *pgy++ = (OutputStorageType)tmpy;
      pv++;
      ph++;
    }
    pv+=step;
    ph+=step;
    pgx+=step;
    pgy+=step;
    lend2+=width;
  }

  return 0;
}

template <class InputStorageType, class OutputStorageType>
void GradientSobel3x3<InputStorageType, OutputStorageType>::
_DeleteInternalMem()
{
  if (this->_fwork!=NULL) { 
    delete[] this->_fwork[0]; this->_fwork[0]=NULL;
    delete[] this->_fwork; this->_fwork=NULL; 
  }
  if (_fwork2!=NULL) { 
    delete[] _fwork2[0]; _fwork2[0]=NULL;
    delete[] _fwork2; _fwork2=NULL; 
  }
  if (this->_iwork!=NULL) { 
    delete[] this->_iwork[0]; this->_iwork[0]=NULL;
    delete[] this->_iwork; this->_iwork=NULL; 
  }
  if (_iwork2!=NULL) { 
    delete[] _iwork2[0]; _iwork2[0]=NULL;
    delete[] _iwork2; _iwork2=NULL; 
  }
  this->_InternalMemWidth=this->_InternalMemHeight=0;
}

template <class InputStorageType, class OutputStorageType>
void GradientSobel3x3<InputStorageType, OutputStorageType>::
_AllocInternalMem(const int width, const int height)
{
#ifdef BIAS_DEBUG
  if ((this->_fwork!=NULL) || (_fwork2!=NULL) || 
      (this->_iwork!=NULL) || (_iwork2!=NULL)){
    BIASERR("memory leak, reallocating without deleting!");
    BIASASSERT(false);
  } 
#endif
  int size=width*height;

  if (_Conv.CalculationInFloat()){
    this->_fwork = new CONV_FLOAT*[height];
    this->_fwork[0] = new CONV_FLOAT[size];
    _fwork2 = new CONV_FLOAT*[height];
    _fwork2[0] = new CONV_FLOAT[size];
    for (int i=1; i<height; i++) {
      this->_fwork[i]=this->_fwork[i-1]+width;
      _fwork2[i]=_fwork2[i-1]+width;
    }
  } else {
    this->_iwork = new CONV_INT*[height];
    this->_iwork[0] = new CONV_INT[size];
    _iwork2 = new CONV_INT*[height];
    _iwork2[0] = new CONV_INT[size];
    for (int i=1; i<height; i++) {
      this->_iwork[i]=this->_iwork[i-1]+width;
      _iwork2[i]=_iwork2[i-1]+width;
    }
  }

  this->_InternalMemWidth=width;
  this->_InternalMemHeight=height;
}
