/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __Label_HH__
#define __Label_HH__

#include <new>

#include <Base/Debug/Debug.hh>
#include <Base/Image/Image.hh>
#include <Base/Math/Vector2.hh>
#include "./FilterBase.hh"


#define LABEL_STORAGE_TYPE ImageBase::ST_int
#define LABEL_CALC_TYPE int

typedef struct _FilterBorders{
  int FirstLeft;
  int LastRight;
  int FirstUpper;
  int LastBottom;
 } FilterBorders;

#define D_INIT        0x0001
#define D_LABEL        0x0002

namespace BIAS
{
  // for sorting SortVec_
  class CCAStruct {
  public:
    CCAStruct() {};
    CCAStruct(const CCAStruct& s){
      bc=s.bc; tl=s.tl; br=s.br; }
    int bc;
    Vector2<int> tl, br;
    CCAStruct& operator=(const CCAStruct& s) {
      bc=s.bc; tl=s.tl; br=s.br; return *this; }
    bool operator<(const CCAStruct &right) const {
      return bc>right.bc; };
    bool operator==(const CCAStruct &right) const {
      return bc==right.bc; };
  };

  /** @class Label
      @ingroup g_filter
      @brief does a connected component analysis and labels all connected 
      regions
      
      there is no difference between fore and background. Each region that
      has a different color than all its neighbouring pixels is its own region.
      Each pixel is assigned to one region.
      
      There are utility functions to get Blobcounts and bounding boxes.
      So far only 4neighborhoud and grey value images only.
      Important: When using Filter(), the resulting image is only valid
      if there are less than 255 regions!!
      @author woelk
  */
  class Label : public BIAS::FilterBase<unsigned char, unsigned char>
  {
  public:
    /// standard constructor
    Label();

    /// standard destructor
    ~Label();
       
    /** call this if you don't need the output image, is faster than Filter
     */
    int Process(const Image<unsigned char>& im);

    virtual int Filter(const Image<unsigned char>& src, 
                       Image<unsigned char>& dst);

    /** labels an unsigned short int using an 4er neighbourhood and
        returns it in DestinationImage
        if DestinationImage is allocated it is not reallocated 
        @status untested 
        @author Felix Woelk
    */
    int Label4Neighbour(const BIAS::Image<unsigned char>& Image, 
                        BIAS::Image<unsigned char>& DestinationImage);



    /** labels an unsigned short int using an 4er neighbourhood and
        returns it in DestinationImage
        if DestinationImage is allocated it is not reallocated */
    int Label4Neighbour(const BIAS::Image<LABEL_CALC_TYPE>& Image, 
                        BIAS::Image<LABEL_CALC_TYPE>& DestinationImage);    

    /** returns a list with the size of each blob, largest first,
        if no blobs are found a list with a single 0 is returned      
        @author Daniel Grest, Aug. 2004
    */
    inline std::vector<int> GetBlobCounts()
    { return BlobCount_; }

    /** Retruns a map thta hold the blob size for each label. */
    inline std::map<unsigned char, int> GetBlobMap()
    { return BlobMap_; }

    /** returns in li the image where each region has a different grey value,
        value zero is the background region
     */
    inline void GetLabelImage(Image<LABEL_CALC_TYPE> &li)
    { li=LabelResultImage_; }

    /** returns two lists,one with the topLeft coordinates of the 
        bounding boxes of each region and one with the botrights, 
        sorted by size of regions  (in pixel)
        @return 0, in case of no error
        @author woelk
    */
    inline
    int GetBoundingBoxes(std::vector<Vector2<int> >& topLefts,
                         std::vector<Vector2<int> >& botRights)
    { topLefts = TopLefts_; botRights = BotRights_; return 0; };


  protected:

    // privates for Label filter 
    bool LabelInitialized_;
    Image<LABEL_CALC_TYPE> LabelSumImage_;
    Image<LABEL_CALC_TYPE> LabelResultImage_;
    void LabelInit_(const Image<unsigned char>& Image);
    void LabelInit_(const Image<LABEL_CALC_TYPE>& Image);
    void LabelRelease_();
    int Label4Neighbour_();
    // for CCA funxtionality
    std::vector<int> Num_;
     
    std::vector<CCAStruct> SortVec_;
    std::vector<int> BlobCount_;
    std::map<unsigned char,int> BlobMap_;
    std::vector<Vector2<int> > TopLefts_, BotRights_;
  
  };

} // namespace

#endif // __FILTER_HH__
