/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __HMatrix_hh__
#define __HMatrix_hh__
#include "bias_config.h" 

#include <Base/Geometry/HMatrixBase.hh>
#include <Base/Image/Image.hh>


namespace BIAS{

#define DHMATRIX_COMPUTEMLESAC 1 << 0;

  /** @class HMatrix
      @ingroup g_geometry
      @brief a 3x3 Matrix describing transformations in images
  */
  class HMatrix : public HMatrixBase
  {
  public:
    HMatrix();

    HMatrix(Matrix3x3<double> Mat);
    
    ~HMatrix();

    inline HMatrix GetHinv() const
    { return _Hinv; };



    HMatrix& operator=(const HMatrix& mat);

    /** calculatinges H using the basic linear transformation algorithm
        (see Hartley, Zisserman "Multiple View Geometry", p.73)
        p'=Hp -> (homg. coo) [p']x Hp = 0 
        3 row linear system, but rank 2 -> take only first two rows
        p=1...numpoints -> 2*numpoints equations
        solve A*h=0 with SVD
        @author Felix Woelk */
    int Compute(unsigned int numpoints, HomgPoint2D frompoint[], 
                HomgPoint2D topoints[]);

    /** assumes w=1 and calls the above */
    int Compute(unsigned int numpoints, Vector2<double> frompoint[], 
                Vector2<double> topoints[]);
    /** same as above for std::vectors. 
        @author Jan-Michael Frahm 
    */
    int Compute(std::vector<BIAS::HomgPoint2D>& frompoints, 
                std::vector<BIAS::HomgPoint2D>& topoints);

    /** maps source to dest using this 
        backward mapping is used 
        input: source, this(H)
        output: dest
        @deprecated will be removed, use BIAS/Image/HomographyMapping */
    int Map(Image<unsigned char> &source, Image<unsigned char> &dest);

    /** maps source to dest using this 
        backward mapping is used 
        the image coordinate system of dest is translated according to origin
        input: source, this(H)
        output: dest, origin 
        @deprecated will be removed, use BIAS/Image/HomographyMapping
    */
    int Map(Image<unsigned char> &source, Image<unsigned char> &dest, 
            int origin[2]);

    /** maps the rectangle (0,0),(OrigWidth,OrigHeight) with this H
        and computes mapped range as corner + size. JW 07/2003 */
    void GetMappedRange(const unsigned int OrigWidth, 
                        const unsigned int OrigHeight,
                        BIAS::Vector2<double> & corner,
                        double & sizeWidth, double & sizeHeight) const;

    /** @return the squared distance between the mapped point from
        the first picture and the measured point in the second picture
        @author mdunad 02 2004 */
    double GetMappedError(const BIAS::HomgPoint2D &PointPicOne,
                          const BIAS::HomgPoint2D &PointPicTwo) const;

    /** @brief returns jacobian of H: R^2 --> R^2
        
        H as a mapping of 2D to 2D is not always a linear mapping, because of
        the nonlinear homogenization. This function returns the 2x2 Jacobian
        matrix at the point x. You have to pass homogenized points and a
        correctly sized jacobian

        @param x homogenized point (real point in R^2) for the jacobian
        @param Jac (output) 2x2 jacobian at x
        @author koeser 04/2005 */
    void GetJacobian(const HomgPoint2D& x, Matrix<double>& Jac) const;

  protected:
    HMatrixBase _Hinv;

  };


}// namespace

#endif // __HMatrix_hh__
