/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "Triangulation.hh"
#include <Base/Geometry/RMatrixBase.hh>
#include <Base/Math/Random.hh>
#include <MathAlgo/PolynomialSolve.hh>

#include <cmath>

using namespace BIAS;
using namespace std;


Triangulation::Triangulation() : Debug() {
  _UpdateCovWeights();
}


Triangulation::~Triangulation()
{}

int Triangulation::Optimal(PMatrix &P1, PMatrix &P2, HomgPoint2D& p1, 
                           HomgPoint2D& p2, HomgPoint3D& point3d)
{
  FMatrix F(P1, P2);
  
  int res=CorrectCorrespondences(F, p1, p2);
  if (res==0){
    Vector3<double> r=F*p2;
    BIASCDOUT(D_TR_OPT, "p1: "<<p1<<"tp2: "<<p2<<endl);
    BIASCDOUT(D_TR_OPT, "p1' * F * p2 = "<<r.ScalarProduct(p1)<<endl);
    res=TriangulateLinear(P1, P2, p1, p2, point3d);
  } else {
    BIASERR("CorrectCorrespondences failed "<<res);
  }
  return res;
}

int 
Triangulation::TriangulateLinear(PMatrix &P1, PMatrix &P2,
                                 HomgPoint2D& p1, HomgPoint2D& p2,
                                 HomgPoint3D& point3d) {

  // set up DLT according to Zisserman: "Multiple View Geometry", p.297
  Matrix4x4<double> A;
  
  A[0][0] = P1[0][0] * p1[2] - P1[2][0] * p1[0];
  A[0][1] = P1[0][1] * p1[2] - P1[2][1] * p1[0];
  A[0][2] = P1[0][2] * p1[2] - P1[2][2] * p1[0];
  A[0][3] = P1[0][3] * p1[2] - P1[2][3] * p1[0];
    
  A[1][0] = P1[1][0] * p1[2] - P1[2][0] * p1[1];
  A[1][1] = P1[1][1] * p1[2] - P1[2][1] * p1[1];
  A[1][2] = P1[1][2] * p1[2] - P1[2][2] * p1[1];
  A[1][3] = P1[1][3] * p1[2] - P1[2][3] * p1[1];

  A[2][0] = P2[0][0] * p2[2] - P2[2][0] * p2[0];
  A[2][1] = P2[0][1] * p2[2] - P2[2][1] * p2[0];
  A[2][2] = P2[0][2] * p2[2] - P2[2][2] * p2[0];
  A[2][3] = P2[0][3] * p2[2] - P2[2][3] * p2[0];
    
  A[3][0] = P2[1][0] * p2[2] - P2[2][0] * p2[1];
  A[3][1] = P2[1][1] * p2[2] - P2[2][1] * p2[1];
  A[3][2] = P2[1][2] * p2[2] - P2[2][2] * p2[1];
  A[3][3] = P2[1][3] * p2[2] - P2[2][3] * p2[1];

  
  A.NormalizeRows();
  
  SVD svd(A);
  BIAS::Matrix<double> VT = svd.GetVT();
  point3d = (BIAS::Vector4<double>) VT.GetRow(3);
  point3d.Homogenize();
  return 0;
}

int Triangulation::CorrectCorrespondences(const FMatrix& F, HomgPoint2D& p1, 
                                          HomgPoint2D& p2, bool refine) 
{
#ifdef BIAS_DEBUG  
  BIASCDOUT(D_TR_OPT_F, "p1: "<<p1<<"\tp2: "<<p2<<endl);
  BIASCDOUT(D_TR_OPT_F, "before: p2^T * F * p1 = "<<(F*p1).ScalarProduct(p2)
            <<endl);
#endif

  p1.Homogenize();
  p2.Homogenize();
  Matrix3x3<double> T, Tp, Ti, Tpi;
  T.SetIdentity();
  Tp.SetIdentity();
  Ti.SetIdentity();
  Tpi.SetIdentity();
  T[0][2]=-p1[0];
  T[1][2]=-p1[1];
  Tp[0][2]=-p2[0];
  Tp[1][2]=-p2[1];
  Ti[0][2]=p1[0];
  Ti[1][2]=p1[1];
  Tpi[0][2]=p2[0];
  Tpi[1][2]=p2[1];
  
  BIASCDOUT(D_TR_OPT_TRANSF, "T: "<<T<<"\nTp: "<<Tp<<"\nTi: "<<Ti
            <<"\nTpi: "<<Tpi<<endl);

  FMatrix mF;
  mF=Tpi.Transpose()*F*Ti;
  BIASCDOUT(D_TR_OPT_TRANSF, "orig F: "<<F<<"\nTpi.Transpose() * F * Ti = "
            <<mF<<endl);

  HomgPoint2D e, ep;
  F.GetEpipoles(e, ep);
  ep.Homogenize(); e.Homogenize();
  BIASCDOUT(D_TR_OPT_TRANSF, "orig epipoles: "<<e<<"\t"<<ep<<endl);

  mF.GetEpipoles(e, ep);
  BIASCDOUT(D_TR_OPT_TRANSF, "epipoles: "<<e<<"\t"<<ep<<endl);
  
#ifdef BIAS_DEBUG
  Vector3<double> r;
  mF.TransposedMult(ep, r);
  BIASCDOUT(D_TR_OPT_TRANSF, "F*e = "<<mF*e<<"\tep^T*F = "<<r<<endl);

  // BIASASSERT((mF*e).NormL2()<=1e-10);
  // BIASASSERT(r.NormL2()<=1e-10);
#endif

  double se, sep;
  se=sqrt(e[0]*e[0]+e[1]*e[1]);
  sep=sqrt(ep[0]*ep[0]+ep[1]*ep[1]);
  
  e/=se;
  ep/=sep;

  BIASCDOUT(D_TR_OPT_TRANSF, "normalized epipoles: "<<e<<"\t"<<ep<<endl);

  RMatrixBase R, Rp;
  R.SetIdentity();
  Rp.SetIdentity();

  R[0][0]=R[1][1]=e[0];
  R[0][1]=e[1];
  R[1][0]=-e[1];

  Rp[0][0]=Rp[1][1]=ep[0];
  Rp[0][1]=ep[1];
  Rp[1][0]=-ep[1];

  BIASCDOUT(D_TR_OPT_TRANSF, "R*e = "<<R*e<<"\tRp*ep = "<<Rp * ep<<endl);

  mF = Rp * mF * R.Transpose();

  BIASCDOUT(D_TR_OPT_TRANSF, "Rp * Tpi.Transpose() * F * Ti * R.Transpose() = "
            <<mF<<endl);

#ifdef BIAS_DEBUG  
  Vector3<double> r4(0.0, 0.0, 1.0);
  BIASCDOUT(D_TR_OPT_F, "before: (0, 0, 1) * mF * (0, 0, 1)^T = "
            <<(mF*r4).ScalarProduct(r4)<<endl);
  BIASCDOUT(D_TR_OPT_TRANSF, "Ti * R.Transpose() * (0,0,1): "
            <<Ti * R.Transpose() * r4
            <<"\t(Ti * R.Transpose() * (0,0,1))-p1 ="
            <<(Ti * R.Transpose() * r4)-p1<<endl);
  BIASCDOUT(D_TR_OPT_TRANSF, "Tpi * Rp.Transpose() * (0,0,1): "
            <<Tpi * Rp.Transpose() * r4<<"\t(Tpi * Rp.Transpose() * r4)-p2 = "
            <<(Tpi * Rp.Transpose() * r4)-p2<<endl);
#endif
  
  double f, fp, a, b, c, d;
  f=e[2];
  fp=ep[2];
  a=mF[1][1];
  b=mF[1][2];
  c=mF[2][1];
  d=mF[2][2];
  BIASCDOUT(D_TR_OPT, "ffpd: "<<f*fp*d<<"\t-fpc: "<<-fp*c<<"\t-fpd: "<<-fp*d
            <<"\n-fb: "<<-f*b<<"\ta: "<<a<<"\tb: "<<b
            <<"\n-fd: "<<-f*d<<"\tc: "<<c<<"\td: "<<d<<endl<<endl);
  
  BIASCDOUT(D_TR_OPT, "a: "<<a<<"\tb: "<<b<<"\tc: "<<c<<"\td: "<<d<<"\tf: "<<f
            <<"\tfp: "<<fp<<endl);
  /** see mupad
      assume(a, Type::Real):
      assume(b, Type::Real):
      assume(c, Type::Real):
      assume(d, Type::Real):
      assume(f, Type::Real):
      assume(mfp, Type::Real):
      assume(t, Type::Real):
      
      g:=t*((a*t+b)^2+mfp^2*(c*t+d)^2)^2-(a*d-b*c)*(1+f^2*t^2)^2*(a*t+b)*(c*t+d):

      g:=simplify(expand(g)):

      PRETTYPRINT := FALSE:

      res:=poly(collect(g, [t]), [t]);

      fprint(Unquoted, 0, "coeff[0]=", coeff(res, t, 0), "; // t^0"):
      fprint(Unquoted, 0, "coeff[1]=", coeff(res, t, 1), "; // t^1"):
      fprint(Unquoted, 0, "coeff[2]=", coeff(res, t, 2), "; // t^2"):
      fprint(Unquoted, 0, "coeff[3]=", coeff(res, t, 3), "; // t^3"):
      fprint(Unquoted, 0, "coeff[4]=", coeff(res, t, 4), "; // t^4"):
      fprint(Unquoted, 0, "coeff[5]=", coeff(res, t, 5), "; // t^5"):
      fprint(Unquoted, 0, "coeff[6]=", coeff(res, t, 6), "; // t^6"):
  */
  vector<POLYNOMIALSOLVE_TYPE> coeff(7);
  vector<complex<POLYNOMIALSOLVE_TYPE> > csol;
  double a2=a*a, b2=b*b, c2=c*c, d2=d*d, f2=f*f, fp2=fp*fp;
  double a3=a2*a, b3=b2*b, c3=c2*c, d3=d2*d;
  double a4=a2*a2, b4=b2*b2, c4=c2*c2, d4=d2*d2, f4=f2*f2, fp4=fp2*fp2;
  double ab=a*b, cd=c*d;
  coeff[0]=b2*cd - ab*d2; // t^0
  coeff[1]=b4 - a2*d2 + b2*c2 + d4*fp4 + 2*b2*d2*fp2; // t^1
  coeff[2]=4*a*b3 + ab*c2 - a2*cd - 2*ab*d2*f2 + 2*b2*cd*f2 + 4*c*d3*fp4 + 
    4*ab*d2*fp2 + 4*b2*cd*fp2; // t^2
  coeff[3]=8*ab*cd*fp2 + 6*a2*b2 - 2*a2*d2*f2 + 2*b2*c2*f2 + 2*a2*d2*fp2 + 
    2*b2*c2*fp2 + 6*c2*d2*fp4; // t^3
  coeff[4]=4*a3*b + 2*ab*c2*f2 - 2*a2*cd*f2 - ab*d2*f4 + b2*cd*f4 + 
    4*c3*d*fp4 + 4*ab*c2*fp2 + 4*a2*cd*fp2; // t^4
  coeff[5]=a4 + c4*fp4 - a2*d2*f4 + b2*c2*f4 + 2*a2*c2*fp2; // t^5
  coeff[6]=ab*c2*f4 - a2*cd*f4; // t^6

  PolynomialSolve ps;
  if(DebugLevelIsSet(D_TR_OPT))
    ps.AddDebugLevel(D_PS_GSL);
  // crop upper zero coeffizients 
  ps.CheckCoefficients(coeff);
  //  while ((*coeff.rbegin())==0.0) { coeff.pop_back(); }
  int psres=ps.Numeric(coeff, csol);
  BIASCDOUT(D_TR_OPT, "csol.size(): "<<csol.size()<<"\tpsres: "<<psres<<endl);
  if (psres<0){
    BIASERR("error solving polynomial "<<psres);
    abort();
    return psres;
  }
  if (refine)
    ps.NonLinearRefine(coeff, csol);

  // vector<POLYNOMIALSOLVE_TYPE> sol;
  // ps.Solve(coeff, sol);
  // BIASCDOUT(D_TR_OPT, "sol.size(): "<<sol.size()<<endl);
  
  vector<double> s(csol.size()+1);
  
  for (unsigned i=0; i<csol.size(); i++){
    double sol2=real(csol[i])*real(csol[i]);
    double t1=1.0+f2*sol2;
    double t2=c*real(csol[i])+d;
    double t3=a*real(csol[i])+b;
    double t4=c*real(csol[i])+d;
    
    s[i]=sol2/(t1*t1)+(t2*t2)/(t3*t3+fp2*t4*t4);
    BIASCDOUT(D_TR_OPT, "dist["<<i<<"] = "<<s[i]<<" = "<<sol2/(t1*t1)<<" + "
              <<(t2*t2)/(t3*t3+fp2*t4*t4)<<endl);
  }
  s[csol.size()]=1.0/f2+c2/(a2+fp2*c2);
  BIASCDOUT(D_TR_OPT, "dist["<<csol.size()<<"] = "<<s[csol.size()]
            <<" = "<<1.0/f2<<" + "<<c2/(a2+fp2*c2)<<endl);
 
  double costmin=s[0], tmin=real(csol[0]);
  int minres=0;
 

  for (unsigned i=1; i<s.size(); i++){
    if (s[i]<costmin) { costmin=s[i]; minres=i; tmin=real(csol[i]); } 
  }
  BIASCDOUT(D_TR_OPT, "costmin = "<<costmin<<"\tat "<<minres<<"\ttmin: "
            <<tmin<<endl);
  BIASCDOUT(D_TR_OPT, "poli("<<tmin<<") = "<<coeff[0] + tmin*coeff[1] + 
            tmin*tmin*coeff[2]+tmin*tmin*tmin*coeff[3]+
            tmin*tmin*tmin*tmin*coeff[4]+tmin*tmin*tmin*tmin*tmin*coeff[5]+
            tmin*tmin*tmin*tmin*tmin*tmin*coeff[6]<<endl);
  BIASCDOUT(D_TR_OPT, "ps.EvaluatePolynomial("<<tmin<<", coeff) = "
            <<ps.EvaluatePolynomial(tmin, coeff)<<endl);
  

  Vector3<double> l, lp;
  l.Set(tmin*f, 1.0, -tmin);
  lp.Set(-fp*(c*tmin+d), a*tmin+b, c*tmin+d);
  //   BIASCDOUT(D_TR_OPT, "l: "<<l<<"\tlp: "<<lp<<endl);

  HomgPoint2D xh, xhp;
  xh.Set(-l[0]*l[2], -l[1]*l[2], l[0]*l[0]+l[1]*l[1]);
  xhp.Set(-lp[0]*lp[2], -lp[1]*lp[2], lp[0]*lp[0]+lp[1]*lp[1]);
  //BIASCDOUT(D_TR_OPT, "xh: "<<xh<<"\txhp: "<<xhp<<endl);

#ifdef BIAS_DEBUG  
  //BIASCDOUT(D_TR_OPT, "xh: "<<xh<<"\txhp: "<<xhp<<endl);
  BIASCDOUT(D_TR_OPT_F, "optimized: xhp * mF * xh = "
            <<(mF*xh).ScalarProduct(xhp)<<endl);
#endif

  p1 = Ti * R.Transpose() * xh; 
  p2 = Tpi * Rp.Transpose() * xhp;
  p1.Homogenize();
  p2.Homogenize();
  BIASCDOUT(D_TR_OPT, "optimized points "<<p1<<" <--> "<<p2<<endl);

#ifdef BIAS_DEBUG  
  BIASCDOUT(D_TR_OPT_F, "optimized: p2^T * F * p1 = "<<(F*p1).ScalarProduct(p2)
            <<endl);
#endif
    
  return 0;
}


int 
Triangulation::Triangulate(PMatrix &P1, PMatrix &P2,
                           HomgPoint2D& p1, HomgPoint2D& p2,
                           HomgPoint3D& point3d)
{
  int res = 0; 
  Vector3<double> normray1, center1, normray2, center2;
 
  Matrix3x3<double> P1I, P2I;
  P1.GetHinf(P1I);
  P2.GetHinf(P2I);
  normray1 = P1I * p1;
  normray1 /= normray1.norm_L2();
  normray2 = P2I * p2;
  normray2 /= normray2.norm_L2();
 
//   P1.GetC(center1);
//   P2.GetC(center2);
  HomgPoint3D c1, c2;
  P1.GetNullVector(c1);
  P2.GetNullVector(c2);
  c1.Homogenize();
  c2.Homogenize();
  center1.Set(c1[0], c1[1], c1[2]);
  center2.Set(c2[0], c2[1], c2[2]);

  Vector3<double> vecX;
  res = Triangulation::Intersect(center1, normray1, center2, normray2, vecX);
  point3d[0] = vecX[0];
  point3d[1] = vecX[1];   
  point3d[2] = vecX[2];

  // this is guaranteed by the interface:
  point3d[3] = 1.0;

  return res;
}

int 
Triangulation::Triangulate2D(PMatrix& P1, PMatrix& P2, HomgPoint2D& p1, 
                             HomgPoint2D& p2, HomgPoint3D& point3d, 
                             double& dist, double& angle)
{
  int res = 0;
  BIASDOUT(D_TR_ARG, "P1: "<<P1<<"\nP2: "<<P2<<"\np1: "<<p1<<"\tp2: "<<p2);
  Vector3<double> ray1, ray2;
  double gamma, tau;
  double s1t,s2t;
  HomgPoint3D p3d1, p3d2;
  Vector3<double> T;
  //cerr << "C1: "<< P1.GetC()<< "\tC2: "<< P2.GetC() << endl;
  //cerr << "p1: "<< p1 << "\tp2: "<< p2 << endl;
  Vector3<double> C1, C2; 
  if ((P1.GetC(C1) != 0) || (P2.GetC(C2) != 0)){
    BIASERR("error getting camera center "<<P1<<" "<<C1<<"\n"<<P2<<" "<<C2);
    return -3;
  }
  T = C1 - C2; //camera base line
  ray1=P1.GetNormRayWorldCoo(p1);
  ray2=P2.GetNormRayWorldCoo(p2);
  BIASDOUT(D_TR_TR,"center1: "<<C1<<"\t    ray1: "<<ray1);
  BIASDOUT(D_TR_TR,"center2: "<<C2<<"\t    ray2: "<<ray2);

  // compute different scalar products  
  ray1.ScalarProduct(ray2, angle);
  ray1.ScalarProduct(T, s1t);
  ray2.ScalarProduct(T, s2t); 

//   cerr << "( " << setprecision(14) << angle * s2t - s1t <<" ) / ( "
//        << setprecision(14) << 1 - angle * angle << " )"<<endl;
  tau = (angle * s2t - s1t)/(1 - angle * angle);
  gamma = tau * angle + s2t;
  BIASDOUT(D_TR_TR,"T: "<<T<<"\tang: "<<angle<<"\ts1t: "<<s1t<<"\ts2t: "<<s2t
           <<"\ttau: "<<tau<<"\tgamma: "<<gamma);

//   cerr <<"ray1: "<< ray1 << "\tray2: "<<ray2<<"\tT: "<<T<<endl;
//   cerr <<"angle: "<<angle<<"\ts1t: "<<s1t<<"\ts2t: "<<s2t
//        <<"\ttau: "<<tau<<"\tgamma: "<<gamma<<"\t"<<(ray1-ray2).NormL2()<<endl;

  if ( tau < 0 || gamma < 0){
    // no valid point because intersection point is behind
    // one camera or behind both cameras
    BIASDOUT(D_TR_POS,"3d point lies behind one or both cameras"); 
    point3d.Set(0.0, 0.0, 0.0);
    res = -1;
    BIASCDOUT(D_TR_INF, "B"); BIASFLUSHDOUT;
  } else if ((ray1-ray2).norm_L2()<PARALLEL_THRESHOLD) {
    // no valid point because rays do not intersect
    BIASDOUT(D_TR_POS,"no triangulation possible, rays are parallel"); 
    point3d.Set(0.0, 0.0, 0.0);
    res = -2;
    BIASCDOUT(D_TR_INF, "i"); BIASFLUSHDOUT;
  } else {// maybe a correct point
    p3d1=C1 + ray1 * tau;
    p3d2=C2 + ray2 * gamma;
    dist=(p3d1 - p3d2).NormL2();
    BIASDOUT(D_TR_TR,"p3d1: "<<p3d1<<"\tp3d2: "<<p3d2);
    //cerr << "dist: " << dist << endl;
    point3d=(p3d1+p3d2)/2.0;
    //cerr <<"p3d: "<<(p3d1+p3d2)/2.0<<endl;
//     point3d[0] = (P1.GetC()[0] + P2.GetC()[0] + tau * ray1[0] +
// 		  gamma * ray2[0]) / 2.0;
//     point3d[1] = (P1.GetC()[1] + P2.GetC()[1] + tau * ray1[1] +
// 		  gamma * ray2[1]) / 2.0;                
//     point3d[2] = (P1.GetC()[2] + P2.GetC()[2] + tau * ray1[2] +
// 		  gamma * ray2[2]) / 2.0;
//     point3d[3] = 1.0;
    BIASCDOUT(D_TR_INF, "."); BIASFLUSHDOUT;
  }
  return res;
}

int 
Triangulation::Triangulate2D(PMatrix& P1, PMatrix& P2, HomgPoint2D& p1, 
                             HomgPoint2D& p2, HomgPoint3D& point3D,
                             Matrix3x3<double>& CovMatrix, double& dist,
                             double& angle, const double& triangdeltapixel)
{
  int res=0;
  if ((res=Triangulate2D(P1, P2, p1, p2, point3D, dist, angle))==0)
    res=GetCovariance2D(P1, P2, p1, p2, point3D, CovMatrix, triangdeltapixel);
  return res;
}




void Triangulation::_UpdateCovWeights(const double &TriangDeltaPixels,
                                      const double &sigma)
{
  //AddDebugLevel(D_TRIANG_COVWEIGHT);
  _dTriangDeltaPixels = TriangDeltaPixels;
  _dTriangSigmaPixels = sigma;

  BIASCDOUT(D_TRIANG_COVWEIGHT, "triangulating in window with "
            <<TriangDeltaPixels<<" using sigma "<<sigma<<endl);
  // initialize Gauss-weights  
  for (int y=-1; y <= 1; y++){
    for (int x=-1; x <= 1; x++)	 {
      // construct weight for deviation (0,1,2)
      cov_weights_[x+1][y+1] = 
        exp(-((((double)y*TriangDeltaPixels)*((double)y*TriangDeltaPixels) 
             +((double)x*TriangDeltaPixels)*((double)x*TriangDeltaPixels))/
             (2.0*sigma*sigma)));
      
      BIASCDOUT(D_TRIANG_COVWEIGHT, (2.0*sigma*sigma)<<"\t"
                <<(((double)y*TriangDeltaPixels)*
                       ((double)y*TriangDeltaPixels)+
                       ((double)x*TriangDeltaPixels)*
                       ((double)x*TriangDeltaPixels))<<"\t"
                <<exp(-((((double)y*TriangDeltaPixels)*
                             ((double)y*TriangDeltaPixels) 
                             +((double)x*TriangDeltaPixels)*
                             ((double)x*TriangDeltaPixels))/
                        (2.0*sigma*sigma)))<<endl);
      /*1.0/(TriangDeltaPixels*sqrt(2.0*M_PI)*sigma)*exp(-(abs(y)+ abs(x))/
          (2.0*sigma*sigma));*/
      BIASCDOUT(D_TRIANG_COVWEIGHT, "weight["<<x+1<<"]["<<y+1<<"] = "
                <<cov_weights_[x+1][y+1]<<endl);
    }  
  }
}


int 
Triangulation::GetCovariance2D(PMatrix& P1, PMatrix& P2, HomgPoint2D& p1,
                               HomgPoint2D& p2, HomgPoint3D &point3D,
                               Matrix3x3<double>& covariance,
                               const double &triangdeltapixel/*=1.0*/) 
{
#ifdef BIAS_TRIANG_DEBUG
  covpoints_.clear();
  covweights_.clear();
#endif
  int n=0, res;
  //cerr << "point3D: "<<point3D<<endl;
  p1.Homogenize();
  p2.Homogenize();

  if (_dTriangDeltaPixels != triangdeltapixel){
    _UpdateCovWeights(triangdeltapixel);
  }

  covariance.SetZero();
  
  // loop over all combinations x around p1[0] and p2[0]
  // create point cloud of image variance 1 pixel and estimate 
  // 3D covarianve from it  
  HomgPoint3D X; //used in inner loop
  double sumweight = 0;
  double dist, angle, wl, weight;

  HomgPoint2D p3, p4;

  point3D.Homogenize();
  // loop around left pixel
  for (int yl=-1; yl <= 1; yl++)
    for (int xl=-1; xl <= 1; xl++) {
      p3.Set(p1[0] + (double)xl*triangdeltapixel, 
             p1[1] + (double)yl*triangdeltapixel);  

      wl = cov_weights_[xl+1][yl+1];
        
      // loop around right pixel
      for (int yr=-1; yr <= 1; yr++)
        for (int xr=-1; xr <= 1; xr++) {
          // construct total weight
          weight =wl*cov_weights_[xr+1][yr+1];              

          p4.Set(p2[0] + (double)xr*triangdeltapixel, 
                 p2[1] + (double)yr*triangdeltapixel); 

          res = Triangulate2D(P1, P2, p3, p4, X, dist, angle);

          if (res == 0) {  
            X.Homogenize();
            n++;
#ifdef BIAS_TRIANG_DEBUG
            covpoints_.push_back(X);
            covweights_.push_back(weight);
#endif
            //cerr <<X<<" \t X / point3D "<< X[0] / point3D[0]<< " " <<
            // X[1] / point3D[1]<< " " << X[2] / point3D[2]<< " " <<dist<<endl;
            // now Outer product 
            for (unsigned int i=0;i<3;i++) {
              for (unsigned int j=0;j<3;j++) {                           
                covariance[i][j] +=  
                  ((X - point3D)[i]) * ((X - point3D)[j]) * weight ;
              }
            }
            sumweight += weight;
          } else {
            BIASCDOUT(D_TR_INF, "B"); BIASFLUSHDOUT;
          }    
        }
    } 
  // cerr << endl;
  // normalize Cov
  if (sumweight!=0){
    covariance = covariance*(1.0/sumweight);
    // cerr << num_weights<<" weights, sum="<<sumweight<<endl;
    //     for (int i=0; i<3; i++)
    //       cerr << i<<": " << covariance[i][i] << "  ";
    //     cerr <<endl;
    //cerr << "point3D: "<<point3D<<" calculated cov from " << n << " points " << endl;
    return 0;
  } else {
    return -4;
  }
}


#ifdef BIAS_TRIANG_DEBUG
void Triangulation::GetCovPoints(vector<HomgPoint3D>& pvec,
                                 vector<double>& weight)
{
  vector<HomgPoint3D>::iterator pit;
  vector<double>::iterator dit;
  for (dit=covweights_.begin(), pit=covpoints_.begin(); pit!=covpoints_.end();
       pit++, dit++){
    pvec.push_back((*pit));
    weight.push_back((*dit));
  };
}
#endif

int 
Triangulation::GetCovarianceAnalytic(PMatrix& P1, PMatrix& P2, HomgPoint2D& p1,
                                     HomgPoint2D& p2, HomgPoint3D &point3D,
                                     Matrix3x3<double>& covariance,
                                     const double &triangdeltapixel,
                                     const double &sigma) 
{
  // for performance reasons, we dont want to homogenize each time
  // therefore rely on homogenized points and check in debug mode
#ifdef BIAS_DEBUG
  if (!p1.IsHomogenized()) {
    BIASERR("Supplied unhomogenized point p1 for covariance computation !"<<
            " Fix this, since this is only checked in DEBUG mode. Aborting.");
    abort();
  }
  if (!p2.IsHomogenized()) {
    BIASERR("Supplied unhomogenized point p2 for covariance computation !"<<
            " Fix this, since this is only checked in DEBUG mode. Aborting.");
    abort();
  }
  if (!point3D.IsHomogenized()) {
    BIASERR("Supplied unhomogenized 3d point for covariance computation !"<<
            " Fix this, since this is only checked in DEBUG mode. Aborting.");
    abort();
  }
#endif

#ifdef BIAS_TRIANG_DEBUG
  covpoints_.clear();
  covweights_.clear();
#endif
    
  if ((_dTriangDeltaPixels != triangdeltapixel)
      || (_dTriangSigmaPixels != sigma)) {
    _UpdateCovWeights(triangdeltapixel,sigma);
  }

  covariance.SetZero();
  
  // loop over all combinations x around p1[0] and p2[0]
  // create point cloud of image variance 1 pixel and estimate 
  // 3D covarianve from it  
  HomgPoint3D X; //used in inner loop
  double sumweight = 0;
 
  // cout << "cov_weights_:"<<cov_weights_ << endl;
  HomgPoint2D p3, p4;

  // loop around left pixel
  for (int yl=-1; yl <= 1; yl++)
    for (int xl=-1; xl <= 1; xl++) {
      p3.Set(p1[0] + (double)xl*triangdeltapixel, 
             p1[1] + (double)yl*triangdeltapixel);  

      double wl = cov_weights_[xl+1][yl+1];
        
      // loop around right pixel
      for (int yr=-1; yr <= 1; yr++)
        for (int xr=-1; xr <= 1; xr++) {
                     

          p4.Set(p2[0] + (double)xr*triangdeltapixel, 
                 p2[1] + (double)yr*triangdeltapixel); 

            
          if (Triangulate(P1, P2, p3, p4, X)==0) {  
            // construct total weight
            register double weight = wl*cov_weights_[xr+1][yr+1];  
            
            // X.Homogenize(); // not necessary, since guaranteed
            // n++;
#ifdef BIAS_TRIANG_DEBUG
            covpoints_.push_back(X);
            covweights_.push_back(weight);
#endif
            //cerr <<"p3: "<<p3<<"\tp4: "<<p4<<"\tX "<<X<<"\tweight "<<weight<<endl;

            
            // optimized outer product implementation does this (I hope so):
            // for (unsigned int i=0;i<3;i++) 
            //   for (unsigned int j=0;j<3;j++) {                           
            //     covariance[i][j] +=  
            //       ((X - point3D)[i]) * ((X - point3D)[j]) * weight ;
            //   }
            register double DiffVec[3]; double tmp;
            // unrolled loop to compute diagonal elements:
            DiffVec[0]        = X[0]-point3D[0];
            covariance[0][0] += DiffVec[0] * DiffVec[0] * weight;
            DiffVec[1]        = X[1]-point3D[1];
            covariance[1][1] += DiffVec[1] * DiffVec[1] * weight;
            DiffVec[2]        = X[2]-point3D[2];
            covariance[2][2] += DiffVec[2] * DiffVec[2] * weight;

            // non-diag elements of symmetric matrix:
            covariance[0][1] += tmp = DiffVec[0] * DiffVec[1] * weight ;
            covariance[1][0] += tmp;           
            covariance[0][2] += tmp = DiffVec[0] * DiffVec[2] * weight ;
            covariance[2][0] += tmp;   
            covariance[2][1] += tmp = DiffVec[2] * DiffVec[1] * weight ;
            covariance[1][2] += tmp;   
                        
            sumweight += weight;              
          }    
        }
    } 
  // cerr << endl;
  // normalize Cov
  if (sumweight != 0){
    //cout << "covariance "<<covariance<<endl;
    covariance = covariance*(1.0/sumweight);
    //     for (int i=0; i<3; i++)
    //       cerr << i<<": " << covariance[i][i] << "  ";
    //     cerr <<endl;
    //cerr << "point3D: "<<point3D<<" calculated cov from " << n << " points " << endl;
    //cout << "covariance calculation succeeded ("<<sumweight<<")"<<covariance<<endl;
    return 0;
  } else {
    //cout << "covariance calculation failed"<<endl;
    return -1;
  }
}

Vector3<double> 
Triangulation::Intersect(Vector3<double> &a1, Vector3<double> &a2,
                         Vector3<double> &b1, Vector3<double> &b2)
{
  Vector3<double> dirA=a2-a1, dirB=b2-b1; // direction vectors
#ifdef BIAS_DEBUG
  if ( (dirA.NormL2()<=DBL_EPSILON) ||  (dirB.NormL2()<=DBL_EPSILON) )
    BIASERR("no valid intersection because of identical points");
#endif
  dirA.Normalize();
  dirB.Normalize();
  Vector3<double> res;
  Triangulation::Intersect(a1,dirA, b1,dirB, res);
  return res;
}

int 
Triangulation::GetCovariance(PMatrix& P1, PMatrix& P2, 
                             const HomgPoint2D& p1, 
                             const Matrix2x2<double>& cov1,
                             const HomgPoint2D& p2, 
                             const Matrix2x2<double>& cov2,
                             HomgPoint3D& p, Matrix3x3<double>& cov,
                             const int num) 
{
  int res=0;
  Random ran;

  BIASASSERT(!p1.IsAtInfinity());
  BIASASSERT(!p2.IsAtInfinity());
  BIASASSERT(p1.IsHomogenized());
  BIASASSERT(p2.IsHomogenized());

  Vector2<double> vp1(p1[0], p1[1]), vp2(p2[0], p2[1]);
  vector<Vector2<double> > ran1, ran2;

  if (ran.GetNormalDistributed(vp1, cov1, num, ran1)!=0){
    BIASERR("invalid cov1 "<<cov1<<endl);
    res=-1;
  }
  if (ran.GetNormalDistributed(vp2, cov2, num, ran2)!=0){
    BIASERR("invalid cov2 "<<cov2<<endl);
    res=-2;
  }

  /*
  cout << "input: "<<vp1<<"\n"<<cov1<<endl;
  Vector2<double> mr1(0.0, 0.0), mr2(0.0, 0.0);
  cout << "ran1: ";    
  for (int i=0; i<num; i++){
    cout << ran1[i]<<" "; mr1+=ran1[i]; } 
  cout << "  mean "<<mr1/(double)num<<endl;
  cout << "input: "<<vp2<<"\n"<<cov2<<endl;
  cout << "ran2: ";    
  for (int i=0; i<num; i++){
    cout << ran2[i]<<" "; mr2+=ran2[i]; }
  cout << "  mean "<<mr2/(double)num<<endl;
  */

  if (res==0){
    HomgPoint2D tp1, tp2;
    vector<HomgPoint3D> vec3d;
    vec3d.reserve(num*num);
    HomgPoint3D tp3d;

    p.Set(0.0, 0.0, 0.0, 0.0);

    int i, j;
    for (i=num-1; i>=0; i--){
      for (j=num-1; j>=0; j--){
        tp1.Set(ran1[i][0], ran1[i][1]);
        tp2.Set(ran2[j][0], ran2[j][1]);
      
        if (Triangulate(P1, P2, tp1, tp2, tp3d)==0) { 
          BIASASSERT(tp3d.IsHomogenized());
          // implicitly count number of successfull triangulation in 
          // w component of p
          p+=tp3d;
          vec3d.push_back(tp3d);
        } 
      }
    }
  
    int num_successfull = (int)rint(p[3]);
    BIASASSERT(num_successfull==(int)vec3d.size()); // just for debugging

    // get mean by dividing by w component of p. It contains the number of 
    // successfull triangulations
    p.Homogenize(); 

    // now calculate variance
    cov.SetZero();
    vector<HomgPoint3D>::iterator it;
    register double dx, dy, dz;
    for (it=vec3d.begin(); it!=vec3d.end(); it++){
      dx = p[0]-(*it)[0];
      dy = p[1]-(*it)[1];
      dz = p[2]-(*it)[2];
    
      cov[0][0] += dx*dx;
      cov[1][1] += dy*dy;
      cov[2][2] += dz*dz;
      cov[0][1] += dx*dy;
      cov[0][2] += dx*dz;
      cov[1][2] += dy*dz;
    }
    cov[1][0] = cov[0][1];
    cov[2][0] = cov[0][2];
    cov[2][1] = cov[1][2];
  
    if (num_successfull>1) {
      cov /= double(num_successfull-1);
    } 
    res = num_successfull;
  }

  return res;
}

int
Triangulation::Intersect(Vector3<double> &pA, Vector3<double> &dirA,
                         Vector3<double> &pB, Vector3<double> &dirB, 
                         Vector3<double> &res)
{
#ifdef BIAS_DEBUG
  if ( (dirA.NormL2()<=DBL_EPSILON) ||  (dirB.NormL2()<=DBL_EPSILON) ) {
    BIASERR("no valid intersection because direction vector has length zero");
    return -2;
  }
#endif 

  // get planes through origin, pA and dirA (also for B)
  Vector3<double> momA, momB, x;
  pA.CrossProduct(dirA,momA); //momA = a1 x dirA;
  pB.CrossProduct(dirB,momB); //momB = b1 x dirB;
  
  // check whether dirA and dirB conincide
  dirA.CrossProduct(dirB,res);
  register double dummy,lambda;
  dummy = res.NormL2();
  lambda = dummy * dummy;

  if (lambda<=DBL_EPSILON) {
#ifdef BIAS_DEBUG
    BIASERR("cant intersect, lines are parallel");  
#endif
    res[0]=0;res[1]=0;res[2]=0;
    return -1;
  }

  // old code, slow but possibly more readable:
  //   res= ( (res * momB) - 
  //           ( (dirA * dirB) * (x * momA)) ) 
  //           ) /lambda 
  //         ) * dirA + dirA.CrossProduct(momA);

  dummy = ( (res * momB) - ( (dirA * dirB) * (res * momA)) ) / lambda;
  dirA.Multiply(dummy,res);
  dirA.CrossProduct(momA,x);
  res += x;

  // res is the point with minimum distance on line A
  // now get the corresponding point on B:
  lambda = (res-pB)*dirB;
  x = pB + lambda*dirB;

  // set intersection as the point in the middle of x and res
  res = 0.5*(res+x);
 
  // if lambda<0 the point is in negative dirB, 
  // that is, for dirB=optical axis: point is behind camera B
  if (lambda <0.0 ) return 1;

  // check if point is behind camera A
  if (dirA.ScalarProduct(res-pA)<0.0) return 2;
 
  
  return 0;
}

int
Triangulation::Intersect(Vector3<double> &pA, Vector3<double> &dirA,
                         Vector3<double> &pB, Vector3<double> &dirB, 
                         Vector3<double> &res, double &dist)
{
#ifdef BIAS_DEBUG
  if ( (dirA.NormL2()<=DBL_EPSILON) ||  (dirB.NormL2()<=DBL_EPSILON) ) {
    BIASERR("no valid intersection because direction vector has length zero");
    return -2;
  }
#endif 
  Vector3<double> momA, momB, x;
  pA.CrossProduct(dirA,momA); //momA = a1 x dirA;
  pB.CrossProduct(dirB,momB); //momA = a1 x dirA;
  dirA.CrossProduct(dirB,res);
  register double dummy,lambda;
  dummy = res.NormL2();
  lambda = dummy * dummy;

  if (lambda<=DBL_EPSILON) {
#ifdef BIAS_DEBUG
    BIASERR("cant intersect, lines are parallel");  
#endif
    res[0]=0;res[1]=0;res[2]=0;
    return -1;
  }
  
  //   res= ( (res * momB) - 
  //           ( (dirA * dirB) * (x * momA)) ) 
  //           ) /lambda 
  //         ) * dirA + dirA.CrossProduct(momA); 
  dummy = ( (res * momB) - ( (dirA * dirB) * (res * momA)) )  / lambda;

  dirA.Multiply(dummy,res);
  dirA.CrossProduct(momA,x);
  res += x;


  // res is the point with minimum distance on line A
  // now get the correpsonding point on B:
  lambda=(res-pB)*dirB;
  x = pB + lambda*dirB; 
  dist=(res-x).NormL2();
  res = 0.5*(res+x);  
  if (lambda <0.0 ) return 1;

  // check if point is behind camera A
  if (dirA.ScalarProduct(res-pA)<0.0) return 2;  

  return 0;
}





