/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#ifndef __BIAS_IMAGECONTOUR_HH__
#define __BIAS_IMAGECONTOUR_HH__
  
// std includes
#include <vector>

// BIAS includes
#include <Base/Image/Image.hh>
#include <Base/Math/Vector2.hh>

namespace BIAS {
  /** @class Contour
      @ingroup g_image
      @brief a class for calculating the contour(boundary) of a region, 
      the contour is available as a point set or by freeman code        
      @author Dennis Herzog, Jan. 2005, edited by grest
  */
  class Contour { 
  public:          
    Contour() { background_=0; }
    /** creates the contour of the first blob found in the image;
        a pixel is assumed as foreground , if the pixel value is != 0;
        this function works even if the blob is located at the image border<br>
        IMPORTANT: the border of the image is set to zero !        
        @author herzog, edited by grest
    */
    int Process(BIAS::Image<unsigned char> &img);  
   
    /* draws the contour into img, which was calcuated by Process() before
     */
    int Draw(BIAS::Image<unsigned char> &img, 
             unsigned char greyValue=255);

    /** @param code the resulting 
        list of difference vectors coded in FreeManCode,
        that means: <br>
        0 = step right --<br>
        1 = step right & step down --<br>
        2 = step down --<br>
        3 = step left & step down --<br>
        ... --<br>
        6 = step up --<br>
        7 = step right & step up<br>
    */
    inline std::vector<int>& GetFreemanCode() { return directionList_;}

    /** all pixels belonging to the contour are returned
        The first entry corresponds to the first freemanCode,
        that means: the direction from the first to the second pixel
        is the first value in GetFreemanCode()
    */
    inline std::vector< BIAS::Vector2<int> >& GetPointSet() 
    { return contPixels_; }

    inline void SetBackGround(unsigned char backGroundColor)
    { background_=backGroundColor;}

  protected:

    inline bool IsFg(int x, int y) {
      unsigned char *pixel = data_[y] + x;
      return !( pixel[0] == background_ );
    }

    inline bool IsFg(int x, int y, int direction) {
      switch ( (direction+16)%8 ) {
      case 0: return IsFg(x+1,y+0);
      case 2: return IsFg(x+0,y+1);
      case 4: return IsFg(x-1,y+0);
      case 6: return IsFg(x+0,y-1);
      case 1: return IsFg(x+1,y+1);
      case 3: return IsFg(x-1,y+1);
      case 5: return IsFg(x-1,y-1);
      case 7: return IsFg(x+1,y-1);
      }
      BIASERR("IsFg: error");
      return false;
    }

    void FindFirst_(int &pos_x, int &pos_y);
    void Init_(BIAS::Image<unsigned char> &img);

    int img_w;
    int img_h;
    unsigned char **data_;

    unsigned char background_;
  
    std::vector<int> directionList_;
    std::vector< BIAS::Vector2<int> > contPixels_; // the contour pixels    
  };


}
#endif
