/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "Histogram.hh"

#include <Base/Image/ImageBase.hh>
#include <Base/Image/ImageDraw.hh>

namespace BIAS
{
using namespace std;

#define INST(type)\
template int Histogram::AddHist<type>(Image<type> &src, unsigned hist)

INST(float);

#ifdef BUILD_IMAGE_INT
INST(int);
#endif
#ifdef BUILD_IMAGE_CHAR
INST(char);
#endif
#ifdef BUILD_IMAGE_SHORT
INST(short);
#endif
#ifdef BUILD_IMAGE_USHORT
INST(unsigned short);
#endif
#ifdef BUILD_IMAGE_UINT
INST(unsigned int);
#endif
#ifdef BUILD_IMAGE_DOUBLE
INST(double);
#endif



Histogram::Histogram(unsigned short int size)
{
  ColorModel_ = ImageBase::CM_RGB;
  BinCount_ = 256;
  InterleavedDataOrder_ = true;
  HistCount_ = 0;
  HistSize_ = 256;
  _uiBorder = 2;
  _ucBorderColor = 128;
  Factor_ = size;
  ScaleYToMaxEntry_=true;

  Hist_ = NULL;
  Color_=NULL;
  Average_=NULL;
  MinBinVal_=NULL;
  BinSize_=NULL;
  MaxBinVal_=NULL;
  MaxHistEntry_ = NULL;
  NumData_ = NULL;
}

Histogram::~Histogram() 
{
  ReleaseHist();
}

void Histogram::SetFactor(unsigned short Factor)
{ 
  Factor_ = Factor; 
  InitHist(BinCount_, HistCount_);
}

void Histogram::SetColorModel()
{
  BIASERR("Histogram::SetColorModel() not valid for Histogram");
}


void Histogram::SetBinCount(unsigned int bincount)
{ 
  if (bincount != BinCount_){
    InitHist(bincount, HistCount_);
  }
}


void Histogram::SetHistCount(unsigned int histcount)
{
  if (histcount != HistCount_){
    InitHist(BinCount_, histcount);
  }
}

int Histogram::InitHist(unsigned int bincount , unsigned int histcount)
{
  int bc, hs;
  ReleaseHist();
  HistCount_ = histcount;
  BinCount_ = bincount;
  Hist_ = new double*[HistCount_];
  Color_ = new unsigned char*[HistCount_];
  MinBinVal_ = new double[HistCount_];
  MaxBinVal_ = new double[HistCount_];
  BinSize_ = new double[HistCount_];
  MaxHistEntry_ = new double[HistCount_];
  NumData_ = new double[HistCount_];
  Average_ = new double[HistCount_];
  for (register unsigned int i = 0; i < HistCount_; i++){
    Hist_[i] = new double[BinCount_];
    for (register unsigned int j = 0; j < BinCount_; j++){
      Hist_[i][j] = 0.0;
    }
    Color_[i] = new unsigned char[3];
    // deafult Color is black
    for (register int j = 0; j < 3; j++){
      Color_[i][j] = 0;
    }
    MinBinVal_[i]=DBL_MAX;
    MaxHistEntry_[i]=MaxBinVal_[i]=-DBL_MAX;
    NumData_[i]=0.0;
    Average_[i]=0.0;
  }

  // image width and height must be dividable by 4
  if (BinCount_ % 4 != 0)
    bc = BinCount_ + 4 - (BinCount_ % 4);
  else
    bc = BinCount_;
  if (HistSize_ % 4 != 0)
    hs = HistSize_ + 4 - (HistSize_ % 4);
  else
    hs = HistSize_;

  // scale image with Factor_
  Init((bc + 2 * _uiBorder) * Factor_ , (hs + 2 * _uiBorder) * Factor_ , 3);
  // white background
  FillImageWithConstValue(255);
  return 0;
}


int Histogram::ReleaseHist()
{
  if (Hist_){
    for (register int i =0; i < HistCount_; i++){
      delete[] Hist_[i];
      delete[] Color_[i];
    }
    delete[] Hist_; Hist_=NULL;
  }
  HistCount_=0;
  if (Color_){ delete[] Color_; Color_=NULL; }
  if (Average_){ delete[] Average_; Average_=NULL; }
  if (MinBinVal_){ delete[] MinBinVal_; MinBinVal_=NULL; }
  if (BinSize_){ delete[] BinSize_; BinSize_=NULL; }
  if (MaxBinVal_){ delete[] MaxBinVal_; MaxBinVal_=NULL; }
  if (MaxHistEntry_){ delete[] MaxHistEntry_; MaxHistEntry_ = NULL; }
  if (NumData_){ delete[] NumData_; NumData_ = NULL; }
  Release();
  return 0;
}

template <class StorageType>
int Histogram::AddHist(Image<StorageType>& Image, 
                       unsigned int hist )
{
#ifdef BIAS_DEBUG
  if (Image.GetChannelCount() != 1)
    BIASERR("Histogram::AddHist only supports one channel images");
  if ( hist >= HistCount_)
    BIASERR("Histogram::AddHist invalid hist "<<hist);
  if (Image.IsEmpty())
    BIASERR("Histogram::AddHist called with empty image");
#endif
  register StorageType *ImageData = Image.GetImageData();
  register StorageType *ImageDataEnd = ImageData + Image.GetPixelCount();
  StorageType min, max, diff;
  int bin;
  Image.GetMinMaxPixelValue(min, max);
  diff=max-min;
  MinBinVal_[hist]=(double)min;
  MaxBinVal_[hist]=(double)max;
  BinSize_[hist]=(MaxBinVal_[hist]-MinBinVal_[hist])/(double)BinCount_;

  while (ImageData < ImageDataEnd){
    if (*ImageData==max){
      bin=BinCount_-1;
    } else {
      bin=(unsigned)floor((double)BinCount_ * ((double)*ImageData-MinBinVal_[hist])/(double)diff);
    }  
#ifdef BIAS_DEBUG
      if (bin<0)
        BIASERR("invalid index "<<bin);
      if (bin>=(int)BinCount_)
        BIASERR("invalid index "<<bin<<" "
                <<*ImageData<<" "<<min<<" "<<BinSize_[hist]<<" "<<BinCount_);
#endif
    Hist_[hist][bin]++;
    if (Hist_[hist][bin]>MaxHistEntry_[hist]) 
      MaxHistEntry_[hist]=Hist_[hist][bin];
    ImageData++;
    
  }
  NumData_[hist]+=(double)Image.GetPixelCount();
  CalcAverage_(hist);
  
  return 0;
}

template <>
int Histogram::AddHist(Image<unsigned char>& Image, 
                       unsigned int hist )
{
#ifdef BIAS_DEBUG
  if (Image.GetChannelCount() != 1)
    BIASERR("Histogram::AddHist only supports one channel images");
  if ( hist >= HistCount_)
    BIASERR("Histogram::AddHist invalid hist "<<hist);
  if (Image.IsEmpty())
    BIASERR("Histogram::AddHist called with empty image");
#endif
  register unsigned char *ImageData = Image.GetImageData();
  register unsigned char *ImageDataEnd = ImageData + Image.GetPixelCount();

  while (ImageData < ImageDataEnd){
    if (Hist_[hist][BinCount_ * *ImageData / 256] + 1 > MaxHistEntry_[hist] ){
      MaxHistEntry_[hist] = Hist_[hist][BinCount_ * *ImageData++ / 256]++;
    } else {
      Hist_[hist][BinCount_ * *ImageData++ / 256]++;
    }
  }
  CalcAverage_(hist);
  MinBinVal_[hist]=0.0;
  MaxBinVal_[hist]=255.0;
  NumData_[hist]+=(double)Image.GetPixelCount();
  BinSize_[hist]=(MaxBinVal_[hist]-MinBinVal_[hist])/(double)BinCount_;
  return 0;
}


template <class DataType>
int Histogram::AddHist(vector<DataType>& data, unsigned int hist)
{
  typename vector<DataType>::iterator it;
  DataType max=data.front(), min=data.front();
  BIASCDOUT(D_HIST_VEC, "data.size() = "<<data.size()<<endl);
  for (it=data.begin(); it!=data.end(); it++){
    if ((*it)>max) max=*it;
    if ((*it)<min) min=*it;
    BIASCDOUT(D_HIST_VEC, (*it)<<" ");
  }
  BIASCDOUT(D_HIST_VEC, "\nmin/max "<<min<<"/"<<max);
  return AddHist(data, min, max, hist);
}

template <class DataType>
int Histogram::AddHist(vector<DataType>& data, DataType min, DataType max, 
                       unsigned int hist)
{


  if (data.size()==0){
    BIASERR("data has size 0");
    return -1;
  }
  if (hist>HistCount_){
    BIASERR("hist="<<hist<<" to big. HistCount_="<<HistCount_);
    return -2;
  }

  typename vector<DataType>::iterator it;
  int bin;
  BIASCDOUT(D_HIST_VEC, "data.size() = "<<data.size()<<endl);
  
  BIASCDOUT(D_HIST_VEC, "\nmin/max "<<min<<"/"<<max);
  MinBinVal_[hist]=(double)min;
  MaxBinVal_[hist]=(double)max;
  BIASCDOUT(D_HIST_VEC, "MinBinVal_["<<hist<<"]/MaxBinVal_["<<hist<<"] "<<
            MinBinVal_[hist]<<"/"<<MaxBinVal_[hist]);
  BinSize_[hist]=(MaxBinVal_[hist]-MinBinVal_[hist])/(double)BinCount_;
  BIASCDOUT(D_HIST_VEC, "min: "<<MinBinVal_[hist]<<"   max: "<<MaxBinVal_[hist]
            <<"   binsize: "<<BinSize_[hist]<<endl);
  for (it=data.begin(); it!=data.end(); it++){
    if (*it==max){
      bin=BinCount_-1;
    } else {
      bin=(int)floor(double((*it)-min)/BinSize_[hist]);
    }
#ifdef BIAS_DEBUG
    if (bin<0){
      BIASERR("invalid index "<<bin<<"  data "<<(*it)<<"  min "<<min
              <<"  binsize "<<BinSize_[hist]);
      BIASASSERT(false);
    }
    if (bin>=(int)BinCount_){
      BIASERR("invalid index "<<bin<<" "
              <<*it<<" "<<min<<" "<<BinSize_[hist]<<" "<<BinCount_);
      BIASASSERT(false);
    }
#endif
    Hist_[hist][bin]++;
    if (Hist_[hist][bin]>MaxHistEntry_[hist]) 
      MaxHistEntry_[hist]=Hist_[hist][bin];
  }
  BIASCDOUT(D_HIST_VEC, "MaxHistEntry_["<<hist<<"] = "<<MaxHistEntry_[hist]
            <<endl);
  NumData_[hist]+=(double)data.size();
  CalcAverage_(hist);
  return 0;
}


template int Histogram::AddHist<double>(vector<double>& data, double min, double max, unsigned hist);
template int Histogram::AddHist<float>(vector<float>& data, float min, float max, unsigned hist);
template int Histogram::AddHist<int>(vector<int>& data, int min, int max, unsigned hist);
template int Histogram::AddHist<unsigned int>(vector<unsigned>& data, unsigned min, unsigned max, unsigned hist);
template int Histogram::AddHist<double>(vector<double>& data, unsigned hist);
template int Histogram::AddHist<float>(vector<float>& data, unsigned hist);
template int Histogram::AddHist<int>(vector<int>& data, unsigned hist);
template int Histogram::AddHist<unsigned int>(vector<unsigned>& data, unsigned hist);

template <class DataType>
int Histogram::AddWeightedHist(vector<DataType>& data, vector<double>& weight, 
                               unsigned int hist)
{
#ifdef BIAS_DEBUG
  if (data.size() != weight.size()){
    BIASERR("data and weight have different sizes "<<data.size()
            <<" "<<weight.size());
    BIASASSERT(data.size() == weight.size());
  }
#endif
  typename vector<DataType>::iterator it;
  vector<double>::iterator wit;
  int bin;
  DataType max=data.front(), min=data.front();
  for (it=data.begin(); it!=data.end(); it++){
    if ((*it)>max) max=*it;
    if ((*it)<min) min=*it;
  }
  MinBinVal_[hist]=(double)min;
  MaxBinVal_[hist]=(double)max;
  BinSize_[hist]=(MaxBinVal_[hist]-MinBinVal_[hist])/(double)BinCount_;
  
  for (it=data.begin(), wit=weight.begin(); it!=data.end(); it++, wit++){
    if (*it==max){
      bin=BinCount_-1;
    } else {
      bin=(int)floor(double((*it)-min)/BinSize_[hist]);
    }
#ifdef BIAS_DEBUG
    if (*it!=max){
      if (bin<0)
        BIASERR("invalid index "<<bin);
      if (bin>=(int)BinCount_)
        BIASERR("invalid index "<<bin<<" "
                <<*it<<" "<<min<<" "<<BinSize_[hist]<<" "<<BinCount_);
    }
#endif
    Hist_[hist][bin]+=(*wit);
    if (Hist_[hist][bin]>MaxHistEntry_[hist]) 
      MaxHistEntry_[hist]=Hist_[hist][bin];
  }

  NumData_[hist]+=(double)data.size();
  CalcAverage_(hist);
  return 0;
}

template int Histogram::AddWeightedHist<double>(vector<double>& data, vector<double>& weight, unsigned hist);
template int Histogram::AddWeightedHist<float>(vector<float>& data, vector<double>& weight, unsigned hist);
template int Histogram::AddWeightedHist<int>(vector<int>& data, vector<double>& weight, unsigned hist);
template int Histogram::AddWeightedHist<unsigned>(vector<unsigned>& data, vector<double>& weight, unsigned hist);

int Histogram::CalcAverage_(unsigned int hist)
{
  Average_[hist] = 0.0;
  for (unsigned int i=0; i<BinCount_; i++)
    Average_[hist] += Hist_[hist][i];
  //cerr << "sum Hist["<<hist<<"]="<<Average_[hist]<<endl;
  Average_[hist] /= BinCount_;  
  //cerr << "average Hist["<<hist<<"]="<<Average_[hist]<<endl;
  return 0;
}



int Histogram::ZeroHist(unsigned int hist)
{
  MaxHistEntry_[hist]=0.0;
  NumData_[hist]=0.0;
  Average_[hist]=0.0;
  MaxBinVal_[hist]=MinBinVal_[hist]=BinSize_[hist]=0.0;
  if (Hist_!=NULL)
    memset(Hist_[hist], 0, BinCount_ * sizeof(double));
  FillImageWithConstValue(255);
  return 0;
}


int Histogram::RedrawHist_(unsigned int hist)
{
  return DeleteHist_(hist) + DrawHist_(hist);
}


int Histogram::DrawHist_(unsigned int hist)
{
  unsigned int BorderOffset = Factor_ * _uiBorder * GetChannelCount();
  register unsigned char **ImageDataArray = GetImageDataArray();
  unsigned start[2], end[2], st[2];
 
  // now draw hist
  register int y, x, kk;
  start[0]=Factor_ * _uiBorder;
  start[1]=Factor_ * _uiBorder + Factor_ * (HistSize_-1);
  double maxy=(ScaleYToMaxEntry_)?(MaxHistEntry_[hist]):(NumData_[hist]);
  for (register unsigned int i = 0; i < BinCount_ ; i++){
    y = Factor_ * _uiBorder + Factor_ * 
      (((int)HistSize_-1)-(int)rint(((int)HistSize_-1)*Hist_[hist][i] / 
                                    maxy));
    x = BorderOffset + Factor_ * 3 * i;
#ifdef BIAS_DEBUG
    if (y<0 || y>=(int)GetHeight()){
      BIASERR("invalid y="<<y<<" HistSize_="<<HistSize_<<"  Hist_["<<hist<<"]["
              <<i<<"]="<<Hist_[hist][i]<<"  MaxHistEntry_="
              <<MaxHistEntry_[hist]);
      BIASASSERT(false);
    }
#endif
    switch (Mode_ ) {
    case HIST_POINTS:
      for (register unsigned int k = 0; k < Factor_ ; k++){
        kk = k*3;
        for (register unsigned int j = 0; j < Factor_ ; j++){
          ImageDataArray[y + j][x + kk] = Color_[hist][0];
          ImageDataArray[y + j][x + kk + 1] = Color_[hist][1];
          ImageDataArray[y + j][x + kk + 2] = Color_[hist][2];
        }
      }
      break;
    case HIST_BARS:
      for (register unsigned int k = 0; k < Factor_ ; k++){
        kk = k*3;
        for (register unsigned int j = y; 
             j < GetHeight() - Factor_ * _uiBorder; j++){
          ImageDataArray[j][x + kk] = Color_[hist][0];
          ImageDataArray[j][x + kk + 1] = Color_[hist][1];
          ImageDataArray[j][x + kk + 2] = Color_[hist][2];
        }
      }
      break;
    case HIST_LINES:
      for (register unsigned int k = 0; k < Factor_ ; k++){
        for (register unsigned int j = 0; j < Factor_ ; j++){
          st[0]=start[0]+k;
          st[1]=start[1]+j;
          end[0]=x/3+k;
          end[1]=y+j;
          ImageDraw<unsigned char>::Line(*this, st, end, Color_[hist]);
        }
      }
      start[0]=x/3;
      start[1]=y;
      break;
    }
  } 
  // draw line of average 
  y = Factor_ * (_uiBorder + (HistSize_ - 1) - 
                 (int)rint((HistSize_ - 1) * Average_[hist] / 
                           maxy)) +
    Factor_ * _uiBorder;
  //cerr << "average of Hist_["<<hist<<"]="<<Average_[hist]<<" at "<<y<<endl;
  for (unsigned int j=BorderOffset; 
       j < (BinCount_ - 1) * Factor_ * 3 + BorderOffset; 
       j+=3){
	  ImageDataArray[y][j] = Color_[hist][0];
	  ImageDataArray[y][j+1] = Color_[hist][1];
	  ImageDataArray[y][j+2] = Color_[hist][2];
  }
  return 0;
}

int Histogram::DrawHistLog_(unsigned int hist)
{
  unsigned int BorderOffset = Factor_ * _uiBorder * GetChannelCount();
  register unsigned char **ImageDataArray = GetImageDataArray();
  unsigned start[2], end[2], st[2];
 
  // now draw hist
  register int y, x, kk;
  start[0]=Factor_ * _uiBorder;
  start[1]=Factor_ * _uiBorder + Factor_ * (HistSize_-1);
  double maxy=(ScaleYToMaxEntry_)?(MaxHistEntry_[hist]):(NumData_[hist]);
  for (register unsigned int i = 0; i < BinCount_ ; i++){
    y = Factor_ * _uiBorder + Factor_ * 
      (((int)HistSize_-1)-
       (int)rint(((int)HistSize_-1)*
                 log10((Hist_[hist][i]/maxy)*9.0+1.0)));
    x = BorderOffset + Factor_ * 3 * i;
#ifdef BIAS_DEBUG
    if (y<0 || y>=(int)GetHeight()){
      BIASERR("invalid y="<<y<<" HistSize_="<<HistSize_<<"  Hist_["<<hist<<"]["
              <<i<<"]="<<Hist_[hist][i]<<"  MaxHistEntry_="
              <<MaxHistEntry_[hist]);
      BIASASSERT(false);
    }
#endif
    switch (Mode_ ) {
    case HIST_POINTS:
      for (register unsigned int k = 0; k < Factor_ ; k++){
        kk = k*3;
        for (register unsigned int j = 0; j < Factor_ ; j++){
          ImageDataArray[y + j][x + kk] = Color_[hist][0];
          ImageDataArray[y + j][x + kk + 1] = Color_[hist][1];
          ImageDataArray[y + j][x + kk + 2] = Color_[hist][2];
        }
      }
      break;
    case HIST_BARS:
      for (register unsigned int k = 0; k < Factor_ ; k++){
        kk = k*3;
        for (register unsigned int j = y; 
             j < GetHeight() - Factor_ * _uiBorder; j++){
          ImageDataArray[j][x + kk] = Color_[hist][0];
          ImageDataArray[j][x + kk + 1] = Color_[hist][1];
          ImageDataArray[j][x + kk + 2] = Color_[hist][2];
        }
      }
      break;
    case HIST_LINES:
      for (register unsigned int k = 0; k < Factor_ ; k++){
        for (register unsigned int j = 0; j < Factor_ ; j++){
          st[0]=start[0]+k;
          st[1]=start[1]+j;
          end[0]=x/3+k;
          end[1]=y+j;
          ImageDraw<unsigned char>::Line(*this, st, end, Color_[hist]);
        }
      }
      start[0]=x/3;
      start[1]=y;
      break;
    }
  } 
  // draw line of average 
  y = Factor_ * (_uiBorder + (HistSize_ - 1) - 
                 (int)rint((HistSize_ - 1) * Average_[hist] / 
                           maxy)) +
    Factor_ * _uiBorder;
  //cerr << "average of Hist_["<<hist<<"]="<<Average_[hist]<<" at "<<y<<endl;
  for (unsigned int j=BorderOffset; 
       j < (BinCount_ - 1) * Factor_ * 3 + BorderOffset; 
       j+=3){
	  ImageDataArray[y][j] = Color_[hist][0];
	  ImageDataArray[y][j+1] = Color_[hist][1];
	  ImageDataArray[y][j+2] = Color_[hist][2];
  }
  return 0;
}


int Histogram::DeleteHist_(unsigned int hist)
{
  register unsigned char *r = GetImageData(); 
  register unsigned char *g = r + 1; 
  register unsigned char *b = g + 1;
  register unsigned char *end = r + GetChannelCount() * GetPixelCount();

  while (r < end){
    if ( (*r == Color_[hist][0]) && (*g == Color_[hist][1]) && 
	 (*b == Color_[hist][2]) )
      *r = *g = *b = 255;
    r+=3;
    g+=3;
    b+=3;
  }
  return 0;
}


int Histogram::SetColor(unsigned char R, unsigned char G, unsigned char B, 
			   unsigned int hist)
{
#ifdef BIAS_DEBUG
  if (hist + 1 > HistCount_ )
    BIASERR("Histogram::SetColor() inavlid hist "<<hist);
#endif
  Color_[hist][0] = R;
  Color_[hist][1] = G;
  Color_[hist][2] = B;
  
  return 0;
}


int Histogram::Draw()
{
  register unsigned char **ImageDataArray = GetImageDataArray();
  // draw border
  for (unsigned int i=0; i< Factor_ * _uiBorder; i++){
    memset((void *)(ImageDataArray[i]), _ucBorderColor, GetWidthStep() * 
           sizeof(unsigned char));
    memset((void *)(ImageDataArray[GetHeight()-1-i]), _ucBorderColor, 
           GetWidthStep() * sizeof(unsigned char) );
    for (unsigned int j=Factor_ * _uiBorder; j<GetHeight()-Factor_ * _uiBorder;
         j++){
      for (unsigned int c=0; c<GetChannelCount(); c++){
        ImageDataArray[j][i*GetChannelCount()+c]=_ucBorderColor;
        ImageDataArray[j][(GetWidth()-1-i)*GetChannelCount()+c]=_ucBorderColor;
      }
    }
  }
  for (int i = 0; i < HistCount_; i++)
    if (MaxHistEntry_[i]>0)
      DrawHist_(i);
  return 0;
}

int Histogram::DrawLog()
{
  register unsigned char **ImageDataArray = GetImageDataArray();
  // draw border
  for (unsigned int i=0; i< Factor_ * _uiBorder; i++){
    memset((void *)(ImageDataArray[i]), _ucBorderColor, GetWidthStep() * 
           sizeof(unsigned char));
    memset((void *)(ImageDataArray[GetHeight()-1-i]), _ucBorderColor, 
           GetWidthStep() * sizeof(unsigned char) );
    for (unsigned int j=Factor_ * _uiBorder; j<GetHeight()-Factor_ * _uiBorder;
         j++){
      for (unsigned int c=0; c<GetChannelCount(); c++){
        ImageDataArray[j][i*GetChannelCount()+c]=_ucBorderColor;
        ImageDataArray[j][(GetWidth()-1-i)*GetChannelCount()+c]=_ucBorderColor;
      }
    }
  }
  for (int i = 0; i < HistCount_; i++)
    DrawHistLog_(i);
  return 0;
}

double Histogram::GetBin(unsigned int bin,unsigned int hist)
{
  if (bin < BinCount_ && hist < HistCount_) return Hist_[hist][bin];
  else return -1;
}

int Histogram::SetBin(unsigned int bin,double value, unsigned int hist)
{
  if (bin < BinCount_ && hist < HistCount_) {
    Hist_[hist][bin] = value;
    MaxHistEntry_[hist] =0.0;
    
    for (unsigned int i=0; i<BinCount_; i++)
      if (MaxHistEntry_[hist] < Hist_[hist][i]) 
          MaxHistEntry_[hist] = Hist_[hist][i];
    CalcAverage_(hist);
    return 0;
  }
  else return -1;
  
}

void Histogram::Dump(ostream& os)
{
  double *s=new double[HistCount_];
  memset((void *)s, 0, sizeof(double)*HistCount_);
  for (int b=0; b<(int)BinCount_; b++){
    os << setw(4) << b <<": ";
    for (int h=0; h<(int)HistCount_; h++){
      os << setw(8) << Hist_[h][b];
      s[h]+=Hist_[h][b];
    }
    os << endl;
  }
  os << " sum: ";
  for (int h=0; h<(int)HistCount_; h++){
    os << setw(8) << s[h];
  }
  os << endl;
  os << " avg: ";
  for (int h=0; h<(int)HistCount_; h++){
    os << setw(8) << Average_[h];
  }
  os << endl;
  delete[] s;
}

bool Histogram::IsEmpty()
{
  bool result=true;
  for (int h=0; h<(int)HistCount_; h++){
    result = result && (MaxHistEntry_[h]==0);
  }
  return result;
}

int Histogram::WriteASCII(const std::string& filename)
{
  // first write data file
  ofstream of(filename.c_str());
  if (!of){
    BIASERR("error opening "<<filename);
    return -1;
  }

  WriteASCII(of);

  if (!of){
    BIASERR("error writing "<<filename);
    return -2;
  }
  of.close();

  return 0;
}

void Histogram::WriteASCII(ostream& os)
{
  os << "# "<<HistCount_<<" histogram(s) with "<<BinCount_<<" bins"<<endl;
  os << "# bin   - number of the bin\n";
  os << "# binb  - binborders: [binb, binb+1)\n"; 
  os << "# binv  - number of values contained in bin\n"; 
  os << "#"<<setw(7)<<"bin";
  double *bs=new double[HistCount_];
  for (int h=0; h<(int)HistCount_; h++){
    bs[h]=(MaxBinVal_[h]-MinBinVal_[h])/(double)BinCount_;
    os <<setw(15)<<"binb"<<setw(12)<<"binv";
  }
  os << endl;
  for (int b=0; b<(int)BinCount_; b++){
    os << setw(8) << b;
    for (int h=0; h<(int)HistCount_; h++){
      os <<setw(15) << MinBinVal_[h]+b*bs[h] << setw(12) << Hist_[h][b];
    }
    os << endl;
  }
  delete[] bs;
}

void Histogram::CooTransf(double datacoo[2], double imcoo[2], unsigned hist)
{
  double d[2];
  double maxy=(ScaleYToMaxEntry_)?(MaxHistEntry_[hist]):(NumData_[hist]);
  
  d[0]=(datacoo[0]<MinBinVal_[hist])?(MinBinVal_[hist]):
    ((datacoo[0]>MaxBinVal_[hist])?(MaxBinVal_[hist]):(datacoo[0]));
  d[1]=(datacoo[1]<0.0)?(0.0):
    ((datacoo[1]>maxy)?(maxy):(datacoo[1]));

  imcoo[0]=Factor_ * 
    ((double)_uiBorder+(d[0]-MinBinVal_[hist])/BinSize_[hist]);
  imcoo[1]=Factor_ * ((double)_uiBorder+double((int)HistSize_-1)*
     (1.0-d[1]/maxy));
}

void Histogram::CooTransf(double datacoo[2], unsigned imcoo[2], unsigned hist)
{
  double d[2];
  CooTransf(datacoo, d, hist);
  imcoo[0]=(unsigned)rint(d[0]);
  imcoo[1]=(unsigned)rint(d[1]);
}

/////////////////////////////////////////////////////////////////////////
///   Histogram2D
/////////////////////////////////////////////////////////////////////////

Histogram2D::Histogram2D(unsigned short int zoomfactor) 
  : Image<unsigned char>()
{
  ColorModel_ = ImageBase::CM_RGB;
  InterleavedDataOrder_ = true;
  Factor_=zoomfactor;
  XBinCount_=YBinCount_=0;
  Hist_=NULL;
  MaxHistEntry_=XMinBinVal_=XMaxBinVal_=YMinBinVal_=YMaxBinVal_=0.0;
  XBinSize_=YBinSize_=0.0;
  _uiBorder=2;
  _ucBorderColor=128;
}

Histogram2D::~Histogram2D()
{
  _ReleaseHist();
}

void Histogram2D::SetColorModel()
{
  BIASERR("Histogram::SetColorModel() not valid for Histogram");
}

void Histogram2D::SetFactor(unsigned short Factor)
{ 
  Factor_ = Factor; 
  InitHist(XBinCount_, YBinCount_);
}

void Histogram2D::_ReleaseHist()
{
  if (Hist_!=NULL){
    delete[] Hist_[0];
    delete[] Hist_;
  }
  XBinCount_=YBinCount_=0;
  Hist_=NULL;
  MaxHistEntry_=XMinBinVal_=XMaxBinVal_=YMinBinVal_=YMaxBinVal_=0.0;
  XBinSize_=YBinSize_=0.0;
  Release();
}

void Histogram2D::InitHist(unsigned xbincount, unsigned ybincount)
{
  if (Hist_!=NULL){
    _ReleaseHist();
  }
  XBinCount_=ybincount;
  YBinCount_=xbincount;
  Hist_=new double*[YBinCount_];
  Hist_[0]=new double[XBinCount_*YBinCount_];
  for (unsigned i=1; i<YBinCount_; i++)
    Hist_[i]=Hist_[i-1]+XBinCount_;
  // scale image with Factor_
  Init((XBinCount_ + 2 * _uiBorder) * Factor_ , 
       (YBinCount_ + 2 * _uiBorder) * Factor_ , 3);
  // white background
  FillImageWithConstValue(255);
}

template <class DataType>
void Histogram2D::AddHist(vector<DataType>& xdata, vector<DataType>& ydata)
{
  DataType maxx=xdata.front(), minx=xdata.front(), maxy=ydata.front(), 
    miny=ydata.front();
  typename vector<DataType>::iterator itx, ity;
  int xbin, ybin;

  for (itx=xdata.begin(); itx!=xdata.end(); itx++){
    if (*itx>maxx) maxx=*itx;
    if (*itx<minx) minx=*itx;
  }
  for (ity=ydata.begin(); ity!=ydata.end(); ity++){
    if (*ity>maxy) maxy=*ity;
    if (*ity<miny) miny=*ity;
  }
  XMinBinVal_=(double)minx;
  XMaxBinVal_=(double)maxx;
  YMinBinVal_=(double)miny;
  YMaxBinVal_=(double)maxy;
  
  XBinSize_=(XMaxBinVal_-XMinBinVal_)/(double)XBinCount_;
  YBinSize_=(YMaxBinVal_-YMinBinVal_)/(double)YBinCount_;
#ifdef BIAS_DEBUG
  if (xdata.size()!=ydata.size()){
    BIASERR("data vectors have different size "<<xdata.size()<<" "
            <<ydata.size());
  }
#endif
  for (itx=xdata.begin(), ity=ydata.begin(); itx!=xdata.end(); itx++, ity++){
    if (*itx==maxx){
      xbin=(int)XBinCount_-1;
    } else {
      xbin=(int)floor(double((*itx)-minx)/XBinSize_);
    }
    if (*ity==maxy){
      ybin=YBinCount_-1;
    } else {
      ybin=(int)floor(double((*ity)-miny)/YBinSize_);
    }
#ifdef BIAS_DEBUG
    if (xbin<0 || ybin<0)
      BIASERR("invalid index "<<xbin<<" "<<ybin<<" "<<XBinCount_
              <<" "<<YBinCount_<<" "<<*itx<<" "<<*ity);
    if (xbin>=(int)XBinCount_ || ybin>=(int)YBinCount_ )
      BIASERR("invalid index "<<xbin<<" "<<ybin<<"   "<<*itx<<" "<<minx<<" "
              <<maxx<<" "<<XBinSize_<<" "<<XBinCount_<<"   "<<*ity<<" "<<miny
              <<" "<<maxy<<" "<<YBinSize_<<" "<<YBinCount_);
#endif
    Hist_[ybin][xbin]++;
    if (Hist_[ybin][xbin]>MaxHistEntry_) MaxHistEntry_=Hist_[ybin][xbin];
  }
  /*
  cerr << "MaxHistEntry_="<<MaxHistEntry_<<endl;
  double av=0.0;
  for (unsigned y=0; y<YBinCount_; y++){
    for (unsigned x=0; x<XBinCount_; x++){
      av+=Hist_[y][x];
    }
  }
  cerr << "sum="<<av<<endl;
  cerr << "Average="<<av/double(YBinCount_*XBinCount_)<<endl;
  */
}

template void 
Histogram2D::AddHist<double>(vector<double>& xdata, vector<double>& ydata);
template void 
Histogram2D::AddHist<float>(vector<float>& xdata, vector<float>& ydata);
template void 
Histogram2D::AddHist<int>(vector<int>& xdata, vector<int>& ydata);
template void 
Histogram2D::AddHist<unsigned>(vector<unsigned>& xdata, vector<unsigned>& ydata);

void Histogram2D::ZeroHist()
{
  BIASASSERT(Hist_!=NULL);
  memset((void *)Hist_[0], 0, sizeof(double)*XBinCount_*YBinCount_);
}

void Histogram2D::Draw()
{
  BIASASSERT(Hist_!=NULL);
  unsigned char **ida=GetImageDataArray(), r, g, b=0;
  unsigned mx, my, a, f3=3*Factor_;

  // draw border
  for (unsigned int i=0; i< Factor_ * _uiBorder; i++){
    memset((void *)(ida[i]), _ucBorderColor, GetWidthStep() * 
           sizeof(unsigned char));
    memset((void *)(ida[GetHeight()-1-i]), _ucBorderColor, 
           GetWidthStep() * sizeof(unsigned char) );
    for (unsigned int j=Factor_ * _uiBorder; j<GetHeight()-Factor_ * _uiBorder;
         j++){
      for (unsigned int c=0; c<GetChannelCount(); c++){
        ida[j][i*GetChannelCount()+c]=_ucBorderColor;
        ida[j][(GetWidth()-1-i)*GetChannelCount()+c]=_ucBorderColor;
      }
    }
  }

  // draw histogram
  for (unsigned y=0; y<YBinCount_; y++){
    my=(_uiBorder+y)*Factor_;
    for (unsigned x=0; x<XBinCount_; x++){
      mx=(_uiBorder+x)*3*Factor_;
      if (Hist_[y][x]==0.0){
        r=g=0; 
      } else {
        a=(unsigned)rint(Hist_[y][x]/MaxHistEntry_*255.0);
        if (Hist_[y][x]>MaxHistEntry_) {
          a=255;
          BIASERR("error in drawing depth");
        } // from green over yellow to red
        r=(a>=128)?255:(a*2);
        g=(a>=128)?((255-a)*2):255;
      }
//       if (r==255 && g==0) 
//         cerr <<setw(4)<<x<<setw(4)<<y<<setw(10)<<Hist_[y][x]<<setw(6)<<a<<setw(6)<<(int)r<<setw(6)<<(int)g<<endl;
      for (int ro=0; ro<Factor_; ro++){
        for (unsigned c=0; c<f3; c+=3){
          ida[my+ro][mx+c]=r;
          ida[my+ro][mx+c+1]=g;
          ida[my+ro][mx+c+2]=b;
        }
      }
    }
  }
}

void Histogram2D::DrawLog()
{
  BIASASSERT(Hist_!=NULL);
  unsigned char **ida=GetImageDataArray(), r, g, b=0;
  unsigned mx, my, a, f3=3*Factor_;

  // draw border
  for (unsigned int i=0; i< Factor_ * _uiBorder; i++){
    memset((void *)(ida[i]), _ucBorderColor, GetWidthStep() * 
           sizeof(unsigned char));
    memset((void *)(ida[GetHeight()-1-i]), _ucBorderColor, 
           GetWidthStep() * sizeof(unsigned char) );
    for (unsigned int j=Factor_ * _uiBorder; j<GetHeight()-Factor_ * _uiBorder;
         j++){
      for (unsigned int c=0; c<GetChannelCount(); c++){
        ida[j][i*GetChannelCount()+c]=_ucBorderColor;
        ida[j][(GetWidth()-1-i)*GetChannelCount()+c]=_ucBorderColor;
      }
    }
  }

  // draw histogram
  for (unsigned y=0; y<YBinCount_; y++){
    my=(_uiBorder+YBinCount_-1-y)*Factor_;
    for (unsigned x=0; x<XBinCount_; x++){
      mx=(_uiBorder+x)*3*Factor_;
      if (Hist_[y][x]==0.0){
        r=g=0; 
      } else {
        a=(unsigned)rint(log10((Hist_[y][x]/MaxHistEntry_)*9.0+1.0)*255.0);
        if (Hist_[y][x]>MaxHistEntry_) {
          a=255;
          BIASERR("error in drawing depth");
        } // from green over yellow to red
        r=(a>=128)?255:(a*2);
        g=(a>=128)?((255-a)*2):255;
      }
//       if (r==255 && g==0) 
//         cerr <<setw(4)<<x<<setw(4)<<y<<setw(10)<<Hist_[y][x]<<setw(6)<<a<<setw(6)<<(int)r<<setw(6)<<(int)g<<endl;
      for (int ro=0; ro<Factor_; ro++){
        for (unsigned c=0; c<f3; c+=3){
          ida[my+ro][mx+c]=r;
          ida[my+ro][mx+c+1]=g;
          ida[my+ro][mx+c+2]=b;
        }
      }
    }
  }
}

void Histogram2D::Dump(std::ostream& os)
{
  os <<"# 2D Histogram with "<<XBinCount_<<" x "<<YBinCount_<<" bins\n"
     <<"# x bins from "<<XMinBinVal_<<" to "<<XMaxBinVal_<<endl
     <<"# y bins from "<<YMinBinVal_<<" to "<<YMaxBinVal_<<endl
     <<"# x, y        - coo of bin\n"
     <<"# xval, yval  - the value of the entries of the bin\n"
     <<"# count       - number of entries in bin\n"
     <<"#\n"
     <<"# "<<setw(4)<<"y"<<setw(15)<<"yval"<<setw(6)<<"x"<<setw(15)<<"xval"
     <<setw(15)<<"val"<<setw(6)<<"x"<<setw(15)<<"xval"<<setw(15)<<"val"
     <<setw(6)<<"x"<<setw(15)<<"xval"<<setw(15)<<"count"
     <<setw(6)<<"...\n";
  for (unsigned y=0; y<YBinCount_; y++){
    os << setw(6) << y << setw(15) << YMinBinVal_+y*YBinSize_;
    for (unsigned x=0; x<XBinCount_; x++){
      os << setw(6) << x << setw(15) << XMinBinVal_+x*XBinSize_
         << setw(15) << Hist_[y][x];
    }
    os << endl;
  }
}

void Histogram2D::WriteASCII(std::ostream& os)
{
  os <<"# 2D Histogram with "<<XBinCount_<<" x "<<YBinCount_<<" bins\n"
     <<"# x bins from "<<XMinBinVal_<<" to "<<XMaxBinVal_<<endl
     <<"# y bins from "<<YMinBinVal_<<" to "<<YMaxBinVal_<<endl
     <<"# x, y        - coo of bin\n"
     <<"# xval, yval  - the value of the entries of the bin\n"
     <<"# count       - number of entries in bin\n"
     <<"#\n"
     <<"# "<<setw(6)<<"y"<<setw(15)<<"yval"<<setw(8)<<"x"<<setw(15)<<"xval"
     <<setw(15)<<"count\n";
  for (unsigned y=0; y<YBinCount_; y++){
    for (unsigned x=0; x<XBinCount_; x++){
      os << setw(8) << y << setw(15) << YMinBinVal_+y*YBinSize_
         << setw(8) << x << setw(15) << XMinBinVal_+x*XBinSize_
         << setw(15) << Hist_[y][x] << endl;
    }
    os << endl << endl << endl;
  }
}

void Histogram2D::CooTransf(double datacoo[2], double imcoo[2])
{
  double d[2];
  d[0]=(datacoo[0]<XMinBinVal_)?(XMinBinVal_):
    ((datacoo[0]>XMaxBinVal_)?(XMaxBinVal_):(datacoo[0]));
  d[1]=(datacoo[1]<YMinBinVal_)?(YMinBinVal_):
    ((datacoo[1]>YMaxBinVal_)?(YMaxBinVal_):(datacoo[1]));

  imcoo[0]=Factor_ * (_uiBorder + (d[0]-XMinBinVal_)/XBinSize_);
  imcoo[1]=Factor_ * (_uiBorder + (d[1]-YMinBinVal_)/YBinSize_);

}

void Histogram2D::CooTransf(double datacoo[2], unsigned imcoo[2])
{
  double d[2];
  CooTransf(datacoo, d);
  imcoo[0]=(unsigned)rint(d[0]);
  imcoo[1]=(unsigned)rint(d[1]);
}

}
