/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "HomographyMapping.hh"
#include <Geometry/RMatrix.hh>
#include <Base/Image/ImageConvert.hh>
#include <Base/Geometry/HomgPoint2D.hh>


using namespace BIAS;

template class HomographyMapping<unsigned char, unsigned char>;
template class HomographyMapping<float, float>;
template class HomographyMapping<unsigned char, float>;

#if defined(BUILD_IMAGE_CHAR)
template class HomographyMapping<unsigned char, char>;
template class HomographyMapping<char, char>;
#endif

#if defined(BUILD_IMAGE_USHORT)
template class HomographyMapping<unsigned short,unsigned short>;
#endif

#if defined(BUILD_IMAGE_SHORT)
template class HomographyMapping<short, short>;
#endif

#if defined(BUILD_IMAGE_SHORT)&&defined(BUILD_IMAGE_USHORT)
template class HomographyMapping<unsigned short, short>;
#endif

#if defined(BUILD_IMAGE_INT)
template class HomographyMapping<int,int>;
#endif

#if defined(BUILD_IMAGE_USHORT)
template class HomographyMapping<unsigned short,float>;
#endif

#if defined(BUILD_IMAGE_USHORT) && defined(BUILD_IMAGE_INT)
template class HomographyMapping<unsigned short,int>;
#endif

#if defined(BUILD_IMAGE_DOUBLE)
template class HomographyMapping<double,double>;
#endif


template <class InputStorageType, class OutputStorageType>
int HomographyMapping<InputStorageType, OutputStorageType>::
RotateToGrey(Image<InputStorageType>& src, 
             Image<OutputStorageType>& sink,
             float PhiX, float PhiY, float PhiZ, 
             const Matrix3x3<double> &K) {
  Matrix3x3<double> Kinv;
  Matrix3x3<double> KRKinv;
  Image<InputStorageType> source;

  double centerx,centery;
    
  if (!src.IsEmpty()){
      
    if(src.GetColorModel() != ImageBase::CM_Grey)
      ImageConvert::Convert(src, source, ImageBase::CM_Grey);
    else
      source=src;
      
    if(!sink.IsEmpty())
      sink.Release();
      
    sink.Init(source.GetWidth(),source.GetHeight(),source.GetChannelCount());
      
    // Prepare inverse camera mattix
    Kinv.set_identity();
    Kinv[0][0] = 1 / K[0][0];
    Kinv[1][1] = 1 / K[1][1];
    Kinv[0][2] = -K[0][2]/K[0][0];
    Kinv[1][2] = -K[1][2]/K[1][1];
      
    centerx = K[0][0];
    centery = K[1][1];
      
    RMatrix R;
    R.SetXYZ(PhiX,PhiY,PhiZ);
    KRKinv = K*R*Kinv;
      
    /*cout << "K=" << K << endl;
      cout << "R=" << R << endl;
      cout << "Kinv = " << Kinv << endl;
      cout << "KRKinv=" << KRKinv << endl;*/
    
    return Map(src, sink, KRKinv);
  } else{
    return -1;
  }
}

template <class InputStorageType, class OutputStorageType>
int HomographyMapping<InputStorageType, OutputStorageType>::
MapTrilinear(const Image<InputStorageType>& source, 
             Image<OutputStorageType>& sink,
             const HMatrix& H) {

#ifdef BIAS_DEBUG
  if (sink.IsEmpty()){
    BIASERR("sink is empty !");
    abort();
  }
  if (sink.GetChannelCount()!=source.GetChannelCount() ||
      sink.GetColorModel()!=source.GetColorModel() ||
      sink.GetStorageType()!=source.GetStorageType()) {
    BIASERR("different image types !");
    abort();
  }
  if (source.IsPlanar() && source.GetChannelCount()>1 ||
      sink.IsPlanar() && sink.GetChannelCount()>1) {
    BIASERR("not implemented for planar images !");
    abort();
  }

#endif
  Pyramid_.Init(source);
  const double log2inv = 1.0 / log(2.0);
  HomgPoint2D p_sink(0,0), p_source(0,0);
  p_sink[2] = 1;
  Matrix<double> Jacobian(2,2);  

  OutputStorageType** Data=sink.GetImageDataArray();
  // iterate over sink and calculate the corresponding pix in source
  for (unsigned int y=0; y<sink.GetHeight(); y++) {
    for (unsigned int x=0; x<sink.GetWidth(); x++) {
      // image point in homogenous coords
      p_sink[0] = x; 
      p_sink[1] = y;
      
      p_source = H * p_sink;
      p_source.Homogenize();

      //if (source.IsInROI(p_source[0],p_source[1]))
{
        H.GetJacobian(p_sink, Jacobian);
        const double 
          dx = sqrt(Jacobian[0][0]*Jacobian[0][0]+
                    Jacobian[1][0]*Jacobian[1][0]),
          dy = sqrt(Jacobian[0][1]*Jacobian[0][1]+
                    Jacobian[1][1]*Jacobian[1][1]);
        
        // use larger step to select pyramid level
        double scale = (dx > dy)? log(dx)*log2inv:log(dy)*log2inv;
        for (unsigned int j=0; j<source.GetChannelCount(); j++) {
          OutputStorageType tmp = (OutputStorageType)
            GetTrilinearImageValue_(p_source[0], p_source[1], scale, j,0);
          if (tmp > (OutputStorageType) 0)
            Data[y][x*sink.GetChannelCount()+j] =tmp;
        }
      }
    }
  }
#ifdef BIAS_DEBUG
  if (!RangeChecked_) {
    RangeChecked_ = true;
    BIASERR("You may have got artefacts in your image because overflow "
            <<"checking of your data types combination "
            <<typeid(InputStorageType).name()<<" -> "
            <<typeid(InputStorageType).name()<<" is not yet implemented."
            <<" Please do it in CheckValueRange_().");
  }
#endif
  return 0;
}

template <class InputStorageType, class OutputStorageType>
double HomographyMapping<InputStorageType, OutputStorageType>::
GetTrilinearImageValue_(const double& xsource, 
                        const double& ysource, 
                        const double& scale, 
                        unsigned int channel,
                        OutputStorageType defvalue/*=(OutputStorageType) 0*/) {
  double value = defvalue, value1 =  defvalue, value2 =  defvalue;
  double weight = scale - int(scale);
  int scale1 = int(scale);
  if (scale1>=(int)Pyramid_.size()) {
    scale1 = Pyramid_.size();
    weight = 0.0;
  } 
  if (scale1<0) {
    scale1 = 0;
    weight = 0.0;
  }

  // compute coordinates in smaller image
  double x1 = xsource, y1 = ysource;
  for (int s=0; s<scale1; s++) {
    // -0.5 necessary because of offset in 2^n downsampling
    x1 = (x1 - 0.5) * 0.5; 
    y1 = (y1 - 0.5) * 0.5;
  }
  const int scale2 = scale1 +1;
  const double x2 = (x1 - 0.5) * 0.5; 
  const double y2 = (y1 - 0.5) * 0.5;
  
  double weight1 = 1.0 - weight, weight2 = weight;

  // compute value at first scale
  if (x1 < (double)Pyramid_[scale1]->GetWidth()-2.0 && 
      y1 < (double)Pyramid_[scale1]->GetHeight()-2.0 &&
      x1 >= 1.0 && y1 >= 1.0 ) {
    value1 = Pyramid_[scale1]->BicubicInterpolation(x1, y1, channel);
   
  } else if (x1 < (double)Pyramid_[scale1]->GetWidth()-1.0 && 
             y1 < (double)Pyramid_[scale1]->GetHeight()-1.0 &&
             x1 > 0.0 && y1 > 0.0 ) {
    value1 = Pyramid_[scale1]->BilinearInterpolation(x1, y1, channel);

  } else {
    weight1 = 0;
  }
  // compute second scale if necessary
  if (weight>0.0) {
    if (x2 < (double)Pyramid_[scale2]->GetWidth()-2.0 && 
        y2 < (double)Pyramid_[scale2]->GetHeight()-2.0 &&
        x2 >= 1.0 && y2 >= 1.0 ) {
      value2 = Pyramid_[scale2]->BicubicInterpolation(x2, y2, channel);
    } else if (x2 < (double)Pyramid_[scale2]->GetWidth()-1.0 && 
               y2 < (double)Pyramid_[scale2]->GetHeight()-1.0 &&
               x2 > 0.0 && y2 > 0.0 ) {
      value2 = Pyramid_[scale2]->BilinearInterpolation(x2, y2, channel);
    } else {
      weight2 = 0;
    }
  }
  if ((weight = weight1 + weight2)<=DBL_EPSILON) return  defvalue;
  value = (weight1 * value1 + weight2 * value2) / weight;
  return CheckValueRange_(value);
}



template <>
int HomographyMapping<unsigned char, unsigned char>::
Map(Image<unsigned char> &source, Image<unsigned char> &dest, 
    const HMatrix& H) {
  int res=0;
#ifdef BIAS_DEBUG
  if (source.GetChannelCount() != 1){
    BIASERR("unable to map image with more than one channel "
	    <<source.GetChannelCount());
    return -2;
  }    
#endif

  if (!dest.IsEmpty())
    dest.Release();
    
  unsigned char **destida;
  HomgPoint2D dstpt, srcpt;

  if (!dest.IsEmpty()){
    if (!source.SamePixelAndChannelCount(dest)){
      dest.Release();
      dest.Init(source.GetWidth(), source.GetHeight(), 
                source.GetChannelCount());
    }
  } else {
    dest.Init(source.GetWidth(), source.GetHeight(), source.GetChannelCount());
  }
  
  destida=dest.GetImageDataArray();

  for (register unsigned int x=0; x<dest.GetWidth(); x++){
    for (register unsigned int y=0; y<dest.GetHeight(); y++){
      dstpt.Set((double)x,(double)y);
      srcpt= H *dstpt;
      srcpt.Homogenize();
      if (!srcpt.IsAtInfinity()){
        if (srcpt[0] > 0 && srcpt[0] < source.GetWidth()-1 &&
            srcpt[1] > 0 && srcpt[1] < source.GetHeight()-1){
          destida[y][x]=
            (unsigned char)rint(source.BilinearInterpolationGrey(srcpt[0], 
                                                                 srcpt[1]));
        }
      }
    } 
  }

  return res;
}


template <class InputStorageType, class OutputStorageType>
int HomographyMapping<InputStorageType, OutputStorageType>::
Map(Image<InputStorageType> &source, 
    Image<OutputStorageType> &dest, const HMatrix& H)
{
  int res=0;
#ifdef BIAS_DEBUG
  if (source.GetChannelCount() != 1){
    BIASERR("unable to map image with more than one channel "
	    <<source.GetChannelCount());
    return -2;
  }    
#endif

  if (!dest.IsEmpty())
    dest.Release();
    
  OutputStorageType **destida;
  HomgPoint2D dstpt, srcpt;

  if (!dest.IsEmpty()){
    if (!source.SamePixelAndChannelCount(dest)){
      dest.Release();
      dest.Init(source.GetWidth(), source.GetHeight(), 
                source.GetChannelCount());
    }
  } else {
    dest.Init(source.GetWidth(), source.GetHeight(), source.GetChannelCount());
  }
  
  destida = (OutputStorageType**) dest.GetImageDataArray();

  for (register unsigned int x=0; x<dest.GetWidth(); x++){
    for (register unsigned int y=0; y<dest.GetHeight(); y++){
      dstpt.Set((double)x,(double)y);
      srcpt= H *dstpt;
      srcpt.Homogenize();
      if (!srcpt.IsAtInfinity()){
        if (srcpt[0] > 0 && srcpt[0] < source.GetWidth()-1 &&
            srcpt[1] > 0 && srcpt[1] < source.GetHeight()-1){
          destida[y][x]=
            (OutputStorageType)(source.BilinearInterpolationGrey(srcpt[0], 
                                                                 srcpt[1]));
        }
      }
    }
  }

  return res;
}

// generic implementation does nothing
template <class InputStorageType, class OutputStorageType>
double HomographyMapping<InputStorageType, OutputStorageType>::
CheckValueRange_(const double& value) {
#ifdef BIAS_DEBUG
  RangeChecked_ = false;
#endif
  return value;
}

template <>
double HomographyMapping<unsigned char, unsigned char>::
CheckValueRange_(const double& value) {
  if (value<0.0)   return 0.0;
  if (value>255.0) return 255.0;
  return value;
}

template <>
double HomographyMapping<float, float>::
CheckValueRange_(const double& value) {
  return value;
}
