/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __HomographyMapping_hh__
#define __HomographyMapping_hh__

#include <Base/Image/Image.hh>
#include <Geometry/HMatrix.hh>
#include <Image/PyramidImage.hh>

namespace BIAS {

  /** @class HomographyMapping
      @ingroup g_filter
      @brief  Maps image src to image sink with homography H (software
      implementation)
    
      @todo check reimplementation, fast color image mapping
     
  */ 
  template <class InputStorageType, class OutputStorageType> 
  class HomographyMapping  {
  public:
  
    HomographyMapping() {
#ifdef BIAS_DEBUG
      RangeChecked_ = true;
#endif
    }

    /** @brief given a calibration and 3 rot angles, compute mapped image 
        Mainly computes the homography and calls map */
    int RotateToGrey(Image<InputStorageType>& src, 
                     Image<OutputStorageType>& sink,
                     float PhiX, float PhiY, float PhiZ, 
                     const Matrix3x3<double> &K);
  
    /** @brief map image avoiding aliasing with arbitrary homography
    
    This is almost correct regarding the sampling theorem, because it uses
    trilinear filtering over a pyramid. The local sampling frequency
    is estimated by analyzing the jacobian of the homography as a R^2->R^2
    mapping. This is very slow. If possible bicubic interpolation is done in
    the image, at borders (high pyramid levels) only bilinear.

    @author koeser 04/2005  */
    int MapTrilinear(const Image<InputStorageType>& src, 
                     Image<OutputStorageType>& sink,
                     const HMatrix& H);

    /** @brief fast mapping of grey images using bilinear interpolation 
        disregarding sampling frequency */
    int Map(Image<InputStorageType> &source, 
            Image<OutputStorageType> &dest, 
            const HMatrix& H);

  protected:
    double GetTrilinearImageValue_(const double& xsource, 
                                   const double& ysource, 
                                   const double& scale, 
                                   unsigned int channel,
                                   OutputStorageType defvalue/*=(OutputStorageType) 0*/);

    /** @brief implement 0<=x<=255 checks here */
    double CheckValueRange_(const double& value);

    PyramidImage<InputStorageType> Pyramid_;
#ifdef BIAS_DEBUG
    bool RangeChecked_;
#endif

  };

#ifdef WIN32

template <>
double HomographyMapping<float, float>::
CheckValueRange_(const double& value);

template <>
double HomographyMapping<unsigned char, unsigned char>::
CheckValueRange_(const double& value);

template <>
int HomographyMapping<unsigned char, unsigned char>::
Map(Image<unsigned char> &source, Image<unsigned char> &dest, 
    const HMatrix& H);
#endif


} // namespace 


#endif
