/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __MatlabTBUndistortion_hh__
#define __MatlabTBUndistortion_hh__

#include <Base/Image/Image.hh>
#include <Base/Geometry/HomgPoint2D.hh>
#include <Base/Geometry/KMatrix.hh>

#include <Image/CameraParam.hh>

#include <vector>
#include <string>

namespace BIAS{

  class MatlabTBUndistortion{

    //    static std::string CameraModelName[];
    // static std::string LensModelName[];

  public:
    MatlabTBUndistortion();
    MatlabTBUndistortion(const CameraParam &cp);

    ~MatlabTBUndistortion();

    // @brief default initialization
    int Init(const CameraParam &cp);

    inline bool IsInited() const {return IsInited_;};
    
    /** @brief True, if camera has a zoom lens.
        @author streckel 06/2004*/
    bool IsZoomCamera() const;

    /** @brief Set the current camera zoom, the focal length will be 
        calculated.
        @author buck */
    void SetZoom(const double abszoom);

    /** @brief get the current camera focal length
        @author buck */
    void SetFocalLength(const double f);

   /** @brief Set CCD skew
        @author streckel */
    void SetSkew(const double s);

    /** @brief Set the zoom and focal length of a zoom camera to the 
        default settings.
        @author streckel 06/2004*/
    void SetUnzoomed();

    /** @brief Set the current camera zoom, the focal length will be 
        calculated.
        @author buck */
    double GetZoom() const;

    /** @brief Set the current camera focal length, the zoom will be 
        calculated.
        @author buck */
    double GetFocalLength() const;

    /** @brief get camera image width.
        @author buck */
    unsigned int GetWidth() const; 
    /** @brief get camera image height.
        @author buck */
    unsigned int GetHeight() const;

    /** @brief get principal point, x
        @author buck */
    double GetPrincipalX() const; 
    /** @brief get principal point, y
        @author buck */
    double GetPrincipalY() const; 

    /** @brief Set active area to WidthxHeight+xoff+yoff.
        @author buck */
    int SetActiveArea(const float w,const float h, 
                      const float xoff, const float yoff);
 
    /** @brief Set subsampling, if required.
        @author buck */
    int SetSubsampling(const float subx, const float suby);

    /** @brief Get the Cameras KMatrix corresponding to the current 
        zoom setting
        @author streckel */
    BIAS::KMatrix GetKMatrix() const;

   /** @brief Using the Matlab camera calibration toolbox parameters 
        for an undistortion of distorted coordinates - correct implementation
        (Error in the Christian Buck implementation: Iteration does 
        not converge?!)
        @author streckel 06/2004 */
    void Undistort(BIAS::HomgPoint2D& point2d) const;
    void Undistort(double& origx,double& origy) const;

    /** @brief Using the Matlab camera calibration toolbox parameters 
        for a distortion of undistorted coordinates. Inverse to the 
        undistort function.
        @author streckel 06/2004 */
    void Distort(BIAS::HomgPoint2D& point2d) const;
    void Distort(double& origx,double& origy) const;

    
   /** @brief Using the Matlab camera calibration toolbox parameters 
        for an undistortion of a distorted camera image
        @author streckel 06/2004 */
    template <class StorageType>
    void DistortImage(const Image<StorageType>& src, 
                      Image<StorageType>& dst) const;

   /** @brief Using the Matlab camera calibration toolbox parameters 
        for a distortion of an already undistorted image. Inverse to the 
        UndistortImage function.
        @author streckel 06/2004 */
    template <class StorageType>
    void UndistortImage(const Image<StorageType>& src, 
                        Image<StorageType>& dst) const;

    int SetupProjectionLUT(const unsigned int& srcwidth,const 
                           unsigned int& srcheight,
                           const unsigned int &ChanCnt,
                           const unsigned int& SizeX, 
                           const unsigned int& SizeY);

    //template <class StorageType>
    //int UndistortFast(const Image<StorageType>& src,
 		//      Image<StorageType>& dst);
    int UndistortFast(const Image<unsigned char>& src,
		      Image<unsigned char>& dst);

  private:
	
    bool IsInited_;
      
    // do we have zoom or fixed optics? 
    bool zoomCamera_;

    // width and height of the camera image
    unsigned int width_,height_;
    
    // array of camera calibration constants for zoom cameras
    std::vector<CPDiscreteParam> knownparams_vect_;
    
    // filled by inheriting camera model
    double aspectratio_;
    double skew_;

    // Principal Point Coordinates (one for all zoom settings)
    double principalX_, principalY_;

    // calibration settings for the current zoom setting of a zoom camera
    // zoom and focallengthZoom are dependend:
    // if zoom is set, focallengthZoom will be calculated and vice versa
    double zoom_;
    double focallengthZoom_;
    double kc1Zoom_,kc2Zoom_,kc3Zoom_,kc4Zoom_;
        
    // standard settings for non zoom lenses or
    // for zoom default state
    double focallengthDef_;
    double kc1Def_,kc2Def_,kc3Def_,kc4Def_;
    /*
    // if only a portion of the CCD is used
    BIAS::Vector2<float> ActiveArea_; // width and height
    BIAS::Vector2<float> ActiveAreaOffset_; // ofset from top-left corner
    
    // if camera or frame grabber applies a subsampling
    BIAS::Vector2<float> Subsampled_;
    */
    // illumination vignette correction
    // x is distance in pixel from principal point
    // y is illumination correction factor
    std::vector<double> IlluCorrX_;
    std::vector<double> IlluCorrY_;

    double InterpolateFocalLength_(); // w/ given abszoom
    double InterpolateAbsZoom_(); // w/ given focal length
    void InterpolateDistortionFromIndex_(unsigned int i,double relnorm);

    // structure for Look-up-table for bilinear interpolation 
    // during backward warping
    struct LuTEntry {
      unsigned int offset[4];
      unsigned int  weight[2];
    };
    LuTEntry *LuT_;
    unsigned int CutOutWidth_, CutOutHeight_;
  };

};

#endif
