/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef WIN32
#  define _USE_MATH_DEFINES
#endif //WIN32

#include <math.h>
#include <Geometry/RMatrix.hh>
#include <Base/Geometry/Quaternion.hh>
#include <Base/Image/ImageConvert.hh>
#include <Base/Image/ImageDraw.hh>
#include <Base/Image/ImageIO.hh>

#include "SphericalUndistortion.hh"

#include <iostream>
#include <limits> // JW


using namespace BIAS;
using namespace std;

#define INST(type)\
\
template int SphericalUndistortion::ProjectImage(const Image<type>& src, Image<type>& dst, const double& SizeX, const double& SizeY, const double& FovX, const double& FovY, const double& CenterX, const double& CenterY, const int projectionType, const double& zoomFactor, KMatrix& KMatDst);\
template int SphericalUndistortion::ProjectImageCylindric(const Image<type>& src, Image<type>& dst, double viewCenterX, double viewCenterY, double zoomFactor);\
template int SphericalUndistortion::DrawCircles(Image<type>& src, const double& DegreeStep);\
template int SphericalUndistortion::CorrectVignette(Image<type>& src);\


INST(unsigned char);
INST(float);
#ifdef BUILD_IMAGE_INT
INST(int);
#endif
#ifdef BUILD_IMAGE_CHAR
INST(char);
#endif
#ifdef BUILD_IMAGE_SHORT
INST(short);
#endif
#ifdef BUILD_IMAGE_USHORT
INST(unsigned short);
#endif
#ifdef BUILD_IMAGE_UINT
INST(unsigned int);
#endif
#ifdef BUILD_IMAGE_DOUBLE
INST(double);
#endif


//#define LOCDEB



SphericalUndistortion::SphericalUndistortion()
{
 LuT_ = NULL;
  isInitialized_=false;
}
    
SphericalUndistortion::SphericalUndistortion(const SphericalUndistortion& sp) {
   LuT_ = NULL;
  AngleCorrX_ = sp.AngleCorrX_;
  AngleCorrY_ = sp.AngleCorrY_;
  LuT_ = NULL;
  IlluCorrX_ = sp.IlluCorrX_;
  IlluCorrY_ = sp.IlluCorrY_;
  
  interpolatorAngle_ = sp.interpolatorAngle_;
  interpolatorAngleBack_ = sp.interpolatorAngleBack_;
  interpolatorIllu_ = sp.interpolatorIllu_;
  
  MaxCamAngle_ = sp.MaxCamAngle_;
  CenterX_ = sp.CenterX_;
  CenterY_ = sp.CenterY_;
  radius_ = sp.radius_;
  isInitialized_ = sp.isInitialized_;
}

SphericalUndistortion::SphericalUndistortion(const double& centerX, 
                                             const double& centerY, 
                                             const double& radius,
                                             const vector<double> &AngleCorrX, 
                                             const vector<double> &AngleCorrY,
                                             const vector<double> &IlluCorrX, 
                                             const vector<double> &IlluCorrY) 
{
  LuT_ = NULL;
  Init(centerX, centerY, radius, AngleCorrX, AngleCorrY, IlluCorrX, IlluCorrY);
}

SphericalUndistortion::SphericalUndistortion(const CameraParam& cp) {
  LuT_ = NULL;
  Init(cp);
}

SphericalUndistortion::~SphericalUndistortion() {
  if (LuT_ !=NULL)  delete[] LuT_;
}

int SphericalUndistortion::Init(const CameraParam& cp) {
  if (cp.IsSpherical()) {
    return Init(cp.GetPrincipalX(), cp.GetPrincipalY(), cp.GetRadius(), 
                cp.GetAngleCorrX(), cp.GetAngleCorrY(),
                cp.GetIlluCorrX(), cp.GetIlluCorrY());
  } else {
    isInitialized_=false;
    return -1;
  }
}

int SphericalUndistortion::Init(const double& centerX, const double& centerY, 
                                const double& radius,
                                const std::vector<double> &AngleCorrX, 
                                const std::vector<double> &AngleCorrY,
                                const std::vector<double> &IlluCorrX, 
                                const std::vector<double> &IlluCorrY) {
  
  AngleCorrX_.clear();
  AngleCorrY_.clear();
  IlluCorrX_.clear();
  IlluCorrY_.clear();

  CenterX_ = centerX;
  CenterY_ = centerY;
  radius_ = radius;
  AngleCorrX_ = AngleCorrX;
  AngleCorrY_ = AngleCorrY;
  IlluCorrX_ = IlluCorrX;
  IlluCorrY_ = IlluCorrY;

  if (AngleCorrX_.size()<2 || AngleCorrY_.size()<2 || 
      AngleCorrX_.size()!=AngleCorrY_.size()) {
    isInitialized_=false;
    return -1;
  }

  double dum;
  interpolatorAngle_.SetKnotPoints(AngleCorrX_);  //t
  interpolatorAngle_.SetControlPoints(AngleCorrY_); //y = f(t)
  interpolatorAngle_.InitSpline(); 
  interpolatorAngle_.Spline(dum, AngleCorrX[0],4);  
  MaxCamAngle_ = AngleCorrX_[AngleCorrX_.size()-1];
  int lutsamples = int(radius_*100);
  interpolatorAngle_.PrepareLuTSpline(AngleCorrX[0],MaxCamAngle_,
				      lutsamples,4 );

  interpolatorAngleBack_.SetKnotPoints(AngleCorrY_);  
  interpolatorAngleBack_.SetControlPoints(AngleCorrX_);  
  interpolatorAngleBack_.InitSpline(); 
  interpolatorAngleBack_.Spline(dum, AngleCorrY[0],4);  
  interpolatorAngleBack_.PrepareLuTSpline(AngleCorrY[0],
					  AngleCorrY[AngleCorrY.size()-1], 
					  lutsamples,4 );

  if (IlluCorrX_.size()>1 && IlluCorrY_.size()>1 && 
      IlluCorrX_.size()==IlluCorrY_.size()) {
    interpolatorIllu_.SetKnotPoints(IlluCorrX_);  
    interpolatorIllu_.SetControlPoints(IlluCorrY_);
    interpolatorIllu_.InitSpline(); 
    interpolatorIllu_.Spline(dum, IlluCorrX[0],4);  
  } else {
    IlluCorrX_.clear();
    IlluCorrY_.clear();
  }

  isInitialized_=true;
  return 0;
}



int SphericalUndistortion::GetSphereCenter(double& centerX, double& centerY) const
{
  if (isInitialized_) {
    centerX = CenterX_;
    centerY = CenterY_;
    return 0;
  } else {
    return -1;
  }
}

int SphericalUndistortion::GetSphereCenter(double& centerX, double& centerY, double& radius) const
{
  if (isInitialized_) {
    centerX = CenterX_;
    centerY = CenterY_;
    radius = radius_;
    return 0;
  } else {
    return -1;
  }
}

double SphericalUndistortion::GetSphereRadius() const
{
  if (isInitialized_) {
    return radius_;
  } else {
    return -1;
  }
}


double SphericalUndistortion::GetMaxCamAngle() const
{
  if (isInitialized_) {
    return MaxCamAngle_;
  } else {
    return -1;
  }
}

KMatrix SphericalUndistortion::GetKMatrix() { 	 
  	 
   KMatrix K; 	 
   K.SetIdentity(); 	 
  	 
   if (!isInitialized_) { 	 
 #ifdef BIAS_DEBUG 	 
     BIASERR("Sphere Projections was not initialized!"); 	 
 #endif 	 
     return K; 	 
   } 	 
   double cx, cy; 	 
   GetSphereCenter(cx, cy); 	 
   double sx, sy; 	 
   TransfCoordSphere2Image( 80.0/180.0*M_PI, 0.0, sx, sy); 	 
   K[0][0] = (sx-cx)/tan(80.0/180.0*M_PI); 	 
   K[1][1] = (sx-cx)/tan(80.0/180.0*M_PI); 	 
   K[0][1] = 0; //skew=0! 	 
   K[0][2] = CenterX_; 	 
   K[1][2] = CenterY_; 	 
  	 
   return K; 	 
 }

template <class StorageType>
int SphericalUndistortion::ProjectImage(const Image<StorageType>& src, 
                                        Image<StorageType>& dst,
                                        const double& SizeX, 
                                        const double& SizeY, 
                                        const double& FovX, 
                                        const double& FovY,
                                        const double& CenterX, 
                                        const double& CenterY, 
                                        const int projectionType, 
                                        const double& zoomFactor,
                                        KMatrix& KMatDst)
{
  if (!isInitialized_) {
#ifdef BIAS_DEBUG
    BIASERR("Sphere Projections was not initialized!");
#endif
    return -1;
  }

  double viewCenterX = CenterX;
  double viewCenterY = CenterY;
  if (viewCenterX==0) viewCenterX = CenterX_;
  if (viewCenterY==0) viewCenterY = CenterY_;
  
  double imageSizeX = SizeX;
  double imageSizeY = SizeY;
  //cout << "ROI: " << startx << "," << starty 
  //     << "," << endx << "," << endy << endl;
  //cout << "viewCenter: " << viewCenterX << "," << viewCenterY << endl;

  double fovXrad, fovYrad;//, dum;
  if (FovX!=0) fovXrad = FovX/360.0*M_PI;
  else fovXrad = 0;
  
  if (FovY!=0) fovYrad = FovY/360.0*M_PI;
  else fovYrad = 0;

  //DrawCircles(src, 20.0);
  
  // either result image size or FoV is fixed
  KMatrix K;
  K.SetIdentity();
  if (fovXrad!=0 && imageSizeX!=0) {
    K[0][0] = (imageSizeX*zoomFactor)/(tan(fovXrad)*2.0);
  }
  if (fovYrad!=0 && imageSizeY!=0) {
    K[1][1] = (imageSizeY*zoomFactor)/(tan(fovYrad)*2.0);
  } 
  if (K[0][0]==1) {
    K[0][0]=K[1][1];
  }
  if (K[1][1]==1) {
    K[1][1]=K[0][0];
  }
  K[0][2] = (imageSizeX*zoomFactor)/2.0;
  K[1][2] = (imageSizeY*zoomFactor)/2.0;
  //cout << "K:" << K << endl;

  /*if (imageSizeX==0) { 
    //same scale in both directions
    //x-direction is responsible for scale
    HomgPoint2D sp1,sp2, dp1, dp2;
    sp1[0] = sp2[0] = (sizeX*zoomFactor)/2.0;
    sp1[1] = CenterY_-sizeY/2.0;
    sp2[1] = CenterY_+sizeY/2.0;
    sp1[2] = sp2[2] = 1;
    ProjectCoordBack(projectionType, K, sp1, dp1);
    ProjectCoordBack(projectionType, K, sp2, dp2);
    K[1][1] = K[0][0];
    ProjectCoord(projectionType, K, dp1, sp1);
    ProjectCoord(projectionType, K, dp2, sp2);
        imageSizeX = rint(sp2[1]-sp1[1]); 
      }
  if (imageSizeY==0) { 
    imageSizeY = 
      }
  */
  //dst.Release();
  if (dst.IsEmpty()) {
    dst.Init((unsigned int)(imageSizeX*zoomFactor), 
             (unsigned int)(imageSizeY*zoomFactor), 
             src.GetChannelCount());
  }
  //cout << (endx-startx) << "," <<  (endy-starty) << endl;
  StorageType** dataDst=dst.GetImageDataArray();
  
  //cout << "Roi for correct scale: " << startx << "," << starty 
  //     << "," << endx << "," << endy << endl;
  //cout << "K:" << K << endl;

  // go through pixels in dest image (backward mapping)
  for (int y=0; y<imageSizeY*zoomFactor; y++) {
    for (int x=0; x<imageSizeX*zoomFactor; x++) {

      //TransfRectImage2Sphere(x,y,projectionType, zoomFactor, theta, phi);

      double xSrc, ySrc;
      int ret=0; 
      HomgPoint2D Source, Dest;
      Source[0]=(double)x;
      Source[1]=(double)y;
      Source[2]=1;
      if (ProjectCoordBack(projectionType, K, Source, Dest)<0) {
        if (src.GetChannelCount() == 1) 
          dataDst[y][x] = 0;
        else
          for (unsigned int j=0; j<src.GetChannelCount(); j++) {
            dataDst[y][x * dst.GetChannelCount()+j] = 0;
          }
      } else {
        if ((viewCenterX!=0 && viewCenterX!=CenterX_) || 
            (viewCenterY!=0 && viewCenterY!=CenterY_)) {
          double t, p;
          //cout << "viewCenter: " << t << "," << p << endl;
          ret = TransfCoordImage2Sphere (Dest[0], Dest[1], t, p);
          ret = TransfCoordRotateSphere2Image(t, p, 
                                              viewCenterX, viewCenterY,
                                              xSrc, ySrc);
        } else {
          xSrc=Dest[0];
          ySrc=Dest[1];
        }
        
        if (ret==-1) {
          xSrc=ySrc=0;
        }
        
        if (xSrc<0 || xSrc>src.GetWidth()-1) xSrc=0;
        if (ySrc<0 || ySrc>src.GetHeight()-1) ySrc=0;
        if (src.GetChannelCount() == 1) 
          dataDst[y][x] = 
            (StorageType) src.FastBilinearInterpolationGrey(xSrc, ySrc);
        else
          for (unsigned int j=0; j<src.GetChannelCount(); j++) {
            dataDst[y][x * dst.GetChannelCount()+j] = 
              (StorageType) src.BilinearInterpolation(xSrc, ySrc,j);
          }
      }
    }
  }
  
  KMatDst=K;

  return 0;
}

template <class StorageType>
int SphericalUndistortion::ProjectImageCylindric(const Image<StorageType>& src,
                                                 Image<StorageType>& dst,
                                                 double viewCenterX,
                                                 double viewCenterY,
                                                 double zoomFactor)
{
  if (!isInitialized_) {
#ifdef BIAS_DEBUG
    BIASERR("Sphere Projections was not initialized!");
#endif
    return -1;
  }

  if  (zoomFactor<0.01) zoomFactor=1;
  if (viewCenterX<CenterX_-radius_ || viewCenterX>CenterX_+radius_ ||
      viewCenterY<CenterY_-radius_ || viewCenterY>CenterY_+radius_) {
    viewCenterX = CenterX_;
    viewCenterY = CenterY_;
  }

  double imageSizeX = 2 * M_PI * radius_;
  double imageSizeY = radius_;

  //DrawCircles(src, 20.0);
  
  // either result image size or FoV is fixed
  if (dst.IsEmpty()) {
    dst.Init((unsigned int)(imageSizeX*zoomFactor), 
             (unsigned int)(imageSizeY*zoomFactor), 
             src.GetChannelCount());
  }

  StorageType** dataDst=dst.GetImageDataArray();
  
  // go through pixels in dest image (backward mapping)
  for (int y=0; y<imageSizeY*zoomFactor; y++) {
    for (int x=0; x<imageSizeX*zoomFactor; x++) {

      double theta, phi;
      double xSrc, ySrc;

      theta = y/(imageSizeY*zoomFactor)*MaxCamAngle_;
      phi = (imageSizeX*zoomFactor-x)/(imageSizeX*zoomFactor)*2*M_PI-M_PI;

      //TransfCoordSphere2Image(theta, phi, xSrc, ySrc);
      TransfCoordRotateSphere2Image(theta, phi, 
                                    viewCenterX, viewCenterY,
                                    xSrc, ySrc);

      if (xSrc<0 || xSrc>src.GetWidth()-1) xSrc=0;
      if (ySrc<0 || ySrc>src.GetHeight()-1) ySrc=0;
      if (src.GetChannelCount() == 1) 
        dataDst[y][x] = 
          (StorageType) src.FastBilinearInterpolationGrey(xSrc, ySrc);
      else
        for (unsigned int j=0; j<src.GetChannelCount(); j++) {
          dataDst[y][x * dst.GetChannelCount()+j] = 
            (StorageType) src.BilinearInterpolation(xSrc, ySrc,j);
        }
    }
  }
    
  return 0;
}



int SphericalUndistortion::ProjectCoord(const int projectionType, 
                                    const KMatrix& kMatrix,
                                    const HomgPoint2D& Source,
                                    HomgPoint2D &Dest)
{

  if (!isInitialized_) {
#ifdef BIAS_DEBUG
    BIASERR("Sphere Projections was not initialized!");
#endif
    return -1;
  }
         
  
  double theta,phi;
  if (TransfCoordImage2Sphere (Source, theta, phi)!=0) {
    return -1;
  }
  //cout << "ProjectCoord Src:" << Source << "," << theta 
  //<< "," << phi << endl;
 
  double rDst;
  switch (projectionType) {
  case (PERSPECTIVE):
    rDst = tan(theta);
    break; 
  case (STEREOGRAPHIC):
    rDst = tan(theta/2);
    break; 
  case (EQUIDISTANT):
    rDst = theta;
    break; 
  case (SINE_LAW):
    rDst = sin(theta);
    break; 
  case (EQUI_SOLID_ANGLE):
    rDst = sin(theta/2);
    break; 
  default:
#ifdef BIAS_DEBUG
    BIASERR("Invalid projection type!");
#endif
    return -1;
  }

  Dest.Set(rDst * cos(phi), rDst * sin(phi));
  //cout << "ProjectCoord Src:" << rDst << "," << Dest << endl;
  Dest = kMatrix * Dest;
  //cout << "ProjectCoord Src:" << rDst << "," << Dest << endl;

  return 0;
}

int SphericalUndistortion::ProjectCoordBack(const int projectionType, 
                                        const KMatrix& kMatrix,
                                        const HomgPoint2D& Source,
                                        HomgPoint2D &Dest)
{
  
  if (!isInitialized_) {
#ifdef BIAS_DEBUG
    BIASERR("Sphere Projections was not initialized!");
#endif
    return -1;
  }
  
  HomgPoint2D SourceK;
  SourceK = kMatrix.Invert()  * Source;
  
  //cout << "K" << kMatrix <<endl;
  //cout << "Source" << Source << "," << SourceK << endl;
  // distance of dst-point to center
  
  double rDst = sqrt(SourceK[0]*SourceK[0]+SourceK[1]*SourceK[1]);
  //cout << "rDst" << rDst << "," << SourceK << endl;
  // phi: angle between positive x-axis and dst-point
  double phi, theta;
  phi = atan2(SourceK[1],SourceK[0]);
  
  // theta: calibrated angle between optical axis and source image point
  switch (projectionType) {
  case (PERSPECTIVE):
    theta = atan(rDst);
    break; 
  case (STEREOGRAPHIC):
    //rDst = sqrt(Source[0]*Source[0]+Source[1]*Source[1]);    
    theta = 2.0*atan(rDst);
    break; 
  case (EQUIDISTANT):
    {double FoV_X = atan(kMatrix[0][2]/kMatrix[0][0]);
    rDst = sqrt((Source[0]-kMatrix[0][2])*(Source[0]-kMatrix[0][2])+
                (Source[1]-kMatrix[1][2])*(Source[1]-kMatrix[1][2]));
    double rDstMax_X = sqrt(kMatrix[0][2]*kMatrix[0][2]);
    //cout <<FoV_X << endl;
    //cout <<rDst << endl;
    //cout <<rDstMax_X << endl;
    theta = rDst/rDstMax_X*FoV_X;}
    //cout <<"theta"<<theta << endl;
    break; 
  case (SINE_LAW):
    if (rDst>1) {
      theta = M_PI;
    } else { 
      theta = asin(rDst);
    }
    break; 
  case (EQUI_SOLID_ANGLE):
    if (rDst>1) {
      theta = M_PI;
    } else { 
      theta = 2*asin(rDst);
    }
    break; 
  default:
#ifdef BIAS_DEBUG
    BIASERR("Invalid projection type!");
#endif
    return -1;
  }
  if (theta>MaxCamAngle_) {
    return -2;
  }
  
  TransfCoordSphere2Image(theta, phi, Dest[0], Dest[1]);
  Dest[2]=1;
  //cout << Source << "," << Dest << "," << theta << "," << phi << endl;
  return 0;
  
}

template <class StorageType>
int SphericalUndistortion::DrawCircles(Image<StorageType>& src, 
                                       const double& DegreeStep) {

  if (!isInitialized_) {
#ifdef BIAS_DEBUG
    BIASERR("Sphere Projections was not initialized!");
#endif
    return -1;
  }

  StorageType* Color;
  Color = new StorageType[src.GetChannelCount()];
  Color[0]=127;
  for (unsigned int j=1; j<src.GetChannelCount(); j++) {
    Color[j]=0;
  }
  for (double theta=DegreeStep/180.0*M_PI; theta<=MaxCamAngle_; theta+=DegreeStep/180.0*M_PI) {
    double CalMaxCamAngle;
    double CalTheta;
    // calculate calibrated angles
    interpolatorAngle_.SplineFromLuT(CalMaxCamAngle, MaxCamAngle_);  
    interpolatorAngle_.SplineFromLuT(CalTheta, theta);  
    // apply calculate calibrated radius from calibrated angles
    double rSrc = radius_ * CalTheta/CalMaxCamAngle;
    
    if (ImageDraw<StorageType>::CircleCenter(src, (int)CenterX_, (int)CenterY_, (int)rSrc, Color)!=0){
      delete [] Color;
      return -1;
    } 
  }
  delete [] Color;
  return 0;  
}


int SphericalUndistortion::TransfCoordImage2Sphere (const HomgPoint2D& Source, 
                                                double &theta, 
                                                double &phi) {
  double SourceX, SourceY;
  SourceX = Source[0]/Source[2];
  SourceY = Source[1]/Source[2];
  theta = phi = 0;

  return TransfCoordImage2Sphere (SourceX, SourceY, theta, phi);

}


int SphericalUndistortion::TransfCoordImage2Sphere(const double& SourceX, 
                                               const double& SourceY,
                                               double &theta, 
                                               double &phi) {
  if (!isInitialized_) {
#ifdef BIAS_DEBUG
    BIASERR("Sphere Projections was not initialized!");
#endif
    return -1;
  }

  theta = phi = 0;

  double rSrc = sqrt((SourceX-CenterX_)*(SourceX-CenterX_)+(SourceY-CenterY_)*(SourceY-CenterY_));

  if (rSrc>radius_) {
#ifdef BIAS_DEBUG
    //BIASERR("point not in circle image!");
#endif
    return -1;
  }
//   cout << CenterX_ << " " <<  CenterY_ << " " 
//   <<  MaxCamAngle_ << " " << radius_ << endl;
  phi = atan2(SourceY-CenterY_,SourceX-CenterX_);
  
  double CalMaxCamAngle=0;
  // calculate calibrated maximum angle
  //interpolatorAngle_.Spline(CalMaxCamAngle, MaxCamAngle_, 4);  
  interpolatorAngle_.SplineFromLuT(CalMaxCamAngle, MaxCamAngle_);  

  // calculate calibrated maximum radius
  double RadiusCalMax = MaxCamAngle_/CalMaxCamAngle*radius_;
  // calculate calibrated theta and inverse
  double CalTheta = rSrc/RadiusCalMax*MaxCamAngle_;
  //  interpolatorAngleBack_.Spline(theta, CalTheta,4); 
  interpolatorAngleBack_.SplineFromLuT(theta, CalTheta); 

  return 0;
}

int SphericalUndistortion::TransfCoordRotateSphere2Image(double theta, 
                                                     double phi,
                                                     const double& viewCenterX, 
                                                     const double& viewCenterY,
                                                     HomgPoint2D &Source)
{
  double sx, sy;
  sx=sy=0;
  if (TransfCoordRotateSphere2Image(theta, phi, 
                                    viewCenterX, viewCenterY,
                                    sx, sy)<0) {
    return -1;
  } else {
    Source[0] = sx;
    Source[1] = sy;
    Source[2] = 1;
    return 0;
  }
}

int 
SphericalUndistortion::TransfCoordRotateSphere2Image(double theta, 
                                                     double phi,
                                                     const double& viewCenterX,
                                                     const double& viewCenterY,
                                                     double &SourceX, 
                                                     double &SourceY)
{
  if (!isInitialized_) {
#ifdef BIAS_DEBUG
    BIASERR("Sphere Projections was not initialized!");
#endif
    return -1;
  }
  
  SourceX = SourceY = 0;
  double viewTheta, viewPhi;
  if (TransfCoordImage2Sphere(viewCenterX, viewCenterY, 
                              viewTheta, viewPhi)!=0) { 
    return -1;
  }
  // avoid z-rotation by changing the rotation axes
  Vector3<double> vecPoint(1.0,phi,theta);
  Vector3<double> vecView(1.0,viewPhi,viewTheta);
  //Vector3<double> vecCenter(0,0,1);
  Vector3<double> vecAxis(cos(viewPhi+M_PI/2),sin(viewPhi+M_PI/2),0);
  Vector3<double> vecRot(0,0,0);
  vecPoint = vecPoint.CoordSphereToEuclidian();
  vecView = vecView.CoordSphereToEuclidian();
  /*
  // With 1 rotation and moving axis
  RMatrix rMat;
	rMat.Set(vecAxis, viewTheta);
  double x,y,z;
  rMat.GetRotationAnglesZYX(x,y,z);
  vecRot = rMat*vecPoint;
  //cout << x << "," << y << "," << z << endl;
  */

  //With Quaternions
  Quaternion<double> rot;
  rot.SetValueAsAxisRad (vecAxis, viewTheta);
  rot.MultVec(vecPoint, vecRot);
  
  /* 
  //With 2 rotations alpha, beta (wrong rotation!!)
  // viewAlpha: angle between x-z-Plane and vecView
  double viewAlpha = -1 * atan2(vecView[1],vecView[2]);
  // viewBeta: angle between y-z-Plane and vecView
  double viewBeta = atan2(vecView[0],vecView[2]); 
  RMatrix rMat;
  rMat.SetZYX(viewAlpha,viewBeta,0);
  vecRot = rMat*vecPoint;
  */

  vecRot = vecRot.CoordEuclidianToSphere();
  phi = vecRot[1];
  theta = vecRot[2];

  if (theta > MaxCamAngle_) return -1;
  if (phi>M_PI) phi-=2*M_PI;

  double CalMaxCamAngle=0;
  double CalTheta=0;
  // calculate calibrated angles
  interpolatorAngle_.SplineFromLuT(CalMaxCamAngle, MaxCamAngle_);  
  interpolatorAngle_.SplineFromLuT(CalTheta, theta);  
  // apply calculate calibrated radius from calibrated angles
  double rSrc = radius_ * CalTheta/CalMaxCamAngle;
 
  SourceX = rSrc * cos(phi) + CenterX_;
  SourceY = rSrc * sin(phi) + CenterY_;

  return 0;
}


int SphericalUndistortion::TransfCoordSphere2Image(const double& theta, 
                                                   const double& phi, 
                                                   HomgPoint2D &Source)
{
  double sx, sy;
  sx=sy=0;
  if (TransfCoordSphere2Image(theta, phi, sx, sy)<0) {
    return -1;
  } else {
    Source[0] = sx;
    Source[1] = sy;
    Source[2] = 1;
    return 0;
  }
}


int SphericalUndistortion::TransfCoordSphere2Image(const double &theta,
                                                   const double &phi, 
                                                   double &SourceX, 
                                                   double &SourceY) {
  
  if (!isInitialized_) {
#ifdef BIAS_DEBUG
    BIASERR("Sphere Projections was not initialized!");
#endif
    return -1;
  }
  
  SourceX = SourceY = 0;
  
  if (theta<0 || theta>MaxCamAngle_) {
#ifdef BIAS_DEBUG
    BIASERR("Angle theta out of range! theta="<<theta);
#endif
    return -1;
  }

  if (phi<-M_PI || phi>M_PI) {
#ifdef BIAS_DEBUG
    BIASERR("Angle phi out of range! phi="<<phi);
#endif
    return -1;
  }

  double CalMaxCamAngle=0;
  double CalTheta=0;
  // calculate calibrated angles
  interpolatorAngle_.SplineFromLuT(CalMaxCamAngle, MaxCamAngle_);  
  interpolatorAngle_.SplineFromLuT(CalTheta, theta);  
  // apply calculate calibrated radius from calibrated angles
  double rSrc = radius_ * CalTheta/CalMaxCamAngle;
  //cout << CalMaxCamAngle<< ","<< CalTheta << "," << rSrc << endl;
  SourceX = rSrc * cos(phi) + CenterX_;
  SourceY = rSrc * sin(phi) + CenterY_;

  return 0;
}

int SphericalUndistortion::TransfCoordSphere2CalibImage(const double &theta, 
                                                        const double &phi,
                                                        HomgPoint2D &Source)
  const {
  double sx, sy;
  sx=sy=0;
  if (TransfCoordSphere2CalibImage(theta, phi, sx, sy)<0) {
    return -1;
  } else {
    Source[0] = sx;
    Source[1] = sy;
    Source[2] = 1;
    return 0;
  }
}

int SphericalUndistortion::EnforceNormalRange(double& theta, 
                                              double& phi) const {

  // theta and phi are periodic in 2*M_PI, if they are out of a suitable
  // range, try to find best equivalent angle by adding 2*l*M_PI

  if (theta<0.0) {
    const int count = (int) ((double)theta / (M_PI*(-2.0)));
    theta += ((double)count+1.0) * M_PI*2.0;
    if (theta>MaxCamAngle_) {
      BIASERR("Angle theta out of range!"<<theta);
      return -1;
    } 
  } else if (theta>MaxCamAngle_) {
    const int count = (int) ((double)theta / (M_PI*(2.0)));
    theta -= ((double)count+1.0) * M_PI*2.0;
    if (theta<0.0) {
      BIASERR("Angle theta out of range: "<<theta);
      return -2;
    } 
  }
 
  if (phi<-M_PI) {
    const int count = (int) ((double)phi / (M_PI*(-2.0)));
    phi += ((double)count+1.0) * M_PI*2.0;
    if (phi>M_PI) {
      BIASERR("Angle phi out of range:"<<phi);
      return -3;
    } 
  } else if (phi>M_PI) {
    const int count = (int)((double)phi / (M_PI*(2.0)));
    phi -= ((double)count+1.0) * M_PI*2.0;
    if (phi<-M_PI) {
      BIASERR("Angle phi out of range: "<<phi);
      return -4;
    } 
  }
  return 0;
}

int SphericalUndistortion::TransfCoordSphere2CalibImage(const double& theta, 
                                                        const double& phi,
                                                        double &SourceX, 
                                                        double &SourceY) const{
  
  if (!isInitialized_) {
#ifdef BIAS_DEBUG
    BIASERR("Sphere Projections was not initialized!");
#endif
    return -1;
  }
  
  SourceX = SourceY = 0;
  
  if (theta<0 || theta>MaxCamAngle_) {
#ifdef BIAS_DEBUG
    //BIASERR("Angle theta out of range!");
#endif
    return -1;
  }

  if (phi<-M_PI || phi>M_PI) {
#ifdef BIAS_DEBUG
    //BIASERR("Angle phi out of range!");
#endif
    return -1;
  }

  // apply radius from calibrated angles
  double rSrc = radius_ * theta/MaxCamAngle_;

  SourceX = rSrc * cos(phi) + CenterX_;
  SourceY = rSrc * sin(phi) + CenterY_;

  return 0;

}


// JW
#ifdef WIN32
# ifdef max
#   undef max
#  endif // max
#endif //WIN32


template <class StorageType>
int SphericalUndistortion::CorrectVignette(Image<StorageType>& src) {

  if (!isInitialized_) {
#ifdef BIAS_DEBUG
    BIASERR("Sphere Projections was not initialized!");
#endif
    return -1;
  }
  if (IlluCorrX_.size()<2 || IlluCorrY_.size()<2) return 0;
  
  StorageType** Data=src.GetImageDataArray();

  const StorageType maxVal = (StorageType) ::std::numeric_limits<StorageType>::max(); // JW
  double illuFac=0;
  double theta=0, phi=0;
  for (unsigned int y=0; y<src.GetHeight(); y++) {
    for (unsigned int x=0; x<src.GetWidth(); x++) {
      HomgPoint2D hp(x,y);
      TransfCoordImage2Sphere(hp, theta, phi);
      for (unsigned int j=0; j<src.GetChannelCount(); j++) {
        interpolatorIllu_.Spline(illuFac, theta,4);
		// clip overflow range: JW
		// JW TODO: liek this: if ((double)Data[y][x*src.GetChannelCount()+j]) / (double)maxVal >= illuFac ) {
		if (Data[y][x*src.GetChannelCount()+j]/illuFac > maxVal) {
          Data[y][x*src.GetChannelCount()+j] = maxVal;
        } else {
          Data[y][x*src.GetChannelCount()+j] = 
            StorageType((float)Data[y][x*src.GetChannelCount()+j]/illuFac);
        }
      }
    }
  }
  return 0;
}


int SphericalUndistortion::
SetupProjectionLUT(const unsigned int& srcwidth,const unsigned int& srcheight,
		   const unsigned int &ChanCnt,
		   const unsigned int& SizeX, const unsigned int& SizeY,
		   const double& CenterX, const double& CenterY, 
		   const int projectionType, KMatrix& K)
{
  if (!isInitialized_) {
#ifdef BIAS_DEBUG
    BIASERR("Sphere Projections was not initialized!");
#endif
    return -1;
  }
  CutOutWidth_ = SizeX;
  CutOutHeight_ = SizeY;
  double viewCenterX = CenterX;
  double viewCenterY = CenterY;
  if (viewCenterX==0) viewCenterX = CenterX_;
  if (viewCenterY==0) viewCenterY = CenterY_;
  
  double theta1,theta2,phi,fovx,fovy; //fov is half field of view!
  double dSX = double(SizeX)/2.0;
  double dSY = double(SizeY)/2.0;
  TransfCoordImage2Sphere(viewCenterX - dSX,viewCenterY ,theta1,phi);
  TransfCoordImage2Sphere( viewCenterX + dSX,viewCenterY ,theta2,phi);
  fovx = theta2+theta1;
  cout<<"Theta1: "<<theta1<<" theta2: "<<theta2<<" FoVX:"
      <<fovx * 180.0 / M_PI<<endl;

  TransfCoordImage2Sphere(viewCenterX,viewCenterY - dSY,theta1,phi);
  TransfCoordImage2Sphere(viewCenterX,viewCenterY + dSY,theta2,phi);
  fovy = theta2+theta1;
  cout<<"Theta1: "<<theta1<<" theta2: "<<theta2<<" FoVY:"
      <<fovy * 180.0 / M_PI<<endl;
  K.SetIdentity();
  K[0][0] = dSX / tan(fovx/2.0);
  K[1][1] = dSY / tan(fovy/2.0);
  K[0][2] = dSX;  K[1][2] = dSY;

  if (LuT_ !=NULL)  delete[] LuT_;
  LuT_ = new LuTEntry[SizeX * SizeY];
  LuTEntry *currLE = LuT_;

  // go through pixels in dest image (backward mapping)
  for (unsigned int y=0; y<SizeY; y++) {
    for (unsigned int x=0; x<SizeX; x++) {
      double xSrc, ySrc;
      int ret=0; 
      HomgPoint2D Source, Dest;
      Source[0]=(double)x;
      Source[1]=(double)y;
      Source[2]=1;
      ProjectCoordBack(projectionType, K, Source, Dest);
      if ((viewCenterX!=0 && viewCenterX!=CenterX_) || 
          (viewCenterY!=0 && viewCenterY!=CenterY_)) {
        double t, p;
        ret = TransfCoordImage2Sphere (Dest[0], Dest[1], t, p);
        ret = TransfCoordRotateSphere2Image(t, p, 
                                            viewCenterX, viewCenterY,
                                            xSrc, ySrc);
      } else {
        xSrc=Dest[0];
        ySrc=Dest[1];
      }
      
      if (ret==-1) {
        xSrc=ySrc=0;
	abort();
      }
      if (xSrc<0 || xSrc>srcwidth-1) xSrc=0;
      if (ySrc<0 || ySrc>srcheight-1) ySrc=0;
      
      //xSrc,ySrc are the coordinates in the original image typically between 
      // four pixels, so find the neighbours for bilinear interpolation
      unsigned int x_floor = (unsigned int)floor(xSrc);
      unsigned int y_floor = (unsigned int)floor(ySrc);
      unsigned int x_ceil = x_floor+1;
      unsigned int y_ceil = y_floor+1;
      currLE->weight[0] = (unsigned int)rint((xSrc - x_floor)*256);
      currLE->weight[1] = (unsigned int)rint((ySrc - y_floor)*256);
//       currLE->weight[0] = xSrc - x_floor;
//       currLE->weight[1] = ySrc - y_floor;

      // calculate the offsetes in the image data array for all four corners
      // but only for the first channel of ChanCnt
      currLE->offset[0] = (y_floor * srcwidth + x_floor) * ChanCnt;
      currLE->offset[1] = (y_ceil * srcwidth + x_floor) * ChanCnt;
      currLE->offset[2] = (y_ceil * srcwidth + x_ceil) * ChanCnt;
      currLE->offset[3] = (y_floor * srcwidth + x_ceil) * ChanCnt;
      currLE++;
    }
  }
  return 0;
}


//template <> int SphericalUndistortion::
int SphericalUndistortion::
ProjectCutOut(const Image<unsigned char>& src,Image<unsigned char>& dst)
{
  BIASASSERT(!dst.IsEmpty());

  unsigned char *srcp, *dstp;
  srcp = src.GetImageData();
  dstp = dst.GetImageData();
  if (src.GetChannelCount() ==1) {
    for ( unsigned int i =0; i<dst.GetPixelCount(); i++ ) {
      unsigned char ul = *(srcp + LuT_[i].offset[0]); //ida[y_floor][x_floor];
      unsigned char ll = *(srcp + LuT_[i].offset[1]);// ida[y_ceil][x_floor];
      unsigned char lr = *(srcp + LuT_[i].offset[2]);// ida[y_ceil][x_ceil];   
      unsigned char ur = *(srcp + LuT_[i].offset[3]); // ida[y_floor][x_ceil];
//       unsigned int dx = LuT_[i].weight[0];
//       unsigned int dy = LuT_[i].weight[1];
//       *dstp =  (unsigned char)(((ul<<16) + (dy<<8)*(ll - ul) + (dx<<8)*(ur - ul) + dy*dx*(ul - ll - ur + lr))>>16);

//       float dx = (float)LuT_[i].weight[0] / 256.0;
//       float dy = (float)LuT_[i].weight[1] / 256.0;
//       float top = ((float)ul * (1.0-dx)) +(ur * dx); 
//       float bottom = ((float)ll * (1.0-dx)) +(lr * dx);
//       *dstp = (unsigned char)rint(top * (1.0-dy) + bottom * dy);
      *dstp = (ul + ur + ll + lr)/4;
      dstp++;
    }
  }
  else {
    unsigned int cc = src.GetChannelCount();
    for ( unsigned int i =0; i<dst.GetPixelCount(); i++ ) {
      for (unsigned int c=0; c<cc; c++) {
	unsigned char ul = *(srcp + LuT_[i].offset[0] + c); 
	unsigned char ll = *(srcp + LuT_[i].offset[1] + c);
	unsigned char lr = *(srcp + LuT_[i].offset[2] + c);
	unsigned char ur = *(srcp + LuT_[i].offset[3] + c); 
	unsigned int dx = LuT_[i].weight[0];
	unsigned int dy = LuT_[i].weight[1];
	*dstp =  (unsigned char)(((ul<<16) + (dy<<8)*(ll - ul) + (dx<<8)*(ur - ul) + dy*dx*(ul - ll - ur + lr))>>16);
	dstp++;
      }
    }


  }
  return 0;
}

