/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004, 2005    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef _BIAS_BSPLINECURVE_
#define _BIAS_BSPLINECURVE_

#include <cmath>
#include <vector>
#include "Base/Debug/Debug.hh"
#include "Base/Image/Image.hh"
#include "Base/Image/ImageDraw.hh"
#include "Base/Math/Matrix.hh"
#include "Base/Math/Vector.hh"
#include "Base/Math/Vector2.hh"



namespace BIAS {
/** @class BSplineCurve
    @ingroup g_mathalgo
    @brief this class is intended for drawing  of quadratic and cubic 
    regular B-Splines from given control points. 
 
    this class is intended for drawing of quadratic and cubic regular 
    B-Splines from given control points. The Spline doesnt interpolate through 
    the control points!!!

    The Implementation of BSplineCurve is faster then BSpline. If you only want
    to draw a BSpline you should use this implementation.

    usage:
    -# void SetControlPoints(&cPnts)
       for quadratic closed Bspline set cPnt[n],cPnt[0],...,cPnt[n],cPnt[0] 
       for cubic closed Bspline set cPnt[n],cPnt[0],...,cPnt[n],cPnt[0],cPnt[1]
    -# void InitXXXUniformBSpline() - computes coefficientvectors for all knots
    -# void XXXUniformSpline(Vector2<double> &res, double t ) - computes 
       point on the curve (0<=t<=1) 
       t=0 returns the first computed knot of the first segment
       t=1 returns the last computed knot of the last segment

    see also 'Computer Graphics & Geometric Modeling' from David Salomon 
    chapter 4.16

    @author Marcel Lilienthal, Mai 2005
*/
 
class BSplineCurve : public Debug 
{

 public:
  BSplineCurve();
  ~BSplineCurve() {};
  
  /** 
      set the control points, which control the interpolating curve
   */
  inline void SetControlPoints(std::vector<BIAS::Vector2<double> > &cPnt){
    cPnts_=cPnt;
#ifdef BIAS_DEBUG
    initializedCubic_=false;
    initializedQuadratic_=false;
#endif
  };

  /**
     @brief computes coefficients for all segments of a quadratic uniform
     B-spline
     
     this function computes the coefficients for all sements of a quadratic
     uniform B-spline and saves these coefficients in protected vectors 
     xCoeff_, yCoeff_
     method is discribed in David Salomons "Computer Graphic & Geometric 
     Modelling" chapter 4.16.2
     
     @author Marcel Lilienthal, Mai 2005
   */
  void InitQuadraticUniformBSpline();

  /**
     @brief computes coefficients for all segments of a cubic uniform B-spline
     
     this function computes the coefficients for all sements of a cubic
     uniform B-spline and saves these coefficients in protected vectors 
     xCoeff_, yCoeff_.
     method is discribed in David Salomons "Computer Graphic & Geometric 
     Modelling" chapter 4.16.4
     
     @author Marcel Lilienthal, Mai 2005
  */
  void InitCubicUniformBSpline();

  /**
     @brief computes the point on a quadratic uniform B-spline curve at time t
     
     this function computes the point on a quadratic uniform B-spline curve at 
     time t. method is discribed in David Salomons "Computer Graphic & 
     Geometric Modelling" chapter 4.16.2.
     the value of t must be between 0 and 1.
     t is used to compute the corresponding segment and the t_ for the result 
     segment.
     
     @param  res - reference to the result (return value)
     @param  t - time value of which point is to be calculated 
     @author Marcel Lilienthal, Mai 2005
   */
  void QuadraticUniformBSpline(Vector2<double> &res, double t);

  /**
     @brief computes the point on a cubic uniform B-spline curve at time t
     
     this function computes the point on a cubic uniform B-spline curve at time
     t. method is discribed in David Salomons "Computer Graphic & Geometric
     Modelling" chapter 4.16.2.
     the value of t must be between 0 and 1.
     t is used to compute the corresponding segment and the t_ for the result 
     segment.
     
     @param  res - reference to the result (return value)
     @param  t - time value of which point is to be calculated 
     @author Marcel Lilienthal, Mai 2005
   */
  void CubicUniformBSpline(Vector2<double> &res, double t);
  

 protected:
  
#ifdef BIAS_DEBUG
  bool initializedQuadratic_;
  bool initializedCubic_;
#endif
  std::vector<BIAS::Vector2<double> > cPnts_;//control Points
  std::vector<BIAS::Vector<double> > xCoeff_;//coefficient x Vec
  std::vector<BIAS::Vector<double> > yCoeff_;//coefficient y Vec

};//class

}//namespace
#endif
