/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "Minpack.hh"
#include <Base/Math/tnt/fortran.h>

#ifdef __cplusplus 
extern "C" 
{
#endif
  /* hybrd and hybrd1 solve a system of n non-linear equations using the
     powell hybrid method. Jacobian is aproximated numerical */
  void hybrd_(void (*fcn)(int *n, double *x, double *fvec, int *iflag), 
              cfi_ n, fda_ x, fda_ fvec, cfd_ xtol, cfi_ maxfev,
              cfi_ ml, cfi_ mu, cfd_ epsfcn, fda_ diag, cfi_ mode, cfd_ factor,
              cfi_ nprint, fi_ info, fi_ nfev, fda_ fjac, cfi_ ldfjac, fda_ r,
              cfi_ lr, fda_ qtf, fda_ wa1, fda_ wa2, fda_ wa3, fda_ wa4);

  void hybrd1_(void (*fcn)(int *n, double *x, double *fvec, int *iflag), 
               cfi_ n, fda_ x, fda_ fvec, cfd_ tol, fi_ info,
               fda_ wa, cfi_ lwa);

  /* lmder and lmder1 solve a system of n non-linear equations with m 
     parameters using the Levenberg-Marquardt algorithm. A least squares 
     solution is used. Jacobian is given analytically */
  void lmder_(void (*fcn)(int *m, int *n, double *x, double *fvec, 
                          double *fjac, int *ldfjac, int *iflag), 
              cfi_ m, cfi_ n, fda_ x, fda_ fvec, fda_ fjac, 
              cfi_ ldfjac, cfd_ ftol, cfd_ xtol, cfd_ gtol, cfi_ maxfev, 
              fda_ diag, cfi_ mode, cfd_ factor, cfi_ nprint, fi_ info, 
              fi_ nfev, fi_ njev, fia_ ipvt, fda_ qtf, fda_ wa1, fda_ wa2, 
              fda_ wa3, fda_ wa4);


  void lmder1_(void (*fcn)(int *m, int *n, double *x, double *fvec, 
                           double *fjac, int *ldfjac, int *iflag), 
               cfi_ m, cfi_ n, fda_ x, fda_ fvec, fda_ fjac, 
               cfi_ ldfjac, cfd_ tol, fi_ info, fia_ ipvt, fda_ wa, cfi_ lwa);

  /* lmdif and lmdif1 solve a system of n non-linear equations with m 
     parameters using the Levenberg-Marquardt algorithm. A least squares 
     solution is used. Jacobian is aproximated numerical */
  void lmdif_(void (*fcn)(int *m, int *n, double *x, double *fvec, int *iflag),
              cfi_ m, cfi_ n, fda_ x, fda_ fvec, cfd_ ftol, cfd_ xtol, 
              cfd_ gtol, cfi_ maxfev, cfd_ epsfcn, fda_ diag, cfi_ mode,
              cfd_ factor, cfi_ nprint, fi_ info, fi_ nfev, fda_ fjac,
              cfi_ ldfjac, fia_ ipvt, fda_ qtf, fda_ wa1, fda_ wa2, fda_ wa3,
              fda_ wa4);

  void lmdif1_(void (*fcn)(int *m, int *n, double *x, double *fvec, 
                           int *iflag), cfi_ m, cfi_ n, fda_ x, fda_ fvec,
               cfd_ tol, fi_ info, fia_ iwa, fda_ wa, cfi_ lwa, cfi_ maxfev);

#ifdef __cplusplus 
}
#endif

namespace BIAS{
  using namespace std;

  int Powell(void (*fun)(int *n, double *x, double *fvec, int *iflag), 
             Vector<double>& InitialGuess, Vector<double>& Result, 
             double Tolerance)
  {
    int info;
    int n=InitialGuess.Size();
    int lwa=(n*(3*n+13))/2;
    double *wa=new double[lwa];
    Result=InitialGuess;
    Vector<double> FuncVec(n);

    hybrd1_(fun, &n, Result.GetData(), FuncVec.GetData(), &Tolerance, 
           &info, wa, &lwa);

    delete[] wa;
    return (info==1)?(0):(info==0?-1:info);
  }

  int PowellExtended(void (*fun)(int *n, double *x, double *fvec, int *iflag), 
                     Vector<double>& InitialGuess, Vector<double>& Result, 
                     double EpsFcn, double Tolerance)
  {
    int info;
    int n=InitialGuess.Size();
    double *wa=new double[4*n];
    Result=InitialGuess;
    Vector<double> FuncVec(n);
    int maxfev = 200*(n + 1);
    int ml = n - 1;
    int mu = n - 1;
    int mode = 2;
    int lr = (n*(n + 1))/2;
    double factor=1.0;
    int nprint = 0;
    double *diag=new double[n];
    for (int i=0; i<n; i++) diag[i]=1.0;
    int nfev;
    double *fjac=new double[n*n];
    int ldfjac=n;
    double *r=new double[lr];
    double *qtf=new double[n];

    hybrd_(fun, &n, Result.GetData(), FuncVec.GetData(), &Tolerance, 
           &maxfev, &ml, &mu, &EpsFcn, diag, &mode, &factor, &nprint, 
           &info, &nfev, fjac, &ldfjac, r, &lr, qtf, &wa[0], &wa[n], 
           &wa[2*n], &wa[3*n]);
    
    delete[] wa;
    delete[] fjac;
    delete[] r;
    delete[] qtf;
    delete[] diag;
 
   return (info==1)?(0):(info==0?-1:info);
  }

  int LevenbergMarquardt(void (*fun)(int *m, int *n, double *x, 
                                                 double *fvec, double *fjac, 
                                                 int *ldfjac, int *iflag), 
                         int m, int n, Vector<double>& InitialGuess, 
                         Vector<double>& Result, double Tolerance)
  {
#ifdef BIAS_DEBUG
    if (m<n) {
      BIASERR("dimension of function output is bigger than input "
              <<m<<" "<<n);
    }
#endif
    int info;
    Result=InitialGuess;
    Vector<double> FuncVec(m);
    double *fjac=new double[m*n];
    int ldfjac=m;
    int *ipvt=new int[n];
    int lwa = 5*n+m;
    double *wa=new double[lwa];

    lmder1_(fun, &m, &n, Result.GetData(), FuncVec.GetData(), fjac, 
            &ldfjac, &Tolerance, &info, ipvt, wa, &lwa);

    delete[] fjac;
    delete[] ipvt;
    delete[] wa;

    return (info>0 && info<5)?(0):(info==0?-1:info);
  }

  int LevenbergMarquardtExtended(void (*fun)(int *m, int *n, double *x, 
                                             double *fvec, double *fjac, 
                                             int *ldfjac, int *iflag), 
                                 int m, int n, Vector<double>& InitialGuess, 
                                 Vector<double>& Result, double Tolerance)
 {
#ifdef BIAS_DEBUG
    if (m<n) {
      BIASERR("dimension of function output is bigger than input "
              <<m<<" "<<n);
      BIASASSERT(m>=n);
    }
#endif
    int info;
    Result=InitialGuess;
    Vector<double> FuncVec(m);
    int maxfev = 200*(n + 1);
    double ftol = Tolerance;
    double xtol = Tolerance;
    double gtol = 0.0;
    int mode = 1;
    int nprint = 0;
    double *diag=new double[n];
    double factor=1.0;
    int *iptv=new int[n];
    double *qtf=new double[n];
    double *wa=new double[m+3*n];
    int nfev, njfev;
    double *fjac=new double[m*n];
    int ldfjac=m;

    lmder_(fun, &m, &n, Result.GetData(), FuncVec.GetData(), fjac, &ldfjac, 
           &ftol, &xtol, &gtol, &maxfev, diag, &mode, &factor, &nprint, &info, 
           &nfev, &njfev, iptv, qtf, &wa[0], &wa[n], &wa[2*n], &wa[3*n]);
    
    delete[] diag;
    delete[] iptv;
    delete[] qtf;
    delete[] wa;
    delete[] fjac;

    return (info>0 && info<4)?(0):(info==0?-1:info);
  }

  int LevenbergMarquardt(void (*fun)(int *m, int *n, double *x, 
                                     double *fvec, int *iflag), 
                         int m, int n, Vector<double>& InitialGuess, 
                         Vector<double>& Result, double Tolerance, int MaxIter)
  {
#ifdef BIAS_DEBUG
    if (m<n) {
      BIASERR("dimension of function output is bigger than input "
              <<m<<" "<<n);
    }
#endif
    int info;
    Result=InitialGuess;
    Vector<double> FuncVec(m);
    int *iwa=new int[n];
    int lwa = m*n+5*n+m;
    double *wa=new double[lwa];
    
    lmdif1_(fun, &m, &n, Result.GetData(), FuncVec.GetData(), &Tolerance, 
            &info, iwa, wa, &lwa, &MaxIter);

    delete[] iwa;
    delete[] wa;

    return (info>0 && info<5)?(0):(info==0?-1:info);
  }


  int LevenbergMarquardtExtended(void (*fun)(int *m, int *n, double *x, 
                                             double *fvec, int *iflag), 
                                 int num_errors, int num_vars, 
                                 Vector<double>& InitialGuess, 
                                 Vector<double>& Result, double EpsFcn,
                                 double Tolerance, int MaxIter, double* fjac, 
                                 int* ipvt, double *SumOfSquaredErrors)
 {
#ifdef BIAS_DEBUG
    if (num_errors<num_vars) {
      BIASERR("number of params is bigger than number of error functions "
              <<num_errors<<" "<<num_vars);
    }
#endif
    int info;
    Result=InitialGuess;
    Vector<double> FuncVec(num_errors);
    int maxfev;
    if ( MaxIter==LM_DEF_MAX_ITER )
      maxfev = 200*(num_vars + 1);
    else
      maxfev = MaxIter;

    double ftol = Tolerance;
    double xtol = Tolerance;
    double gtol = 0.0;
    int mode = 1;
    int nprint = 0;
    double *diag=new double[num_vars];
    double factor=1.0;
    bool ExhaustiveData = 
      (fjac!=NULL) && (ipvt!=NULL) && (SumOfSquaredErrors!=NULL);
    if (!ExhaustiveData) {
      fjac = new double[num_vars*num_errors];
      ipvt = new int[num_vars];
    }


    int ldfjac  = num_errors;
    double *qtf = new double[num_vars];
    double *wa  = new double[num_errors+3*num_vars];
    int nfev;

    lmdif_(fun, &num_errors, &num_vars, Result.GetData(), 
           FuncVec.GetData(), &ftol, &xtol, 
           &gtol, &maxfev, &EpsFcn, diag, &mode, &factor, &nprint, &info, 
           &nfev, fjac, &ldfjac, ipvt, qtf, &wa[0], &wa[num_vars], 
           &wa[2*num_vars], &wa[3*num_vars]);

    delete[] diag;
    if (!ExhaustiveData) {
      delete[] fjac;
      fjac = NULL;
      delete[] ipvt;
      ipvt = NULL;
    } else {
      *SumOfSquaredErrors = 0;
      for (int i=0; i<num_vars; i++) *SumOfSquaredErrors += qtf[i]*qtf[i];
    }
   
    delete[] qtf;
    delete[] wa;

    return (info>0 && info<5)?(0):(info==0?-1:info);
  }
} // namespace
