/* 
   This file is part of the BIAS library (Basic ImageAlgorithmS).

   Copyright (C) 2003, 2004    (see file CONTACTS for details)
   Multimediale Systeme der Informationsverarbeitung
   Institut fuer Informatik
   Christian-Albrechts-Universitaet Kiel


   BIAS is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation; either version 2.1 of the License, or
   (at your option) any later version.

   BIAS is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with BIAS; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <Base/Image/Image.hh>
#include <Gui/GuiEnhanced.hh>
#include <Base/Image/ImageIO.hh>
#include <Base/Image/ImageConvert.hh>
#include <Image/Camera.hh>
#include <Image/Histogram.hh>
#include <stdio.h>
#include <unistd.h>
#include <getopt.h>
#include <Base/Debug/TimeMeasure.hh>
#include <bias_config.h>

#ifdef BIAS_DAIMLERCHRYSLER
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h> 
#include <fcntl.h>
#include <Base/Image/ImageDraw.hh>
#endif

using namespace BIAS;
using namespace std;

#define SPEED_STEP 10000


// JW
#define  DEFAULT_FIXED_SCALE true
#define  FIXED_MIN 0.0
#define  FIXED_MAX 25.0


void usage(char * name)
{
  cerr << endl << endl << name << " [options] filename ..."<<endl
       << " [-h]      help"<< endl
       << " [-v]      verbose"<<endl
       << " [-t n]    sleep n * 10-6s (n us) between frames" << endl
       << " [-f name] read images from file name"<<endl
       << " [-s]      start in movie mode"<<endl
       << " [-o]      quit after showing each image once"<<endl
       << " right mouse button to zoom"<<endl
       << endl << endl;
}



void showhistogram(ImageBase &img, GuiGTK & histgui)
{
  Histogram hist(2);
  Image<unsigned char> HistImg;
  if (ImageConvert::Convert(img, HistImg, ImageBase::CM_Grey)!=0){
    BIASERR("error converting");
  }
  hist.InitHist();
  hist.SetColor(0, 0, 0, 0);
  hist.SetDrawMode(HIST_BARS);
  hist.AddHist(HistImg);
  hist.Draw();
  histgui.ShowImage(hist);
}



//#include <stdio.h>  // for strcmp JW
void Scale(ImageBase& src, Image<unsigned char>& dst)
{
  switch(src.GetStorageType()){
  case ImageBase::ST_unsignedchar:
    {
      dst=src;
      unsigned char min, max;
      dst.GetMinMaxPixelValue(min, max);
      if (min!=max)
        dst.ScaleShiftBetween(dst.MinSTValue(), dst.MaxSTValue());
    }
    break;
#ifdef BUILD_IMAGE_USHORT
  case ImageBase::ST_unsignedshortint:
    {
      Image<unsigned short> src2(src);
      unsigned short min, max;
      src2.GetMinMaxPixelValue(min, max);
      if (min!=max)
        src2.ScaleShiftBetween(dst.MinSTValue(), dst.MaxSTValue());
      ImageConvert::ConvertST(src2, dst, ImageBase::ST_unsignedchar);
    }
    break;
#endif
  case ImageBase::ST_float:
    {
      Image<float> src2(src);
      float min, max;
      bool fixed=DEFAULT_FIXED_SCALE; // HACK scaleshift or fixed scale
      src2.GetMinMaxPixelValue(min, max);
      //if (argc>=2) if (strcmp(argv[1],"-fixed")==0) fixed=true;
      if (min!=max){
        if (fixed){
          cout<<"DBG using FIXED scale "<<FIXED_MIN<<" .. "
              <<FIXED_MAX<<" scale."<<endl;
          src2.ScaleShiftBetween(FIXED_MIN, FIXED_MAX);
        } else {
          cout<<"DBG using ScaleShiftbetween Min Max "<<endl;
          src2.ScaleShiftBetween(dst.MinSTValue(), dst.MaxSTValue());
        };
      }
      ImageConvert::ConvertST(src2, dst, ImageBase::ST_unsignedchar);
    }
    break;
  default:
    BIASERR("Support for storage type "<<src.GetStorageType()
            <<" not compiled in.");
    break;
  }
}

void PrintKeyCommands()
{
    cout <<"Key commands:"<<endl;
    cout <<" q/ESC : Quit"<<endl;
    cout <<" s     : slide show/film"<<endl;
    cout <<"SPC    : next image"<<endl;
    cout <<"BkSPC  : previous image"<<endl;
    cout <<" c     : color/grey"<<endl;
    cout <<" v     : print pixel value"<<endl;
    cout <<" ,/.   : decrease/increase film speed"<<endl;
    cout <<" h     : histogram"<<endl;
    cout <<" S     : toggle scaling of image values between 0 and 255 "<<endl;
    cout <<" p     : print header of image" << endl;
#ifdef BIAS_DAIMLERCHRYSLER
    cout <<" m     : create a file <file_name>.mo"<<endl;
    cout <<" n     : remove file <file_name>.mo"<<endl;
    cout <<" [     : decrease shift"<<endl;
    cout <<" ]     : increase shift"<<endl;
#endif
    //    cout <<"<     : downsample by 2"<<endl
    cout <<endl;
}

void ShiftAndMark(ImageBase& pic, int& shift, char **images, int nextfile)
{
#ifdef BIAS_DAIMLERCHRYSLER
#  ifdef BUILD_IMAGE_USHORT
      if (pic.GetStorageType()==ImageBase::ST_unsignedshortint){
        Image<unsigned short> dc_pic;
        dc_pic.StealImage(pic);
        // shift the images
        if (shift!=-1){
          //           cout << "shifting about "<<shift<<endl;
          unsigned short int *p, *e;
          p = dc_pic.GetImageData();
          e = p + dc_pic.GetPixelCount()*dc_pic.GetChannelCount();
          while (p<e) {
            *p = ((*p) >>shift);
            if ((*p) > 255) (*p) = 255;
            p++;
          }
        }

        string dc_name=images[nextfile];
        dc_name+=".mo";
        struct stat buf;
        if (stat(dc_name.c_str(), &buf)==0){
          // now mark all 4 corners white/black or red
          unsigned short mbval=(shift==-1)?(4095):(255);
          unsigned short bval[]={mbval, mbval, mbval};
          unsigned short sval[]={0, 0, 0};
          if (dc_pic.GetChannelCount()==3){
            sval[0]=mbval;
            bval[1]=bval[2]=0;
          } 
          
          int dc_size=20;
          int w=pic.GetWidth(), h=pic.GetHeight();

          dc_pic.SetROI(0, 0, dc_size, dc_size);
          dc_pic.FillImageWithConstValue(bval);
          ImageDraw<unsigned short>::
            RectangleCorners(dc_pic, 0, 0, dc_size, dc_size, sval);
          ImageDraw<unsigned short>::
            RectangleCorners(dc_pic, 1, 1, dc_size-1, dc_size-1, sval);
          dc_pic.SetROI(w-dc_size, 0 , w, dc_size);
          dc_pic.FillImageWithConstValue(sval);
          ImageDraw<unsigned short>::
            RectangleCorners(dc_pic, 0, h-dc_size-1, dc_size, h-1, bval);
          ImageDraw<unsigned short>::
            RectangleCorners(dc_pic, 1, h-dc_size-2, dc_size-1, h-2, bval);
          dc_pic.SetROI(0, h-dc_size, dc_size, h);
          dc_pic.FillImageWithConstValue(sval);
          ImageDraw<unsigned short>::
            RectangleCorners(dc_pic, 0, h-dc_size-1, dc_size, h-1, bval);
          ImageDraw<unsigned short>::
            RectangleCorners(dc_pic, 1, h-dc_size-2, dc_size-1, h-2, bval);
          dc_pic.SetROI(w-dc_size, h-dc_size, w, h);
          dc_pic.FillImageWithConstValue(bval);
          ImageDraw<unsigned short>::
            RectangleCorners(dc_pic, w-dc_size-1, h-dc_size-1, w-1, h-1, sval);
          ImageDraw<unsigned short>::
            RectangleCorners(dc_pic, w-dc_size-2, h-dc_size-2, w-2, h-2, sval);
        }

        pic.StealImage(dc_pic);
      }
#  endif
#endif
}

/** @class biasview
    @ingroup g_tools
    @brief tool for displaying images based on GTK
    @author woelk/evers 2001-2004 */
int main (int argc, char *argv[])
{
  int shift=2;
  //   TimeMeasure T1;

  GuiEnhanced win;
  
  //win.SetDebugLevel(D_GUI_MAIN | D_GUI_MUTEX |D_GUI_DESTRUCT|D_GUI_SIGNAL );

  bool keyevent=true, pressed=false, zpressed=false;
  int button=-1, x=-1, y=-1, zx, zy, zbutton, xoffset, yoffset;
  char title[200];
  char c, zc;
  int files;
  int nextfile;
  int offset, count=0;
  long int speed = 10001;
  bool done = false;
  bool loadnewimage = true;
  bool slideshow = false;
  bool color = true;
  bool quiet = true;
  bool showscaled = false;
  //  unsigned char min, max; // minimal and maximal pixel values in displayImg

  ImageBase pic; // the original image
  //Image<unsigned char> zoomimage; // the zoomimage
  ImageBase zoomimage;
  // image values are scaled to a maximal contrast 
  Image<unsigned char> scaledimage; 
  ImageBase displayImg;

  GuiGTK zoomgui;
  char zoomtitle[1024];
  int zoomfactor =10;
  int zoomsize = 20;
  bool zoom=false;
  bool zoomevent, winevent, zkeyevent;

  GuiGTK histgui;
  bool showhist  = false;
  bool once=false; // only cycle onec through the images and exit
  

  struct ::option opts[] = {
    {"geometry", required_argument, 0, 'g'},
    {0, 0, 0, 0}
  };
  char **images=NULL;

  ifstream is;
  char name[1024];

  if (argc <2) {
    usage(argv[0]);
    //exit(1);
    return 0; // for Dart testing 
  }

  images = argv;

  while ((c = getopt_long_only(argc, argv, "vhg:st:f:o", opts, 0)) != EOF)
    {
      switch (c){
      case 'f':
        //  	cerr << optarg << endl;
        is.open(optarg);
        count = 0;
        while (!is.eof()){
          is >> name;
          //  	  cerr << "countig... " << name << endl;
          count++;
        }	
        //  	cerr << count << endl;
        is.clear();
        is.seekg(ios::beg); 
        images = new char*[argc+count+1];
        count=0;
        while (!is.eof()){
          is >> name;
          //  	  cerr << "reading.. " << name << endl;
          images[argc+count] = new char[strlen(name)+1];
          strcpy(images[argc+count], name);
          count++;
        };
        is.close();
        break;
      case 'v':
        quiet = false;
        break;
      case 'h':
        usage(argv[0]);
        return 0;
        break;
      case 't':
        speed = atol(optarg);
        //	cerr << "speed " << speed << endl;
        break;
      case 's':
        slideshow = true;
        //        cerr << "slideshow found" << endl;
        break;
      case 'o':
        once=true;
        break;
      default:
        break;
      }
    }
  

  offset = optind;
  files = argc + count - offset;
  nextfile = optind;
  
  //cerr << argc << " " << offset << " " << files << endl;
  //    for (int i=0; i<argc; i++){
  //      cerr << images[i] << endl;
  //    }
  
  if (!quiet) PrintKeyCommands();
 
  while (!done){ 
    
    if (loadnewimage) {
      showscaled = false;
      if (ImageIO::Load(images[nextfile], pic)!=0){
        BIASERR("Error loading "<<images[nextfile]<<endl
                <<"Displaying black image.");
        Image<unsigned char> tmppic(640,480,1);
        tmppic.FillImageWithConstValue((unsigned char)0);
        pic = tmppic;
      }
      
      sprintf(title,"%s",images[nextfile]);
      win.SetTitle(title);     
      if (!quiet) {
        cout <<"=== "<<title<<" === "<<pic.GetStorageType()<<" : "<<
          pic.GetColorModel()<< endl;
        cout << "found " << pic.GetMetaData()->size() 
             << " different meta datas." << endl;
        if (pic.GetMetaData()->size() >0){
          Camera<unsigned char> cam(pic);         
          //cout << "found " << cam.GetMetaData()->size() 
          //     << " different meta datas in camera." << endl;
          //cam.SetDebugLevel(cam.GetDebugLevel() | D_CAMERA_METADATA);
          cam.ParseMetaData();
          //cam.GetMetaData()->Dump();
          cam.PrintAppData(cout);
        }
      }

      ShiftAndMark(pic, shift, images, nextfile);

      if (!displayImg.IsEmpty()) displayImg.Release();
      if (! color || (pic.GetChannelCount()==1 
                      && pic.GetColorModel()!=ImageBase::CM_DV
                      && pic.GetColorModel()!=ImageBase::CM_Bayer_RGGB
                      && pic.GetColorModel()!=ImageBase::CM_Bayer_GBRG
                      && pic.GetColorModel()!=ImageBase::CM_Bayer_GRBG
                      && pic.GetColorModel()!=ImageBase::CM_Bayer_BGGR)){
        if (ImageConvert::Convert(pic, displayImg, ImageBase::CM_Grey)!=0){
          BIASERR("error converting");
        }
      } else { 
        if (ImageConvert::Convert(pic, displayImg, ImageBase::CM_RGB)!=0){
          BIASERR("error converting");
        }
      }

      Scale(pic, scaledimage);
     
      if (showscaled){
        //cerr << "showing scaled image" <<endl;
        win.ShowImage(scaledimage);
      } else {
        //cerr << "showing displayImg "<<displayImg.GetStorageType() <<endl;
        win.ShowImage(displayImg);
      }
      loadnewimage = false;
      zoomgui.DestroyWindow();
      zoom=false;
      if (showhist)
        showhistogram(displayImg,histgui);
      
    }      
     
    if (!slideshow) {
      if (zoom) {
        zoomevent=winevent=false;
        while (!winevent){
          winevent = 
            win.CheckForEvent(keyevent, c, pressed, button, x, y);
          if (zoomgui.CheckForEvent(zkeyevent, zc, zpressed, zbutton, zx, zy)){
            if (zkeyevent){
              switch(zc){
              case 'v':{
                int x, y;
                int xoff, yoff, fac;
                zoomgui.GetFactorOffset(xoff, yoff, fac);
                x = zoomgui.GetMouseX()/fac+xoff;
                y = zoomgui.GetMouseY()/fac+yoff;
                if (x<(int)pic.GetWidth() && y<(int)pic.GetHeight()){
                  cout <<"Value at: ("<<x<<","<<y<<") : ";
                  switch (pic.GetStorageType()){
                  case ImageBase::ST_unsignedchar: 
                    for (unsigned int c=0; c<pic.GetChannelCount(); c++)
                      cout << (int)(((unsigned char **)pic.GetImageDataArray())
                                    [y][x*pic.GetChannelCount()+c])<<" ";
                    break;
                  case ImageBase::ST_float: 
                    cout << ((float **)pic.GetImageDataArray())[y][x];
                    break;
#ifdef BUILD_IMAGE_UINT
                  case ImageBase::ST_unsignedint: 
                    cout << ((unsigned int **)pic.GetImageDataArray())[y][x];
                    break;
#endif
                  default:
                    break;
                  }
                  cout <<endl;
                }
              }
                break;
              default:
                cout << zc << endl;
                break;
              }
            }
          } else {
            usleep(10001);
          }
        }
        

      } else {
        win.WaitForEvent(keyevent, c, pressed, button, x, y);
      }
    } else {
      c = win.CheckForKeyEvent();
      usleep(speed);
    }

    if (keyevent){
      //cerr << "keyevent: "<<c<<endl;
      switch (c) {
      case ',':
        if (speed - SPEED_STEP > 0)
          speed -= SPEED_STEP;
        break;
      case '.':
        speed += SPEED_STEP;
        break;
      case 'q': 
      case 27: 
        done = true;
        break;
      case ' ':
        nextfile ++;
        if (nextfile >= files+offset) nextfile = offset;
        loadnewimage = true;
        break;
      case 8:
        nextfile --;
        if (nextfile < offset ) nextfile = files + offset-1;
        loadnewimage = true;
        break;
      case 's':
        slideshow = !slideshow;
        break;
      case 'c':
        color = !color;
        loadnewimage = true;
        break;
      case 'h':
        showhist = !showhist;
        if (showhist) 
          showhistogram(displayImg,histgui);
        else
          histgui.DestroyWindow();
        break;
      case 'p': 
        cout << "image header: "<< endl;
        pic.PrintHeader(std::cout);
        cout << endl;
        break;
      case 'v':
        int x,y;
        x = win.GetMouseX();
        y = win.GetMouseY();
        cout <<"Value at: ("<<x<<","<<y<<") : ";
        switch (pic.GetStorageType()){
        case ImageBase::ST_unsignedchar: 
          for (unsigned int c=0; c<pic.GetChannelCount(); c++)
            cout << (int)(((unsigned char **)pic.GetImageDataArray())
                          [y][x*pic.GetChannelCount()+c])<<" ";
          break;
        case ImageBase::ST_float: 
          cout << ((float **)pic.GetImageDataArray())[y][x];
          break;
#ifdef BUILD_IMAGE_UINT
        case ImageBase::ST_unsignedint: 
          cout << ((unsigned int **)pic.GetImageDataArray())[y][x];
          break;
#endif
        default:
          break;
        }
        cout <<endl;
        break;
      case 'S':
        if (pic.GetStorageType() != ImageBase::ST_unsignedchar){
          cerr << "'S' scale image only reasonable with unsigned char image"
               <<endl;
        }
        if (showscaled){
          showscaled=false;
          win.ShowImage(displayImg);
        } else {
          showscaled=true;
          win.ShowImage(scaledimage);
        }
        break;
      case 'k':
        PrintKeyCommands();
        break;
#ifdef BIAS_DAIMLERCHRYSLER
      case 'm':
        {
          string dc_name=images[nextfile];
          dc_name+=".mo";
          int mres=open(dc_name.c_str(), O_RDWR | O_CREAT, 644);
          if (mres==-1){
            BIASERR("error creating "<<dc_name);
            perror("");
          } else {
            cout << "wrote "<<dc_name<<endl;
          }
          nextfile ++;
          if (nextfile >= files+offset) nextfile = offset;
          loadnewimage = true;
        }
        break;
      case 'n':
        {
          string dc_name=images[nextfile];
          dc_name+=".mo";
          int mres=unlink(dc_name.c_str());
          if (mres!=0){
            BIASERR("error removing "<<dc_name);
            perror("");
          } else {
            cout << "removed "<<dc_name<<endl;
          }
          nextfile ++;
          if (nextfile >= files+offset) nextfile = offset;
          loadnewimage = true;
        }
        break;
      case '[':
        shift-=1;
        if (shift<-1) shift=-1;
        cout << "shift = "<<shift<<endl;
        loadnewimage = true;
        break;
      case ']':
        shift+=1;
        if (shift>4) shift=4;
        cout << "shift = "<<shift<<endl;
        loadnewimage = true;
        break;
#endif
      default:
        if (!quiet)
          cout <<"Taste: "<<(int)c<<endl;
        break;
      }
    } else { // button event
      //cerr << "buttonevent: "<<button<<"   at ("<<x<<", "<<y<<")"<<endl;
      if (pressed && button == 3){
        win.GetZoomImage(x, y, zoomfactor, zoomsize, zoomimage, 
                         xoffset, yoffset);
        sprintf(zoomtitle,"%s: zoomfactor (%i, %i)", images[nextfile], x, y);
        zoomgui.SetTitle(zoomtitle);
        zoomgui.SetFactorOffset(xoffset, yoffset, zoomfactor);
        zoomgui.ShowImage(zoomimage);
        zoom=true;
      }
      keyevent = true;
      c = -1;
      pressed = false;
      button = -1;
      x = y = -1;
    }
    

    if (slideshow) {
      nextfile ++;
      if (nextfile >= files+offset) {
        if (once){
          done=true;
        } else {
          nextfile = offset;
        }
      }
      loadnewimage = true;
    }
    if (win.Closed()) { 
      //cerr <<"closed"<<endl; 
      done=true; 
    } else { 
      //cerr <<"not closed"<<endl; 
    }
  } //  while (!done)
  return 0;
}
